{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
  :: QuickReport 3.0 for Delphi and C++Builder               ::
  ::                                                         ::
  :: QRCTRLS.PAS - PRINTABLE CONTROLS                        ::
  ::                                                         ::
  :: Copyright (c) 1999 QuSoft AS                            ::
  :: All Rights Reserved                                     ::
  ::                                                         ::
  :: web: http://www.qusoft.no                               ::
  ::                                                         ::
  :: Updated by Bugs Free QR Team 2001-02                    ::
  :: http://bfqr.da.ru                                       ::
  ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

{$I QRDEFS.INC}

unit Qrctrls;

interface

uses messages, windows, classes, controls, stdctrls, sysutils, graphics,{RSV buttons,}
     forms,{RSV extctrls, dialogs,} printers, db, {RAA DBtables, }ComCtrls, RichEdit,
     QRPrntr, Quickrpt, QR3Const, qrexpr;

{$R-}
{$B-}

type
  { Forward declarations }
  TQRExpr = class;

  { TQRCustomLabel - base class for printable text components }

  TQRLabelOnPrintEvent = procedure (sender : TObject; var Value : string) of object;

  TQRIndentSpace = class(TPersistent)
  private
    FOwner: TQRPrintable;
    FRect: TRect;
    function GetValue(const Index: Integer): Integer;
    procedure SetValue(const Index: Integer; Value: Integer);
    function GetExtValue(const Index: Integer): Extended;
  protected
    procedure Invalidate;
  public
    constructor Create(AOwner: TQRPrintable);
    destructor Destroy; override;
    function IsDefault: Boolean;
    property LeftWidth:    Extended index 0 read GetExtValue;
    property RightWidth:   Extended index 1 read GetExtValue;
    property TopHeight:    Extended index 2 read GetExtValue;
    property BottomHeight: Extended index 3 read GetExtValue;
  published
    property Left:   Integer index 0 read GetValue write SetValue default 1;
    property Right:  Integer index 1 read GetValue write SetValue default 1;
    property Top:    Integer index 2 read GetValue write SetValue default 0;
    property Bottom: Integer index 3 read GetValue write SetValue default 0;
  end;

  TQRCustomLabel = class(TQRrPrintable)
  private
//>RSV
    FDegree: Integer;
    FIndent: TQRIndentSpace;
//<RSV
    FCaptionBased : boolean;
    DoneFormat : boolean;
    FAutoSize : boolean;
    FAutoStretch : boolean;
//RSV    FCaption : string;
    FCaption : TCaption;

    FCurrentLine : integer;
    FPrintCaption : string;
    FFontSize : integer;
    FFormattedLines : TStrings;
    FLines : TStrings;
    FOnPrint : TQRLabelOnPrintevent;
    FWordWrap : boolean;
    UpdatingBounds : boolean;
//>RSV
    procedure SetDegree(Value: Integer);
    procedure SetIndent(Value: TQRIndentSpace);
//<RSV
//RSV    function GetCaption : string;
//RSV    procedure SetCaption(Value : string);
    function GetCaption : TCaption;
    procedure SetCaption(Value : TCaption);
    function GetCaptionBased : boolean; virtual;
    procedure SetAutoStretch(Value : boolean);
//RAA    procedure SetAutoSize(Value: Boolean);
    procedure SetLines(Value : TStrings);
    procedure SetWordWrap(Value : boolean);
    procedure PaintToCanvas(aCanvas : TCanvas; aRect : TRect; CanExpand : boolean; LineHeight : integer);
    procedure PrintToCanvas(aCanvas : TCanvas; aLeft, aTop, aWidth, aHeight, LineHeight : extended;
                            CanExpand : boolean);
//>RSV
   procedure DegreePaint;
   procedure DegreePrint(OfsX, OfsY: Extended);
//<RSV
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
//    procedure SetIndent(const Value: Integer);
    function SaveIndents: Boolean;
  protected
    procedure FormatLines; virtual;
    procedure Loaded; override;
    procedure SetName(const Value: TComponentName); override;
    procedure SetParent(AParent: TWinControl); override;
    procedure DefineProperties(Filer: TFiler); override;
    procedure ReadFontSize(Reader : TReader); virtual;
    procedure WriteFontSize(Writer : TWriter); virtual;
    procedure Paint; override;
    procedure Prepare; override;
    procedure UnPrepare; override;
//rsv    procedure Print(OfsX, OfsY : integer); override;
    function StoredFont: Boolean;
    procedure SetAutoSize(Value: Boolean); {$IFDEF BFQR_D6} override; {$ENDIF} //RAA
    procedure Print(OfsX, OfsY : extended); override;
    procedure SetAlignment(Value : TAlignment); override;
    property OnPrint : TQRLabelOnPrintEvent read FOnPrint write FOnPrint;
    property Degree: Integer read FDegree write SetDegree default 0;
  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    function GetControlsAlignment: TAlignment; override;
    property CaptionBased : boolean read GetCaptionBased;
    property Alignment default taLeftJustify;
    property AutoSize : boolean read FAutoSize write SetAutoSize default True;
    property AutoStretch : boolean read FAutoStretch write SetAutoStretch default False;
//    property Caption : string read GetCaption write SetCaption stored true;
    property Caption : TCaption read GetCaption write SetCaption stored true;
    property Color default clWhite;
    property Font stored StoredFont;
    property Lines : TStrings read FLines write SetLines;
    property WordWrap : boolean read FWordWrap write SetWordWrap default True;
//    property Indent: Integer read FIndent write SetIndent default 0;
    property Indent: TQRIndentSpace read FIndent write SetIndent stored SaveIndents;
    // TSK
    property FormattedLines: TStrings read FFormattedLines;
  end;

  { TQRLabel - printable component with published Caption property }
  TQRLabel = class(TQRCustomLabel)
  protected
{$ifdef USE_QREDITOR}  //RAA
    function GetEditorClass : TQRPrintableEditorClass; override;
{$endif}
  public
  published
    property Alignment;
    property AlignToBand;
    property AutoSize;
    property AutoStretch;
    property BiDiMode;
    property ParentBiDiMode;
    property Caption;
    property Color;
    property Font;
    property OnPrint;
    property ParentFont;
    property Transparent;
    property WordWrap;
    property Indent;
    property Degree;
  end;

  { TQRMemo - printable memo component (published Lines property) }
  TQRMemo = class(TQRCustomLabel)
  protected
    function GetCaptionBased : boolean; override;
  public
    procedure Paint; override;
//rsv    procedure Print(OfsX, OfsY : integer); override;
    procedure Print(OfsX, OfsY : extended); override;
  published
    property Alignment;
    property AlignToBand;
    property AutoSize;
    property AutoStretch;
    property BiDiMode;
    property ParentBiDiMode;
    property Color;
    property Font;
    property Lines;
    property ParentFont;
    property Transparent;
    property WordWrap;
    property Indent;
    property Degree;
  end;

  { TQRDBText }
  TQRDBText = class(TQRCustomLabel)
  private
    FBlankIfZero: Boolean;
    ComboBox : TEdit;
    Field : TField;
    FieldNo : integer;
    FieldOK : boolean;
    DataSourceName : string[30];
    FDataSet : TDataSet;
    FDataField : string;
    FMask : string;
    IsMemo : boolean;
    procedure SetDataSet(Value : TDataSet);
    procedure SetDataField(Value : string);
    procedure SetMask(Value : string);
  protected
    function GetCaptionBased : boolean; override;
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Prepare; override;
//rsv    procedure Print(OfsX, OfsY : integer); override;
    procedure Print(OfsX, OfsY : extended); override;
    procedure UnPrepare; override;
  public
    constructor Create(AOwner : TComponent); override;
    function    UseRightToLeftAlignment: boolean; override;
    property    Fields: TField read Field;
  published
    property Alignment;
    property AlignToBand;
    property AutoSize;
    property AutoStretch;
    property BiDiMode;
    property ParentBiDiMode;
    property Color;
    property DataSet : TDataSet read FDataSet write SetDataSet;
    property DataField : string read FDataField write SetDataField;
    property Font;
    property Mask : string read FMask write SetMask;
    property OnPrint;
    property ParentFont;
    property Transparent;
    property WordWrap;
    property Indent;
    property Degree;
    property BlankIfZero: Boolean read FBlankIfZero write FBlankIfZero default False;
  end;

  { TQRExpr }
  TQRExpr = class(TQRCustomLabel)
  private
// TSK    Evaluator : TQREvaluator;
    FExpression : string;
    FMask : string;
    FMaster : TComponent;
    FResetAfterPrint : boolean;
    FCurrency: Boolean;
    function  GetValue : TQREvResult;
    procedure SetExpression(Value : string);
    procedure SetMask(Value : string);
    procedure SetCurrency(const Value: Boolean);
  protected
    procedure Prepare; override;
    procedure UnPrepare; override;
    procedure QRNotification(Sender : TObject; Operation : TQRNotifyOperation); override;
//rsv    procedure Print(OfsX, OfsY : integer); override;
    procedure Print(OfsX, OfsY : extended); override;
    procedure SetMaster(AComponent : TComponent);
{$ifdef USE_QREDITOR}  //RAA
    function GetEditorClass : TQRPrintableEditorClass; override;
{$endif}
  public
    Evaluator : TQREvaluator;
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    procedure Reset;
    property Value : TQREvResult read GetValue;
  published
    property Alignment;
    property AlignToBand;
    property AutoSize;
    property AutoStretch;
// TSK
    property Currency: Boolean read FCurrency write SetCurrency default False;
    property BiDiMode;
    property ParentBiDiMode;
    property Font;
    property Color;
    property Master : TComponent read FMaster write SetMaster;
    property OnPrint;
    property ParentFont;
    property ResetAfterPrint : boolean read FResetAfterPrint write FResetAfterPrint;
    property Transparent;
    property WordWrap;
    property Expression : string read FExpression write SetExpression;
    property Mask : string read FMask write SetMask;
    property Indent;
    property Degree;
  end;

  { TQRSysData }
  TQRSysDataType = (qrsTime,
                    qrsDate,
                    qrsDateTime,
                    qrsPageNumber,
                    qrsReportTitle,
                    qrsDetailCount,
                    qrsDetailNo);

  TQRSysData = class(TQRCustomLabel)
  private
    FData : TQRSysDataType;
    FText : string;
    procedure SetData(Value : TQRSysDataType);
    procedure SetText(Value : string);
    procedure CreateCaption;
  protected
//rsv    procedure Print(OfsX, OfsY : integer); override;
    procedure Print(OfsX, OfsY : extended); override;
  public
    constructor Create(AOwner : TComponent); override;
  published
    property Alignment;
    property AlignToBand;
    property AutoSize;
//>RSV
    property AutoStretch;
//<RSV
    property BiDiMode;
    property ParentBiDiMode;
    property Color;
    property Data : TQRSysDataType read FData write SetData;
    property Font;
    property OnPrint;
    property ParentFont;
    property Text : string read FText write SetText;
    property Transparent;
    property Indent;
    property Degree;
  end;

  { TQRShape }
//RSV  TQRShapeType = (qrsRectangle,qrsCircle,qrsVertLine,qrsHorLine,qrsTopAndBottom,qrsRightAndLeft);
  TQRShapeType = (qrsRectangle,qrsCircle,qrsVertLine,qrsHorLine,
                   qrsTopAndBottom,qrsRightAndLeft,
                   qrsDownLine, qrsUpLine,
                   qrsTriangle1,qrsTriangle2,qrsTriangle3,qrsTriangle4);


  TQRShape = class(TQRrPrintable)
  private
    FShape : TQRShapeType;
    FBrush : TBrush;
    FPen : TPen;
    procedure SetBrush(Value : TBrush);
    procedure SetPen(Value : TPen);
    procedure SetShape(Value : TQRShapeType);
  protected
    procedure Paint; override;
//rsv    procedure Print(OfsX, OfsY : integer); override;
    procedure Print(OfsX, OfsY : extended); override;
    procedure StyleChanged(sender : TObject);
  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
  published
    property Brush : TBrush read FBrush write SetBrush;
    property Height default 65;
    property Pen : TPen read FPen write Setpen;
    property Shape : TQRShapeType Read FShape write SetShape default qrsRectangle;
    property Width default 65;
  end;

  { TQRImage }
  TQRImage = class(TQRrPrintable)
  private
    FPicture: TPicture;
    FAutoSize: Boolean;
    FStretch: Boolean;
    FCenter: Boolean;
    function GetCanvas: TCanvas;
    procedure PictureChanged(Sender: TObject);
//RAA    procedure SetAutoSize(Value: Boolean);
    procedure SetCenter(Value: Boolean);
    procedure SetPicture(Value: TPicture);
    procedure SetStretch(Value: Boolean);
  protected
    function GetPalette: HPALETTE; override;
    procedure Paint; override;
//rsv    procedure Print(OfsX, OfsY : integer); override;
    procedure Print(OfsX, OfsY : extended); override;
    procedure SetAutoSize(Value: Boolean); {$IFDEF BFQR_D6} override; {$ENDIF} //RAA
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Canvas: TCanvas read GetCanvas;
  published
    property AutoSize: Boolean read FAutoSize write SetAutoSize default False;
    property Center: Boolean read FCenter write SetCenter default False;
    property Picture: TPicture read FPicture write SetPicture;
    property Stretch: Boolean read FStretch write SetStretch default False;
  end;

  { TQRDBImage }
  TQRDBImage = class(TQRrPrintable)
  private
    FField : TField;
    FDataSet : TDataSet;
    FDataField : string;
    FPicture: TPicture;
    FStretch: boolean;
    FCenter: boolean;
    FPictureLoaded: boolean;
    procedure PictureChanged(Sender: TObject);
    procedure SetCenter(Value: Boolean);
    procedure SetDataField(const Value: string);
    procedure SetDataSet(Value: TDataSet);
    procedure SetPicture(Value: TPicture);
    procedure SetStretch(Value: Boolean);
  protected
    function GetPalette: HPALETTE; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Paint; override;
    procedure Prepare; override;
//rsv    procedure Print(OfsX, OfsY : integer); override;
    procedure Print(OfsX, OfsY : extended); override;
    procedure UnPrepare; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure LoadPicture;
    property Field: TField read FField;
    property Picture: TPicture read FPicture write SetPicture;
  published
    property Center: boolean read FCenter write SetCenter default True;
    property DataField: string read FDataField write SetDataField;
    property DataSet: TDataSet read FDataSet write SetDataSet;
    property Stretch: boolean read FStretch write SetStretch default False;
  end;


  { TQRRichEdit - TQRCustomRichEdit descendant with some special settings }
  TQRRichEdit = class(TCustomRichEdit)
  public
    property BorderStyle;
  end;

  { TQRRichText }
  TQRCustomRichText = class(TQRrPrintable)
  private
    LastChar : integer;
    FAutoStretch : boolean;
    FParentRichEdit : TRichEdit;
    FRichEdit : TQRRichEdit;
    function GetAlignment : TAlignment;
    function GetColor : TColor;
    function GetFont : TFont;
    function GetLines : TStrings;
//RSV    procedure SetAlignment(Value : TAlignment); override;
    procedure SetColor(Value : TColor);
    procedure SetFont(Value : TFont);
    procedure SetLines(Value : TStrings);
    procedure SetParentRichEdit(Value : TRichEdit);
  protected
    property Lines : TStrings read GetLines write SetLines;
    property ParentRichEdit : TRichEdit read FParentRichEdit write SetParentRichEdit;
//rsv    procedure Print(OfsX, OfsY : integer); override;
    procedure Print(OfsX, OfsY : extended); override;
//>RSV
    procedure SetAlignment(Value : TAlignment); override;
//<RSV
  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight : integer); override;
    function GetControlsAlignment: TAlignment; override;
  published
    property Alignment : TAlignment read GetAlignment write SetAlignment;
    property AutoStretch : boolean read FAutoStretch write FAutoStretch;
    property Color : TColor read GetColor write SetColor;
    property Font : TFont read GetFont write SetFont;
  end;

  { TQRRichText }
  TQRRichText = class(TQRCustomRichText)
  published
    property Lines;
    property ParentRichEdit;
  end;

  { TQRDBRichText }
  TQRDBRichText = class(TQRCustomRichText)
  private
    Field:      TField;
    FDataField: String;
    FDataSet:   TDataSet;
  protected
    procedure   Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure   SetDataSet(Value : TDataSet);
    procedure   Prepare; override;
    procedure   UnPrepare; override;
//rsv    procedure Print(OfsX, OfsY : integer); override;
    procedure   Print(OfsX, OfsY : extended); override;
  published
    property  DataField: string read FDataField write FDataField;
    property  DataSet: TDataSet read FDataSet write SetDataSet;
  end;

  { TQRExprMemo }
  TQRMerger = class
  private
    FOrgLines : TStrings;
    FStrippedLines : TStrings;
    FMergedLines : TStrings;
    FMerged : boolean;
    FPrepared : boolean;
    Expressions : TList;
    FDataSets : TList;
  protected
    function  GetOrgLines : TStrings;
    function  GetMergedLines : TStrings;
    procedure SetOrgLines(Value : TStrings);
  public
    constructor Create;
    destructor  Destroy; override;
    procedure   Prepare;
    procedure   Merge;
    procedure Unprepare;
    property Lines : TStrings read GetOrgLines write SetOrgLines;
    property MergedLines : TStrings read GetMergedLines;
    property Merged : boolean read FMerged;
    property Prepared : boolean read FPrepared;
    property DataSets :TList read FDataSets write FDataSets;
  end;

  TQRExprMemo = class(TQRCustomLabel)
  private
    Merger : TQRMerger;
    FRemoveBlankLines : boolean;
  protected
    function GetCaptionBased : boolean; override;
    procedure Prepare; override;
    procedure UnPrepare; override;
//rsv    procedure Print(OfsX, OfsY : integer); override;
    procedure Print(OfsX, OfsY : extended); override;
    procedure Paint; override;
  public
    constructor Create(AOwner : TComponent); override;
  published
    property RemoveBlankLines : boolean read FRemoveBlankLines write FRemoveBlankLines;
    property Alignment;
    property AlignToBand;
    property AutoSize;
    property AutoStretch;
    property Color;
    property Font;
    property Lines;
    property ParentFont;
    property Transparent;
    property WordWrap;
    property Degree;
  end;

{ TQRDBCalc - included for backwards compatibility }

  TQRCalcOperation = (qrcSum, qrcCount, qrcMax, qrcMin, qrcAverage);

  TQRDBCalc = class(TQRExpr)
  private
    FDataField : string;
    FDataSource : TDataSource;
    FOperation : TQRCalcOperation;
    FResetBand : TQRBand;
  protected
    function GetPrintMask : string;
    procedure SetDataField(Value : string);
    procedure SetOperation(Value : TQRCalcOperation);
    procedure SetPrintMask(Value : string);
  published
    property DataField : string read FDataField write SetDataField;
    property DataSource : TDataSource read FDataSource write FDataSource;
    property OnPrint;
    property Operation : TQRCalcOperation  read FOperation write SetOperation;
    property ParentFont;
    property PrintMask : string read GetPrintMask write SetPrintMask;
    property ResetBand : TQRBand read FResetBand write FResetBand;
  end;

{TQRCheckBox}
  TQRCheckmarkType = (cmCross, cmHook, cmGlyph);

  TQRCheckBox = class(TQRrPrintable)
  private
    FExpression              : String;
    FBrush                   : TBrush;
    FPen                     : TPen;
    FCtl3d                   : Boolean;
    FChecked                 : Boolean;
    FCheckmarkDistance       : Word;
    FCheckmarkType           : TQRCheckmarkType;
    FGlyphChecked            : TPicture;
    FGlyphUnchecked          : TPicture;
    procedure SetChecked(Value: Boolean);
    procedure SetBrush(Value: TBrush);
    procedure SetPen(Value: TPen);
    procedure SetCtl3d(Value: Boolean);
    procedure SetCheckmarkDistance(Value: Word);
    procedure SetCheckmarkType(Value: TQRCheckmarkType);
    procedure SetGlyphChecked(Value: TPicture);
    procedure SetGlyphUnchecked(Value: TPicture);
  protected
    procedure StyleChanged(Sender : TObject);
    procedure PictureChanged(Sender: TObject);
  public
    constructor Create(AOwner : TComponent); override;
    procedure Paint; override;
    procedure Print(OfsX, OfsY : Extended); override;
    destructor Destroy; override;
  published
    property Expression: String Read FExpression Write FExpression;
    property GlyphChecked: TPicture read FGlyphChecked write SetGlyphChecked;
    property GlyphUnchecked: TPicture read FGlyphUnchecked write SetGlyphUnchecked;
    property CheckmarkType: TQRCheckmarkType read FCheckmarkType write SetCheckmarkType default cmCross;
    property CheckmarkDistance : Word read FCheckmarkDistance Write SetCheckmarkDistance default 3;
    property Checked : Boolean read FChecked write SetChecked default True;
    property Brush : TBrush read FBrush write SetBrush;
    property Pen : TPen read FPen write Setpen;
    property Height default 17;
    property Width default 17;
    property Transparent;
    property Ctl3D: Boolean Read FCtl3d Write SetCtl3d;
  end;

{TQRGrid}
  TVertAlignment = (vaNone, vaTop, vaCenter, vaBottom);
  THorizAlignment  = (haNone, haLeft, haCenter, haRight);

  TQRGridCell = class
    Caption        : String;
    Font           : TFont;
    ParentFont     : Boolean;
    VertAlignment  : TVertAlignment;
    HorizAlignment : THorizAlignment;
    constructor Create;
    destructor Destroy; override;
  end;

  TQRGrid = class(TQRrPrintable)
  private
    FCellHorizAlignment: THorizAlignment;
    FCellVertAlignment : TVertAlignment;
    FBrush             : TBrush;
    FPen               : TPen;
    FRows              : Word;
    FColumns           : Word;
    FDrawFirstRowLine  : Boolean;
    FFirstRowHeight    : Word;
    FFirstColumnWidth  : Word;
    FFirstRowColor     : TColor;
    FFirstColumnColor  : TColor;
    FFirstColumnJoined : Boolean;
    FFirstRowJoined    : Boolean;
    FFirstCellColor    : TColor;
    FColumnSizes       : TStrings;
    FRowSizes          : TStrings;
    {FOnDrawGridCell   : TQRPOnDrawGridCellEvent;}
    FCellColumns       : TList;
    procedure SetBrush(Value : TBrush);
    procedure SetPen(Value : TPen);
    procedure SetRows(Value: Word);
    procedure SetColumns(Value: Word);
    {procedure SetFirstRowHeight(Value: Word);
    procedure SetFirstColumnWidth(Value: Word);}
    procedure SetFirstRowColor(Value: TColor);
    procedure SetFirstCellColor(Value: TColor);
    procedure SetFirstColumnColor(Value: TColor);
    procedure SetFirstColumnJoined(Value: Boolean);
    procedure SetFirstRowJoined(Value: Boolean);
    procedure SetDrawFirstRowLine(Value: Boolean);
    procedure SetRowSizes(Value: TStrings);
    procedure SetColumnSizes(Value: TStrings);
    function GetCells(ACol, ARow: Integer): TQRGridCell;
    procedure SetCells(ACol, ARow: Integer; const Value: TQRGridCell);
  protected
    procedure Paint; override;
//    procedure Print(OfsX, OfsY: Integer); override;
    procedure Print(OfsX, OfsY: Extended); override;
    procedure StyleChanged(sender : TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Cells[ACol, ARow: Integer]: TQRGridCell Read GetCells Write SetCells;
  published
    property CellHorizAlignment: THorizAlignment Read FCellHorizAlignment Write FCellHorizAlignment default haNone;
    property CellVertAlignment: TVertAlignment Read FCellVertAlignment Write FCellVertAlignment default vaNone;
    property Columns: Word Read FColumns Write SetColumns default 3;
    property ColumnSizes: TStrings Read FColumnSizes Write SetColumnSizes;
    property Rows: Word Read FRows Write SetRows default 3;
    property RowSizes: TStrings Read FRowSizes Write SetRowSizes;
    property DrawFirstRowLine: Boolean Read FDrawFirstRowLine Write SetDrawFirstRowLine default True;
(*    property FirstRowHeight: Word Read FFirstRowHeight Write SetFirstRowHeight;
    property FirstColumnWidth: Word Read FFirstColumnWidth Write FFirstColumnWidth; *)
    property FirstRowColor: TColor Read FFirstRowColor Write SetFirstRowColor default clWhite;
    property FirstColumnColor: TColor Read FFirstColumnColor Write SetFirstColumnColor default clWhite;
    property FirstCellColor: TColor Read FFirstCellColor Write SetFirstCellColor default clWhite;
    property FirstColumnJoined: Boolean Read FFirstColumnJoined Write SetFirstColumnJoined default False;
    property FirstRowJoined: Boolean Read FFirstRowJoined Write SetFirstRowJoined default False;
    property Transparent;
    property Brush : TBrush read FBrush write SetBrush;
    property Pen : TPen read FPen write Setpen;
    property Height default 65;
    property Width default 65;
  end;

implementation
{$ifdef USE_QREDITOR}  //RAA
uses QRLablEd, QRExprEd;
{$endif}

const
  BreakChars : set of Char = [' ', #13, '-'];


{ BiDiMode support routines }

function OkToChangeFieldAlignment(AField: TField; Alignment: TAlignment): Boolean;
begin
  { dont change the alignment for these fields:
    ftSmallInt     ftInteger      ftWord         ftFloat        ftCurrency
    ftBCD          ftDate         ftTime         ftDateTime     ftAutoInc }
  if Assigned(AField) then with AField do
    Result := (DataType < ftSmallInt) or
              (DataType = ftBoolean) or
              ((DataType > ftDateTime) and (DataType <> ftAutoInc))
  else
    Result := Alignment <> taCenter;
end;

function QRDBUseRightToLeftAlignment(AControl: TControl; AField: TField): Boolean;
var
  AAlignment: TAlignment;
begin
  if Assigned(AField) then
    AAlignment := AField.Alignment
  else
    AAlignment := taLeftJustify;
  { Calling AControl.UseRightToLeftAlignment cause an endless recursion }
  Result := (AControl.BiDiMode = bdRightToLeft) and
    (OkToChangeFieldAlignment(AField, AAlignment));
end;


function GetTextHeight(Text: String; Font: TFont): Integer;
var
  C: TCanvas;
  DC: HDC;

begin
  C := TCanvas.Create;
  try
    DC := GetDC(0);
    try
      C.Handle := DC;
      C.Font.Assign(Font);
      Result := C.TextHeight(Text);
    finally
      ReleaseDC(0,DC);
    end;
  finally
    C.Free;
  end;
end;

function GetTextWidth(Text: String; Font: TFont): Integer;
var
  C: TCanvas;
  DC: HDC;

begin
  C := TCanvas.Create;
  try
    DC := GetDC(0);
    try
      C.Handle := DC;
      C.Font.Assign(Font);
      Result := C.TextWidth(Text);
    finally
      ReleaseDC(0,DC);
    end;
  finally
    C.Free;
  end;
end;

function EvalBoolExpression(E: String; ParentReport: TQuickRep): Boolean;
var
  aValue : TQREvResult;
  Evaluator : TQREvaluator;
begin
  EvalBoolExpression:=True;
  If E='' Then Exit;
  Evaluator := TQREvaluator.Create;
  Try
    Evaluator.DataSets := ParentReport.AllDataSets;
    Evaluator.Prepare(E);
    aValue := Evaluator.Value;
    If aValue.Kind=resBool Then
      EvalBoolExpression:=aValue.booResult
    Else
      EvalBoolExpression:=True;
    Evaluator.Unprepare;
  Finally
    Evaluator.Free;
  End;
end;

procedure CalculateRotatedPosition(Comp: TQRPrintable;
                                   FontHandle: HFont;
                                   Alignment: TAlignment;
                                   var Rect: TRect;
                                   Zoom : Integer;
                                   Printing: Boolean;
//                                   PrintOfsX, PrintOfsY: Integer;
                                   PrintOfsX, PrintOfsY: Extended;
                                   Text: String);
var
  CHeight, CWidth, Width2, Height2, DX, DY : Integer;
  RelPos : Single;
//  Factor,
  TW, TH, Degree: Integer;
  Size : TSize;
  DeviceContext : HDC;
  H : HFont;
begin
  DeviceContext:=GetDC(0);
  H:=SelectObject(DeviceContext, FontHandle);
  GetTextExtentPoint32(DeviceContext, @Text[1], Length(Text), Size);
  SelectObject(DeviceContext, H);
  ReleaseDC(0, DeviceContext);
  TW:=Size.cx; TH:=Size.cy;
  Degree := 0;
  if Comp is TQRCustomLabel then Degree := TQRCustomLabel(Comp).Degree;
//  Degree:=GetIntegerProperty(Comp,'Degree');

  if (Degree <= 900) then DX:=0 else
  if (Degree <= 1800) then DX:=Trunc(TW * -cos(Degree * Pi/1800)) else
  if (Degree <= 2700) then DX:=Trunc(TW * -cos(Degree * Pi/1800) + TH * sin((Degree - 1800) * Pi/1800))
  else DX:=Trunc(TH * sin((3600 - Degree) * Pi/1800));
  if (Degree <= 900) then DY:=Trunc(TW * Sin(Degree * Pi/1800)) else
  if (Degree <= 1800) then DY:=Trunc(TW * sin(Degree * Pi/1800) + TH * cos((1800 - Degree) * Pi/1800)) Else
  if (Degree <= 2700) then DY:=Trunc(TH * sin((2700 - Degree) * Pi/1800))
  else DY:=0;

  If Printing then begin
    Rect.Left:=Rect.Left + DX;
    Rect.Top:=Rect.Top + DY;

    Width2:=Round( Abs(Int(TW * cos(Degree * Pi/1800))) +
                        Abs(Trunc(TH * sin(Degree * Pi/1800))));
    Height2:=Round( Abs(Int(TW * sin(Degree * Pi/1800))) +
                         Abs(Trunc(TH * cos(Degree * Pi/1800))));

    CWidth:=Comp.Width;
    CHeight:=Comp.Height;
  end
  else begin
{
    Rect.Left:=Rect.Left + Round(DX * Zoom / 100);
    Rect.Top:=Rect.Top + Round(DY * Zoom / 100);
    CWidth:=Comp.Width;
    CHeight:=Comp.Height;
}
    Rect.Left:=Round((Rect.Left + DX) * Zoom / 100);
    Rect.Top:=Round((Rect.Top + DY) * Zoom / 100);
    CWidth:=round(Comp.Width*Zoom/100);
    CHeight:=round(Comp.Height*Zoom/100);
    Width2:=Round( (Abs(Trunc(TW * cos(Degree * Pi/1800))) +
                        Abs(Trunc(TH * sin(Degree * Pi/1800))))*Zoom/100);
    Height2:=Round( (Abs(Trunc(TW * sin(Degree * Pi/1800))) +
                         Abs(Trunc(TH * cos(Degree * Pi/1800))))*Zoom/100);
  end;
  //if GetBooleanProperty(Comp,'Autosize') then
  if (Comp is TQRCustomLabel) and TQRCustomLabel(Comp).Autosize then begin
{
    if Printing
    then Factor := round(TQRCustomLabel(Comp).Indent.Top * Screen.PixelsPerInch / 72)
    else Factor := round(TQRCustomLabel(Comp).Indent.Top * Screen.PixelsPerInch * Zoom / 7200);
    if (Factor < 1) and (TQRCustomLabel(Comp).Indent.Top > 0) then Factor := 1;
    Inc(Height2,Factor);
    if Printing
    then Factor := round(TQRCustomLabel(Comp).Indent.Bottom * Screen.PixelsPerInch / 72)
    else Factor := round(TQRCustomLabel(Comp).Indent.Bottom * Screen.PixelsPerInch * Zoom / 7200);
    if (Factor < 1) and (TQRCustomLabel(Comp).Indent.Bottom > 0) then Factor := 1;
    Inc(Height2,Factor);

    if Printing
    then Factor := round(TQRCustomLabel(Comp).Indent.Left * Screen.PixelsPerInch / 72)
    else Factor := round(TQRCustomLabel(Comp).Indent.Left * Screen.PixelsPerInch * Zoom / 7200);
    if (Factor < 1) and (TQRCustomLabel(Comp).Indent.Left > 0) then Factor := 1;
    Inc(Width2,Factor);
    if Printing
    then Factor := round(TQRCustomLabel(Comp).Indent.Right * Screen.PixelsPerInch / 72)
    else Factor := round(TQRCustomLabel(Comp).Indent.Right * Screen.PixelsPerInch * Zoom / 7200);
    if (Factor < 1) and (TQRCustomLabel(Comp).Indent.Right > 0) then Factor := 1;
    Inc(Width2,Factor);
}
    Comp.Width :=Width2;
    Comp.Height:=Height2;
  end
  else case Alignment of
    taLeftJustify: begin
      if (Degree <= 1800) then Rect.Top:= Rect.Top + CHeight - Height2;
      if (Degree > 900) and (Degree <= 2700) then Rect.Left:= Rect.Left + CWidth - Width2;
    end;
    taCenter: begin
      if Degree = 2700
      then Rect.Left:= Rect.Left + CWidth - Width2
      else if not (Degree = 900)
        then Rect.Left:= Rect.Left + ((CWidth-Width2) div 2);
      if Degree = 1800
      then Rect.Top:= Rect.Top + CHeight - Height2
      else Rect.Top:= Rect.Top + ((CHeight-Height2) div 2);
    end;
    taRightJustify: begin
      if (Degree < 900) or (Degree >= 2700) then Rect.Left:= Rect.Left + CWidth - Width2;
      if (Degree >= 1800) then Rect.Top:= Rect.Top + CHeight - Height2;
    end;
  end;

  if Printing and (Comp.ParentReport.QRPrinter.Destination = qrdPrinter) then
    with Comp.ParentReport.QRPrinter, Rect do  begin
      RelPos:=Left/Right;
      Left:=XPos(PrintOfsX + Comp.Size.Left);
      Right:=XPos(PrintOfsX + Comp.Size.Left + Comp.Size.Width);
      Left:=Round((Right - Left) * RelPos) + Left;

      RelPos:=Top/Bottom;
      Top:=YPos(PrintOfsY + Comp.Size.Top);
      Bottom:=YPos(PrintOfsY + Comp.Size.Top + Comp.Size.Height);
      Top:=Round((Bottom - Top) * RelPos) + Top;
    end;
end;


{ TQRCustomLabel }

constructor TQRCustomLabel.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  FAutoSize := true;
  FAutoStretch := false;
  FWordWrap := true;
  FLines := TStringList.Create;
  FFormattedLines := TStringList.Create;
  DoneFormat := false;
  Caption := '';
  Transparent := false;
  UpdatingBounds := false;
  FFontSize := 0;
  FCaptionBased := true;
  FDegree := 0;
  FIndent := TQRIndentSpace.Create(Self);
end;

destructor TQRCustomLabel.Destroy;
begin
  FLines.Free;
  if Assigned(FIndent) then FIndent.Free;
  FFormattedLines.Free;
  inherited Destroy;
end;

procedure TQRCustomLabel.SetIndent(Value: TQRIndentSpace);
begin
  if Assigned(FIndent) then FIndent.Free;
  FIndent := Value;
  FormatLines;
  Invalidate;
end;
function TQRCustomLabel.GetControlsAlignment: TAlignment;
begin
  Result := Alignment;
end;

function TQRCustomLabel.GetCaption : TCaption;
begin
  result := FCaption;
end;

function TQRCustomLabel.GetCaptionBased : boolean;
begin
  Result := FCaptionBased;
end;

procedure TQRCustomLabel.DefineProperties(Filer: TFiler);
begin
  Filer.DefineProperty('FontSize', ReadFontSize, WriteFontSize, true); // Do not translate
  inherited DefineProperties(Filer);
end;

procedure TQRCustomLabel.ReadFontSize(Reader : TReader);
begin
  FFontSize := Reader.ReadInteger;
end;

procedure TQRCustomLabel.WriteFontSize(Writer : TWriter);
begin
  Writer.WriteInteger(Font.Size);
end;

procedure TQRCustomLabel.Loaded;
begin
  inherited Loaded;
  if FFontSize > 0 then
    Font.Size := FFontSize;
end;

procedure TQRCustomLabel.CMFontChanged(var Message: TMessage);
begin
  inherited;
  DoneFormat := false;
 formatlines;
end;

procedure TQRCustomLabel.Prepare;
begin
  inherited Prepare;
  Caption := copy(Caption, 1, length(Caption));
end;

procedure TQRCustomLabel.Unprepare;
begin
  inherited Unprepare;
end;

procedure TQRCustomLabel.FormatLines;
var
  I, J : integer;
  NewLine : string;
  LineFinished : boolean;
  HasParent : boolean;
  MaxLineWidth : integer;
  AAlignment: TAlignment;
  Factor : Extended;
  a_Indent,
  a_Width: Integer;

  function aLineWidth(Line : string) : integer;
  begin
//RSV
    if HasParent then
//      result := Muldiv(Longint(ParentReport.TextWidth(Font, Line)),Zoom,100)+a_Width
      result := Muldiv(GetTextWidth(Line,Font),Zoom,100)
    else
//      Result := Canvas.TextWidth(Line) + a_Width;
      Result := GetTextWidth(Line,Font);
  end;

  procedure FlushLine;
  begin
    FFormattedLines.Add(NewLine);
    NewLine := '';
  end;

  procedure AddWord(aWord : string);
  begin
    if aLineWidth(NewLine + aWord) > (Width - a_Width) then
    begin
      if NewLine = '' then
      begin
        while aLineWidth(NewLine + copy(aWord, 1, 1)) <= (Width - a_Width) do
        begin
          NewLine := NewLine + copy(aWord, 1, 1);
          Delete(aWord, 1, 1);
        end;
        aWord := '';
      end;
      FlushLine;
      if aLineWidth(aWord) > (Width - a_Width) then
      begin
        if NewLine = '' then
        begin
          if Width = 0 then
            aWord := ''
          else
            while aLineWidth(aWord) > (Width - a_Width) do
                Delete(aWord, Length(aWord), 1);
        end;
        NewLine := aWord;
        FlushLine;
        aWord := '';
      end;
      if not WordWrap then
      begin
        aWord := '';
        LineFinished := true;
      end;
    end;
    NewLine := NewLine + aWord;
  end;

  procedure AddLine(Line : string);
  var
    aPos : integer;
  begin
    while pos(#10, Line) > 0 do
      Delete(Line, Pos(#10, Line), 1);
    aPos := pos(#13, Line);
    if aPos > 0 then
    begin
      repeat
        AddLine(copy(Line, 1, aPos - 1));
        Delete(Line, 1 , aPos);
        aPos := pos(#13, Line);
      until aPos = 0;
      AddLine(Line);
    end else
    begin
      J := 0;
      NewLine := '';
      LineFinished := false;
      if AutoSize then
      begin
        NewLine := Line;
        FlushLine;
        LineFinished := True;
      end else
      begin
        while (J < Length(Line)) and (Length(Line) > 0) do
        begin
          repeat
            inc(J)
          until (Line[J] in BreakChars) or (J >= Length(Line));
          AddWord(copy(Line, 1, J));
          Delete(Line, 1, J);
          J := 0;
        end;
        if not LineFinished then
          FlushLine;
      end;
    end;
  end;

  procedure FormatFromCaption;
  begin
    AddLine(FPrintCaption);
    if not UpdatingBounds and HasParent then
    begin
      UpdatingBounds := true;
      if Height < (longint(ParentReport.TextHeight(Font, 'W') * Zoom div 100) + 1) then
         Height := (longint(ParentReport.TextHeight(Font, 'W')) * Zoom div 100) + 1;
//      if Height < (muldiv(GetTextHeight('Yy',Font),Zoom,100) + 1) then
//         Height := muldiv(GetTextHeight('Yy',Font),Zoom,100) + 1;
      UpdatingBounds := false;
    end
  end;

  procedure FormatFromStringList;
  var
    J : integer;
  begin
    if (FLines.Count <> 0) then
    begin
      if AutoSize then
        FFormattedLines.Assign(FLines)
      else
        for J := 0 to FLines.Count - 1 do
          AddLine(FLines[J]);
    end else
      if csDesigning in ComponentState then
      begin
        FCaption := Name;
        FormatFromCaption;
        FCaption := '';
      end;
  end;

//>RSV
  function CalcHeight: Integer;
  begin
{
    if QRPrinter = nil then begin
      Factor := muldiv(Screen.PixelsPerInch,Zoom,7200);
      a_Width := round(Frame.Width*Factor);
      if a_Width < 1 then a_Width := 1;
      a_Indent := round((FIndent.Top+FIndent.Bottom)*Factor);
      if (a_Indent < 1) and ((FIndent.Top+FIndent.Bottom) > 0) then a_Indent := 1;
      Result := muldiv(GetTextHeight('Wy',Font),Zoom,100)+
        (byte(Frame.DrawTop)+byte(Frame.DrawBottom))*a_Width+(FIndent.Top+FIndent.Bottom);
    end
    else begin
      a_Width := round(QRPrinter.XFactor * Frame.Width * 254 / 72);
      if a_Width < 1 then a_Width := 1;
      a_Indent := round(QRPrinter.XFactor * (FIndent.Top+FIndent.Bottom) * 254 / 72);
      if (a_Indent < 1) and ((FIndent.Top+FIndent.Bottom) > 0) then a_Indent := 1;
      Result := GetTextHeight('Wy',Font)+
        (byte(Frame.DrawTop)+byte(Frame.DrawBottom))*a_Width+(FIndent.Top+FIndent.Bottom);
    end;
}

//    if QRPrinter = nil then
      Factor := Screen.PixelsPerInch /( 72 * 2); //TSK
//    else
//      Factor := {QRPrinter.YSize}(72/254);

      a_Width := round(Frame.Width*Factor);
      if (a_Width < 1) and (Frame.Width>0) then a_Width := 1;
      a_Indent := round(FIndent.Top*Factor);
      if (a_Indent < 1) and (FIndent.Top > 0) then a_Indent := 1;
      if (round(FIndent.Bottom*Factor) < 1) and (FIndent.Bottom > 0)
      then a_Indent := a_Indent + 1
      else a_Indent := a_Indent + round(FIndent.Bottom*Factor);

      Result := GetTextHeight('Wy',Font)+
        (byte(Frame.DrawTop)+byte(Frame.DrawBottom))*a_Width+
        a_Indent;
{
    end
    else begin
      a_Width := round(QRPrinter.YFactor * Frame.Width * 254 / 72);
      if a_Width < 1 then a_Width := 1;
      a_Indent := round(QRPrinter.YFactor * (FIndent.Top+FIndent.Bottom) * 254 / 72);
      if (a_Indent < 1) and ((FIndent.Top+FIndent.Bottom) > 0) then a_Indent := 1;
      Result := GetTextHeight('Wy',Font)+
        (byte(Frame.DrawTop)+byte(Frame.DrawBottom))*a_Width+(FIndent.Top+FIndent.Bottom);
    end;
}
{
    a_Width := muldiv(Frame.Width, Screen.PixelsPerInch, 72);
    Result := GetTextHeight('Wy',Font)+
      (byte(Frame.DrawTop)+byte(Frame.DrawBottom))*a_Width+
      muldiv((FIndent.Top+FIndent.Bottom),Screen.PixelsPerInch,72);
}
  end;
//<RSV

begin
  if csLoading in ComponentState then Exit;
  if Parent <> nil then
  begin
//>RSV
{RSV - PREV}
//    a_Width := muldiv(Frame.Width,254,Screen.PixelsPerInch);
//    a_Width := (byte(Frame.DrawLeft)+byte(Frame.DrawRight))*a_Width+
//      muldiv(FIndent.Left+FIndent.Right,254,Screen.PixelsPerInch);
//    a_Width := (254 div 72) + round(Frame.Width*Screen.PixelsPerInch/72);
{
    if QRPrinter = nil then
      a_Width := GetTextWidth('W',Font)
    else
      a_Width := Canvas.TextWidth('W');
}

//    if QRPrinter = nil then
      Factor := Screen.PixelsPerInch/72 ;
//    else
//    if QRPrinter <> nil then
//      Factor := QRPrinter.XSize(Factor);
    a_Width := 0;
{
//    if AutoSize then begin
      a_Width := round(Frame.Width*Factor) ;
      if (a_Width < 1) and (Frame.Width > 0) then a_Width := 1;
      a_Width := (byte(Frame.DrawLeft)+byte(Frame.DrawRight))*a_Width;
//    end;
}
    a_Indent := round(FIndent.Left*Factor);
    if (a_Indent < 1) and (FIndent.Left > 0) then a_Indent := 1;
    if (round(FIndent.Right*Factor) < 1) and (FIndent.Right > 0)
    then a_Indent := a_Indent + 1
    else a_Indent := a_Indent + round(FIndent.Right*Factor);
    a_Width := a_Width+a_Indent;
{
    Factor := muldiv(Screen.PixelsPerInch,254,72);
    if QRPrinter = nil then begin
      Factor := muldiv(Screen.PixelsPerInch,Zoom,7200);
      a_Width := Frame.Width*Factor;
      if a_Width < 1 then a_Width := 1;
      a_Indent := (FIndent.Left+FIndent.Right)*Factor;
      if (a_Indent < 1) and ((FIndent.Left+FIndent.Right) > 0) then a_Indent := 1;

    end
    else begin
      a_Width := round(QRPrinter.XFactor * Frame.Width * 254 / 72);
      if a_Width < 1 then a_Width := 1;
      a_Indent := round(QRPrinter.XFactor * (FIndent.Left+FIndent.Right) * 254 / 72);
      if (a_Indent < 1) and ((FIndent.Left+FIndent.Right) > 0) then a_Indent := 1;
    end;
}
//    a_Width := (byte(Frame.DrawLeft)+byte(Frame.DrawRight))*a_Width+(FIndent.Left+FIndent.Right);
//    a_Width := a_Width+(FIndent.Left+FIndent.Right);

//<RSV
    if assigned(FFormattedLines) then
      FFormattedLines.Clear
    else
      FFormattedLines := TStringList.Create;
    HasParent := ParentReport <> nil;
    LineFinished := false;
    if CaptionBased then
      FormatFromCaption
    else
      FormatFromStringList;
    if AutoSize and (not UpdatingBounds) and HasParent then
    begin
      MaxLineWidth := 0;
      for I := 0 to FFormattedLines.Count - 1 do
        if aLineWidth(FFormattedLines[I]) + a_Width > MaxLineWidth then
          MaxLineWidth := aLineWidth(FFormattedLines[I]) + a_Width;
{RSV
      if Frame.DrawLeft then
        MaxLineWidth := MaxLineWidth + Frame.Width;
      if Frame.DrawRight then
        MaxLineWidth := MaxLineWidth + Frame.Width;
}
      MaxLineWidth := MaxLineWidth;
      UpdatingBounds := true;
      AAlignment := Alignment;
      if UseRightToLeftAlignment then
        ChangeBiDiModeAlignment(AAlignment);
      case AAlignment of
        taCenter : Left := Left + ((Width - MaxLineWidth) div 2);
        taRightJustify : Left := Left + Width - MaxLineWidth;
      end;
      Width := MaxLineWidth;

      if (FFormattedLines.Count = 0) and (csDesigning in ComponentState) then
        Height := CalcHeight;

      if (Height < CalcHeight) then
        Height := CalcHeight;
{RSV
      if (FFormattedLines.Count = 0) and (csDesigning in ComponentState) then
        Height := (longint(ParentReport.TextHeight(Font, 'Wy')) * Zoom div 100) + 1;

      if (Height < (longint(ParentReport.TextHeight(Font, 'Wy') * Zoom div 100) + 1)) then
        Height := (longint(ParentReport.TextHeight(Font, 'Wy')) * Zoom div 100) + 1;
}
      UpdatingBounds := false;
    end;
  end;
  DoneFormat := true;
end;

procedure TQRCustomLabel.SetLines(Value : TStrings);
begin
  FLines.Assign(Value);
  DoneFormat := false; {xxx}
  Invalidate;
end;

type
  TQRFixFrame = class(TQRFrame);

procedure TQRCustomLabel.PaintToCanvas(aCanvas : TCanvas; aRect : TRect; CanExpand : boolean; LineHeight: integer);
var
  I : integer;
  Factor,
  EndX,
  EndY,
  StartX : integer;
  StartY : integer;
  Cap : string;
  VPos : integer;
  Flags : integer;
  AAlignment: TAlignment;

begin
  FormatLines;
  Flags := 0;
{  if AutoSize then Flags := 0 else Flags := ETO_CLIPPED;}
  if not Transparent then
  begin
    aCanvas.Brush.Color := Color;
    aCanvas.Brush.Style := bsSolid;
    aCanvas.Fillrect(aRect);
  end;
  StartY := aRect.Top;
  StartX := aRect.Left;
  EndY   := aRect.Bottom;
  EndX   := aRect.Right;
//  Factor := muldiv(Screen.PixelsPerInch, Zoom, 7200);
  if Frame.AnyFrame then
  begin
{RSV
    if Frame.DrawTop and (Frame.Width > 0 ) then
      StartY := StartY + round(Frame.Width / 72 * Screen.PixelsPerInch * Zoom / 100);
    if Frame.DrawLeft then
      StartX := StartX + round(Frame.Width / 72 * Screen.PixelsPerInch * Zoom / 100)
}
{RSV - PREV
    StartY := StartY + byte(Frame.DrawTop) * round(Frame.Width / 72 * Screen.PixelsPerInch * Zoom / 100);
    StartX := StartX + byte(Frame.DrawLeft) * round(Frame.Width / 72 * Screen.PixelsPerInch * Zoom / 100);
    EndY := EndY - byte(Frame.DrawBottom) * round(Frame.Width / 72 * Screen.PixelsPerInch * Zoom / 100);
}
//    StartY := StartY + byte(Frame.DrawTop) * round(Frame.Width * Screen.PixelsPerInch * Zoom / 7200);
//    EndY := EndY - byte(Frame.DrawBottom) * round(Frame.Width * Screen.PixelsPerInch * Zoom / 7200);
//    StartX := StartX + byte(Frame.DrawLeft) * round(Frame.Width * Factor);
//    EndX := EndX - byte(Frame.DrawRight) * round(Frame.Width * Factor);
  end;
{RSV - PREV
  StartY := StartY + round(Indent / 72 * Screen.PixelsPerInch * Zoom / 100);
  StartX := StartX + round(Indent / 72 * Screen.PixelsPerInch * Zoom / 100);
  EndY := EndY - round(Indent / 72 * Screen.PixelsPerInch * Zoom / 100);
}
{
  StartY := StartY + Round(FIndent.Top * Factor);
  StartX := StartX + Round(FIndent.Left * Factor);
  EndY := EndY - Round(FIndent.Bottom * Factor);
  EndX := EndX - Round(FIndent.Right * Factor);
}
{
  StartY := StartY + Round(FIndent.Top * Screen.PixelsPerInch * Zoom/7200);
  EndY := EndY - Round(FIndent.Bottom * Screen.PixelsPerInch * Zoom/7200);

  StartX := StartX + Round(FIndent.Left * Screen.PixelsPerInch * Zoom/7200);
  EndX := EndX - Round(FIndent.Right * Screen.PixelsPerInch * Zoom/7200);
}

  Factor := round(FIndent.Top * Screen.PixelsPerInch * Zoom/7200);
  if (Factor < 1) and (FIndent.Top > 0) then Factor := 1;
  StartY := StartY + Factor;

  Factor := round(FIndent.Bottom * Screen.PixelsPerInch * Zoom/7200);
  if (Factor < 1) and (FIndent.Bottom > 0) then Factor := 1;
  EndY := EndY - Factor;

  Factor := round(FIndent.Left * Screen.PixelsPerInch * Zoom/7200);
  if (Factor < 1) and (FIndent.Left > 0) then Factor := 1;
  StartX := StartX + Factor;

  Factor := round(FIndent.Right * Screen.PixelsPerInch * Zoom/7200);
  if (Factor < 1) and (FIndent.Right > 0) then Factor := 1;
  EndX := EndX - Factor;

//  if AutoSize then begin
//    aRect.Right := aRect.Right - aRect.Left;
//    aRect.Bottom := aRect.Bottom - aRect.Top;
//  end
//  else begin
    aRect.Right := EndX - StartX;
    aRect.Bottom := EndY - StartY;
//  end;

  aRect.Left := 0;
  aRect.Top := 0;
  SetBkColor(aCanvas.Handle, not ColorToRGB(aCanvas.Brush.Color));
  SetBkMode(aCanvas.Handle, Windows.Transparent);
  begin
    AAlignment := Alignment;
    if UseRightToLeftAlignment then
      ChangeBiDiModeAlignment(AAlignment);
    case AAlignment of
      TaLeftJustify : SetTextAlign(aCanvas.Handle, TA_Left + TA_Top + TA_NoUpdateCP);
      TaRightJustify: begin
          SetTextAlign(aCanvas.Handle, TA_Right + TA_Top + TA_NoUpdateCP);
          StartX := StartX + aRect.Right;
{
          if AutoSize
          then begin
            Factor := round(FIndent.Right * Screen.PixelsPerInch * Zoom/7200);
            if (Factor < 1) and (FIndent.Right > 0) then Factor := 1;
            StartX := StartX - Factor;
          end;
}
        end;
      TaCenter : begin
          SetTextAlign(aCanvas.Handle, TA_Center + TA_Top + TA_NoUpdateCP);
          StartX := StartX + aRect.Right div 2;
        end;
    end;
  end;


//  SetTextJustification(aCanvas.Handle,72,Screen.PixelsPerInch);
try
  for I := 0 to FFormattedLines.Count - 1 do
  begin
    VPos := StartY + I * LineHeight;
    begin
      Cap := FFormattedLines[I];
//      if (Length(Cap) > 0) and ((VPos+LineHeight) <= EndY) or AutoStretch then
        ExtTextOut(aCanvas.Handle, StartX, VPos, Flags, @aRect, @Cap[1], length(Cap), nil);
    end;
  end;
finally
//  SetTextJustification(aCanvas.Handle,0,0);
end;
//!!!
end;

{procedure TQRFixFrame.PaintFit(ACanvas : TCanvas; ARect : TRect; XFact, YFact : extended);
var
  FWX, FWY : integer;
begin
  FWX := round(XFact / 72 * 254 * Width);
  if ((FWX < 1) and (Width >= 1)) or (Width = -1) then
    FWX := 1;
  FWY := round(YFact / 72 * 254 * Width);
  if ((FWY < 1) and (Width >= 1)) or (Width = -1) then
    FWY := 1;
  ACanvas.Brush.Style := bsSolid;
  ACanvas.Brush.Color := Color;
  SetPen(ACanvas.Pen);
  with aCanvas do
  begin
    if DrawTop then
      FillRect(Rect(ARect.Left, ARect.Top, ARect.Right, ARect.Top + FWY));
    if DrawBottom then
      FillRect(Rect(ARect.Left, ARect.Bottom, ARect.Right, ARect.Bottom - FWY));
    if DrawLeft then
      FillRect(Rect(ARect.Left, ARect.Top, ARect.Left + FWX, ARect.Bottom));
    if DrawRight then
      FillRect(Rect(ARect.Right - FWX, ARect.Top, ARect.Right, ARect.Bottom));
  end;
  ACanvas.Brush.Style := bsClear;
end;}

(*
procedure TQRCustomLabel.PrintToCanvas(aCanvas : TCanvas;
                                       aLeft, aTop, aWidth, aHeight,
                                       LineHeight : extended;
                                       CanExpand : boolean);
var
  aRect : TRect;
  ControlBottom : extended;
  X, Y : extended;
  SavedCaption : string;
  NewCaption : string;
  HasSaved : boolean;
  HasExpanded : boolean;
  Flags : integer;
  TAFlags : integer;
  AAlignment : TAlignment;
  AFExpanded : extended;
  OrgWidth : extended;

  function CanPrint : boolean;
  var
    PrevTop : extended;
  begin
    Result := true;
    //RSV            VVV
    if Y + LineHeight
      +byte( ((FCurrentLine=0) or (FCurrentLine=(FFormattedLines.count-1))) and (FFormattedLines.count<>0) )*
      round(QRPrinter.YFactor*Frame.Width*254/72)  > ControlBottom then
    begin
      if CanExpand and TQRCustomBand(Parent).CanExpand(LineHeight
      //RSV
      +byte( ((FCurrentLine=0) or (FCurrentLine=(FFormattedLines.count-1))) and (FFormattedLines.count<>0) )*
       round(QRPrinter.YFactor*Frame.Width*254/72)) then
      begin
        PrevTop := AFExpanded;
        TQRCustomBand(Parent).ExpandBand(LineHeight
      //RSV
      +byte( ((FCurrentLine=0) or (FCurrentLine=(FFormattedLines.count-1))) and (FFormattedLines.count<>0) )*
       round(QRPrinter.YFactor*Frame.Width*254/72)
        , AFExpanded, HasExpanded);
        ControlBottom := aTop + aHeight + 1 + AFExpanded;
        if ParentReport.FinalPass and not Transparent then
          with aCanvas do
          begin
            Pen.Width := 0;
            Brush.Color := Color;
            Brush.Style := bsSolid;
            FillRect(rect(QRPrinter.XPos(aLeft),
                          QRPrinter.YPos(aTop + AHeight + PrevTop),
                          QRPrinter.XPos(aLeft + aWidth),
                          QRPrinter.YPos(aTop + aHeight + AFExpanded)));
          end;
      end else
        Result := false;
    end;
  end;

  procedure PrintLine(LineNumber : integer);
  begin
    if ParentReport.FinalPass and (Length(FFormattedLines[LineNumber]) > 0) then
    begin
      ExtTextOut(aCanvas.Handle, QRPrinter.XPos(X),
       QRPrinter.YPos(Y),
        Flags, @aRect, @FFormattedLines[LineNumber][1], length(FFormattedLines[LineNumber]), nil);
    end;
    if ParentReport.Exporting then
      ParentReport.ExportFilter.TextOut(X,Y,
       Font, Color, AAlignment, FFormattedLines[LineNumber]);
    Y := Y + LineHeight;
  end;

begin
  Flags := 0;
  AFExpanded := 0;
  OrgWidth := aWidth;
{  if AutoSize then Flags := 0; else Flags := ETO_CLIPPED;}
{  if Transparent then Flags := Flags + ETO_OPAQUE;}
  HasSaved := false;
  if (FPrintCaption <> '') and assigned(FOnPrint) then
  begin
    SavedCaption := FPrintCaption;
    NewCaption := FprintCaption;
    FOnPrint(Self, NewCaption);

    if Font <> aCanvas.Font then
    begin
      aCanvas.Font := Font;
      aHeight := Size.Height;
      LineHeight := aCanvas.TextHeight('Wy') / QRPrinter.YFactor;
    end;

    if NewCaption <> FPrintCaption then
    begin
      FPrintCaption := NewCaption;
      FormatLines;
      HasSaved := true;
    end;
  end;
  FormatLines;
//  aWidth := Width / QRPrinter.XFactor;
  {if not Autosize then }aWidth := Size.Width;
  if ParentReport.FinalPass and not Transparent then
    with aCanvas do
    begin
      Pen.Width := 0;
      Brush.Color := Color;
      Brush.Style := bsSolid;
      FillRect(rect(QRPrinter.XPos(aLeft),
                    QRPrinter.YPos(aTop),
                    QRPrinter.XPos(aLeft + aWidth),
                    QRPrinter.YPos(aTop + aHeight)));
    end;
  if ParentReport.FinalPass then
    if not AutoSize then;
{
  if Frame.AnyFrame then
  begin
    if Frame.DrawTop then
      aTop := aTop ;//+ muldiv(Frame.Width,254,72);//RSV round(Frame.Width / 72 * 254 );
    if Frame.DrawLeft then
      aLeft := aLeft ;//+ muldiv(Frame.Width,254,72);//RSV round(Frame.Width / 72 * 254 )
  end;
}
  { Get our rectangle for the next line }
  aRect := Rect(0, 0, QRPrinter.XSize(aWidth), QRPrinter.YSize(LineHeight));

  AAlignment := Alignment;
  if UseRightToLeftAlignment then
    ChangeBiDiModeAlignment(AAlignment);

  { Calculate some stuff... }
  ControlBottom := aTop + aHeight + 1;
  Y := aTop;
  if not AutoSize then
    X := aLeft
  else
  begin
    case Alignment of
      TaLeftJustify : X := aLeft;
      TaRightJustify: X := aLeft - (aWidth - OrgWidth);
      TaCenter : X := aLeft - ((aWidth - OrgWidth) / 2);
    end;
  end;

  SetBkMode(aCanvas.Handle, Windows.Transparent);
  { Set the attributes and update X for alignment }
  TAFlags := TA_Top + TA_NoUpdateCP;
  AAlignment := Alignment;
  if UseRightToLeftAlignment then
    ChangeBiDiModeAlignment(AAlignment);
  if UseRightToLeftReading then
  begin
    Flags := Flags or ETO_RTLREADING;
    TAFlags := TAFlags + TA_RTLREADING;
  end;
  case AAlignment of
    TaLeftJustify : SetTextAlign(aCanvas.Handle, TA_Left + TAFlags);
    TaRightJustify: begin
      SetTextAlign(aCanvas.Handle, TA_Right + TAFlags);
      X := X + aWidth;
    end;
    TaCenter : begin
      SetTextAlign(aCanvas.Handle, TA_Center + TAFlags);
      X := X + aWidth / 2;
    end;
  end;

  HasExpanded := false;
  if PrintFinished then
    FCurrentLine := 0;

//  if ParentReport.FinalPass and (Length(FFormattedLines[0]) > 0) then
  begin
    X := X + round(Frame.Width / 72 * 254 );
//    if FFormattedLines.Count > 1 then
//      Y := Y + round(Frame.Width / 72 * 254 );
//    X := X +1+ QRPrinter.XFactor*(254*Frame.Width/72);
//    Y := Y +1+ QRPrinter.YFactor*(254*Frame.Width/72);
  end;

  while (FCurrentLine <= FFormattedLines.Count - 1) and CanPrint do
  begin
    PrintLine(FCurrentLine);
    inc(FCurrentLine);
  end;

  if (FCurrentLine <= FFormattedLines.Count - 1) and AutoStretch then
    PrintFinished := false
  else
    PrintFinished := true;
  SelectClipRgn(QRPrinter.Canvas.Handle, 0);
// end new code

  if HasSaved then
    FPrintCaption := SavedCaption;

  if ParentReport.FinalPass and Frame.AnyFrame then
    TQRFixFrame(Frame).PaintIt(aCanvas,
      rect(QRPrinter.XPos(ALeft),
        QRPrinter.YPos(ATop),
        QRPrinter.XPos(ALeft + aWidth),
        QRPrinter.YPos(Atop + Size.height+ AFExpanded)),
      QRPrinter.XFactor,
      QRPrinter.YFactor);
end;
*)


type
  TCCQuickRep = class(TCustomQuickRep);

procedure TQRCustomLabel.PrintToCanvas(aCanvas : TCanvas;
                                       aLeft, aTop, aWidth, aHeight,
                                       LineHeight : extended;
                                       CanExpand : boolean);
var
  aRect : TRect;
  SavedCaption,
  NewCaption : string;
  FirstCall,
//  HasExpanded,
  HasSaved : boolean;
  Flags : integer;
  TAFlags : integer;
  AAlignment : TAlignment;
  X, Y, T,
//  IndentWidth, IndentHeight,
//  FrameWidth, FrameHeight,
  OrgWidth : extended;

  FH: Integer;

  PrevTop,
  Delta : extended;

  function GetControlBottom: extended;
  begin
//    with TQRCustomBand(Parent) do
    Result := 0;
    if Parent = ParentReport then
      Result := ParentReport.AvailableSpace;
    if Parent is TQRCustomBand then with Parent as TQRCustomBand do
      Result := ParentReport.CurrentY + Size.Length+Expanded;
//    Result := ParentReport.CurrentY + TQRCustomBand(Parent).Size.Length + TQRCustomBand(Parent).Expanded;
  end;

  procedure PrintFillRect(Factor: extended);
  var
    a_X: Extended;
  begin
    if ParentReport.FinalPass and not Transparent and (Factor > 0.0)
    then with aCanvas do begin
      PrevTop := Y;
      T := OrgWidth;
      if aWidth > OrgWidth then T := aWidth;
//RSV 0410
      a_X := aLeft;
      if AutoSize and (Alignment = TaRightJustify)
      then a_X := aLeft-(aWidth-OrgWidth)-(FIndent.LeftWidth+FIndent.RightWidth);

      Brush.Color := Color;
      Brush.Style := bsSolid;
      FillRect(rect(QRPrinter.XPos({aLeft}a_X),
                    QRPrinter.YPos(PrevTop),
                    QRPrinter.XPos(a_X{aLeft} + T),
                    QRPrinter.YPos(PrevTop+Factor)));

    end;
  end;

  procedure FirstcallPaint(Factor: extended);
  begin
    if FirstCall and Frame.DrawTop then begin
      Delta := Y;
      Y := aTop;
      PrintFillRect(Factor);
      Y := Delta;
    end;
  end;

  function CanPrint : boolean;
  begin
    Result := true;
    if not AutoStretch then
      Result := (Y + LineHeight <= aTop+aHeight) and ((FCurrentLine+1) <= FFormattedLines.Count)
    else begin
      Delta := Y + LineHeight + FIndent.BottomHeight {+ FrameHeight*byte(Frame.DrawBottom)}-GetControlBottom;
      if Delta > 0.0 then
      begin
        if CanExpand and TQRCustomBand(Parent).CanExpand(Delta) then
        begin
          Delta := Y+LineHeight-GetControlBottom;
          TQRCustomBand(Parent).Expanded := TQRCustomBand(Parent).Expanded + Delta;
        end else
          Result := false;
      end
    end
  end;

  procedure PrintLine(LineNumber : integer);
  var
    Mf: TMetafile;
    MC: TMetafileCanvas;
    DC: HDC;
    R: TRect;
    Exte: Extended;

//    OldCopyMode: TCopyMode;

  begin
    PrintFillRect(LineHeight);
{****
    if ParentReport.FinalPass and (Length(FFormattedLines[LineNumber]) > 0) then
    try
//      if QRPrinter.Destination = qrdMetafile
//      then SetTextJustification(aCanvas.Handle,72,Screen.PixelsPerInch);

      aCanvas.Font.Height := FH;
      aCanvas.Brush.Style := bsClear;
      ExtTextOut(aCanvas.Handle, QRPrinter.XPos(X), QRPrinter.YPos(Y),
        Flags, @aRect, @FFormattedLines[LineNumber][1], length(FFormattedLines[LineNumber]), nil);
    finally
//      SetTextJustification(aCanvas.Handle,0,0);
    end;
}
  if ParentReport.FinalPass and (Length(FFormattedLines[LineNumber]) > 0)
  then if QRPrinter.Destination = qrdPrinter then begin
      Mf := TMetafile.Create;
      try
        Mf.Height := round(LineHeight*Screen.PixelsPerInch/254);
        Mf.Width := round((aWidth+FIndent.LeftWidth+FIndent.RightWidth)*Screen.PixelsPerInch/254);
        DC := GetDC(0);
        MC := TMetafileCanvas.Create(Mf,DC);
        try
          MC.Font.Assign(Font);
          Mf.Transparent := True;
          R := MC.ClipRect;
          MC.Brush.Color := Color;
          if Transparent
          then begin
            MC.Brush.Style := bsClear;
            SetBkMode(MC.Handle, Windows.TRANSPARENT);
          end;
          SetTextAlign(MC.Handle,GetTextAlign(aCanvas.Handle));
{
          MC.TextRect(
            R,
            round(FIndent.LeftWidth*Screen.PixelsPerInch/254),
            0,
            FFormattedLines[LineNumber]);
}

          ExtTextOut(MC.Handle,
            round(FIndent.LeftWidth*Screen.PixelsPerInch/254),
            0,
            Flags {or ETO_OPAQUE},
            @R, @FFormattedLines[LineNumber][1], length(FFormattedLines[LineNumber]), nil);
        finally
          MC.Free;
          ReleaseDC(0,DC);
        end;
//        OldCopyMode := aCanvas.CopyMode;
//        aCanvas.CopyMode := cmNotSrcErase;
        Exte := FIndent.LeftWidth;
        case Alignment of
        TaCenter: Exte := Exte / 2;
        TaRightJustify: Exte := 0;
        end;
        aCanvas.StretchDraw(
                      rect(QRPrinter.XPos(X-Exte),
                      QRPrinter.YPos(Y),
                      QRPrinter.XPos(X{-FIndent.LeftWidth}+FIndent.RightWidth+Size.Width)+1,
                      QRPrinter.YPos(Y+LineHeight)+1)
                            ,Mf);
//        aCanvas.CopyMode := OldCopyMode;
      finally
        Mf.Free;
      end;
    end
    else begin
      aCanvas.Font.Height := FH;
      aCanvas.Brush.Style := bsClear;
      ExtTextOut(aCanvas.Handle, QRPrinter.XPos(X), QRPrinter.YPos(Y),
        Flags, @aRect, @FFormattedLines[LineNumber][1], length(FFormattedLines[LineNumber]), nil);
    end;
    if ParentReport.Exporting then
      ParentReport.ExportFilter.TextOut( X, Y, Font, Color, AAlignment, FFormattedLines[LineNumber]);
    Y := Y+LineHeight;
  end;

begin
  Flags := 0;
  ControlExpanded := 0;
  OrgWidth := aWidth;
{RSV - PREV
  FrameHeight := Frame.Width * 254 / 72;
  FrameWidth  := FrameHeight;
  IndentWidth := Indent * 254 / 72;
  IndentHeight:= IndentWidth;
}

//  FrameHeight := Frame.Width/QRPrinter.YFactor;
//  FrameWidth  := Frame.Width/QRPrinter.XFactor;

//  IndentWidth := Indent/QRPrinter.XFactor;
//  IndentHeight:= Indent/QRPrinter.YFactor;
  FirstCall := True;

  QRPrinter.Canvas.Pen.Color := Frame.Color;

  HasSaved := false;
  if {(FPrintCaption <> '') and} assigned(FOnPrint) then begin
    SavedCaption := FPrintCaption;
    NewCaption := FprintCaption;
    FOnPrint(Self, NewCaption);

    if Font <> aCanvas.Font then begin
      aCanvas.Font := Font;
      aHeight := Size.Height;

//RSV      LineHeight := aCanvas.TextHeight('Wy') / QRPrinter.YFactor;
      LineHeight := GetTextHeight('Wy',Font) * 254 / Screen.PixelsPerInch
    end;

    if NewCaption <> FPrintCaption then begin
      FPrintCaption := NewCaption;
      FormatLines;
      HasSaved := true;
    end;
  end;

  FormatLines;

  FH := Font.Height;
  if QRPrinter.Destination = qrdPrinter
  then FH := (FH*GetDeviceCaps(QRPrinter.Canvas.Handle,LOGPixelsY) div (Screen.PixelsPerInch));

  aWidth := Size.Width;
  if not AutoSize then
  aWidth := aWidth-FIndent.LeftWidth-FIndent.RightWidth;

  { Get our rectangle for the next line }
  aRect := Rect(0, 0, QRPrinter.XSize(aWidth), QRPrinter.YSize(LineHeight));

  AAlignment := Alignment;
  if UseRightToLeftAlignment then ChangeBiDiModeAlignment(AAlignment);

  { Calculate some stuff... }
//  ControlBottom := aTop + aHeight {+ 1} + TQRCustomBand(Parent).Expanded};
//  ControlBottom := aTop + aHeight + dExpanded;
  Y := aTop;
  X := aLeft + FIndent.LeftWidth;

  if AutoSize then begin
    case Alignment of
      TaRightJustify: X := aLeft - aWidth + OrgWidth;
    end;
  end;

  {
  if AutoSize then begin
    case Alignment of
      TaLeftJustify : X := aLeft;
      TaRightJustify: X := aLeft - (aWidth - OrgWidth);
      TaCenter : X := aLeft - ((aWidth - OrgWidth) / 2);
    end;
  end;
}

//  if AutoSize and (Alignment = TaCenter) then X := X - (aWidth - OrgWidth) / 2;

  SetBkMode(aCanvas.Handle, Windows.Transparent);
  { Set the attributes and update X for alignment }
  TAFlags := TA_Top + TA_NoUpdateCP;
  AAlignment := Alignment;
  if UseRightToLeftAlignment then
    ChangeBiDiModeAlignment(AAlignment);
  if UseRightToLeftReading then
  begin
    Flags := Flags or ETO_RTLREADING;
    TAFlags := TAFlags + TA_RTLREADING;
  end;
  case AAlignment of
    TaLeftJustify : begin
        SetTextAlign(aCanvas.Handle, TA_Left + TAFlags);
//        X := X{+FrameWidth*byte(Frame.DrawLeft)+FIndent.LeftWidth};
      end;
    TaRightJustify: begin
      SetTextAlign(aCanvas.Handle, TA_Right + TAFlags);
      X := X + aWidth;
//RSV 0410
//      SetTextAlign(aCanvas.Handle, TA_Left + TAFlags);
//      X := X - (aWidth-OrgWidth) {- FrameWidth*byte(Frame.DrawRight)-FIndent.RightWidth};
//      X := aLeft + OrgWidth-FIndent.RightWidth{ - FrameWidth*byte(Frame.DrawRight)-FIndent.RightWidth};
    end;
    TaCenter : begin
      SetTextAlign(aCanvas.Handle, TA_Center + TAFlags);
//      X := X + aWidth / 2;
      X := aLeft + OrgWidth / 2;
    end;
  end;

  if PrintFinished then FCurrentLine := 0;

{
  if Frame.DrawTop then begin
    if AutoStretch then begin
      Delta := Y + FrameHeight  - GetControlBottom;
      if TQRCustomBand(Parent).CanExpand(Delta)
      then TQRCustomBand(Parent).Expanded := TQRCustomBand(Parent).Expanded + Delta;
    end;
    PrintFillRect(FrameHeight);
    Y := Y + FrameHeight;
  end;

}
  PrintFillRect(Size.Height);

  if FIndent.Top > 0 then begin
    if AutoStretch then begin
      Delta := Y + FIndent.TopHeight - GetControlBottom;
      if TQRCustomBand(Parent).CanExpand(Delta)
      then TQRCustomBand(Parent).Expanded := TQRCustomBand(Parent).Expanded + Delta;
    end;
    PrintFillRect(FIndent.TopHeight);
    Y := Y + FIndent.TopHeight;
  end;

//  if not AutoStretch then PrintFillRect(Size.Height);

//  OffsetRect(aRect, QRPrinter.XSize(aLeft),
//                    QRPrinter.XSize(aTop));
  while (FCurrentLine <= FFormattedLines.Count - 1) and CanPrint do
  begin
//    FirstCallPaint(Size.Height + byte(AutoStretch)*FIndent.TopHeight);
    FirstCallPaint(byte(AutoStretch)*FIndent.TopHeight);
    PrintLine(FCurrentLine);
    inc(FCurrentLine);
//    if Frame.DrawBottom and AutoStretch then PrintFillRect(FrameHeight);
//    if (Indent.Bottom>0) and AutoStretch then PrintFillRect(Indent.BottomHeight);
    FirstCall := False;
  end;

  if (FIndent.Bottom > 0){ and AutoStretch} then begin
    Delta := Y + FIndent.BottomHeight - GetControlBottom;
    if TQRCustomBand(Parent).CanExpand(Delta)
    then begin
      TQRCustomBand(Parent).Expanded := TQRCustomBand(Parent).Expanded + Delta;
//      PrintFillRect(FIndent.BottomHeight);
    end;
    PrintFillRect(FIndent.BottomHeight);
    Y := Y + FIndent.BottomHeight;
  end;
{
  if Frame.DrawBottom and AutoStretch then begin
    Delta := Y + FrameHeight - GetControlBottom;
    if TQRCustomBand(Parent).CanExpand(Delta)
    then begin
      TQRCustomBand(Parent).Expanded := TQRCustomBand(Parent).Expanded + Delta;
      PrintFillRect(FrameHeight);
      Y := Y + FrameHeight;
    end
  end;
}
  if (FCurrentLine <= FFormattedLines.Count - 1) and AutoStretch then
    PrintFinished := false
  else
    PrintFinished := true;

  SelectClipRgn(QRPrinter.Canvas.Handle, 0);
// end new code

  if HasSaved then FPrintCaption := SavedCaption;

  if not AutoStretch then begin
    ControlExpanded := 0;
    Y := aTop + Size.Height;
  end
  else begin
    if Y > GetControlBottom then Y := GetControlBottom;
    if FirstCall then FirstCallPaint(Y-aTop);
    ControlExpanded := Y - aTop - aHeight;
    if ControlExpanded < 0.0 then ControlExpanded := 0;
  end;

  Delta := OrgWidth;
  if aWidth > OrgWidth then Delta := aWidth;
{
  if ParentReport.FinalPass and Frame.AnyFrame// and (FCurrentLine>0)
  then
    TQRFixFrame(Frame).PaintIt(aCanvas,
      rect(QRPrinter.XPos(ALeft),
        QRPrinter.YPos(ATop),
        QRPrinter.XPos(ALeft + Delta),
        QRPrinter.YPos(Atop + Size.Height + ControlExpanded)),
//        QRPrinter.YPos(Y)),
      QRPrinter.XFactor,
      QRPrinter.YFactor);
}

end;

procedure TQRCustomLabel.Paint;
begin
  if FDegree = 0 then begin
    Canvas.Font.Assign(Font);
    if Canvas.Font.Size <> muldiv(Font.Size, Zoom, 100) then
      Canvas.Font.Size := muldiv(Font.Size, Zoom, 100);
    inherited Paint;
    PaintToCanvas(Canvas, rect(0, 0, Width, Height), false, round(Canvas.TextHeight('W')));
  end
  else DegreePaint;
  TQRFixFrame(Frame).PaintIt(Canvas, rect(0, 0, Width, Height), Screen.PixelsPerInch / 254 * Zoom/100, Screen.PixelsPerInch / 254 * Zoom/100);
  PaintCorners;
end;

procedure TQRCustomLabel.DegreePrint(OfsX, OfsY: Extended);
var
  R, ClipRect : TRect;
  LogFont : TLogFont;
  OldFont: HFont;
  PrintStr: PChar;
  SavedCaption,
  NewCaption: String;

begin
  R :=Rect(ParentReport.QRPrinter.XPos(OfsX + Size.Left),
           ParentReport.QRPrinter.YPos(OfsY + Size.Top),
           ParentReport.QRPrinter.XPos(OfsX + Size.Left + Size.Width),
           ParentReport.QRPrinter.YPos(OfsY + Size.Top + Size.Height));

  if assigned(FOnPrint) then
  begin
    if CaptionBased then begin
      SavedCaption := FPrintCaption;
      NewCaption := FprintCaption;
    end
    else begin
      SavedCaption := Lines.Text;
      NewCaption := Lines.Text;
    end;

    FOnPrint(Self, NewCaption);

    if Font <> ParentReport.QRPrinter.Canvas.Font then
      ParentReport.QRPrinter.Canvas.Font := Font;

    if NewCaption <> FPrintCaption then begin
      if CaptionBased
      then FPrintCaption := NewCaption
      else Lines.Text := NewCaption;
    end;
  end;

  if not Transparent then begin
    ParentReport.QRPrinter.Canvas.Brush.Style:=bsSolid;
    ParentReport.QRPrinter.Canvas.Brush.Color:=Color;
    ParentReport.QRPrinter.Canvas.FillRect(R);
  end;
  ParentReport.QRPrinter.Canvas.Brush.Style:=bsClear;
  with ParentReport.QRPrinter do begin
      FormatLines;
      if CaptionBased
      then PrintStr := PChar(FPrintCaption)
      else PrintStr := PChar(Lines.Text);
      Canvas.Font:=Font;
      Canvas.Font.Color:=Font.Color;
      ClipRect:=R;
      if (ParentReport.QRPrinter.Destination = qrdPrinter) then R:=ClientRect;

      CalculateRotatedPosition(Self, Font.Handle, Alignment, R,
                               Zoom, True, OfsX, OfsY, PrintStr);
      GetObject(Canvas.Font.Handle, SizeOf(LogFont), @LogFont);
      LogFont.lfEscapement :=Degree;
//      LogFont.lfOrientation:=Degree;
      OldFont:=SelectObject(Canvas.Handle, CreateFontIndirect(LogFont));

//      Canvas.TextRect(ClipRect, R.Left, R.Top, PrintStr);

      Canvas.TextOut(R.Left, R.Top, PrintStr);

      FCurrentLine := 1;
      DeleteObject(SelectObject(Canvas.Handle, OldFont));
      PrintFinished := True;
  end;
end;

//RSV procedure TQRCustomLabel.Print(OfsX, OfsY : integer);
procedure TQRCustomLabel.Print(OfsX, OfsY : extended);
var
  aCanvas : TCanvas;
begin
  if IsEnabled then begin
    if FDegree = 0 then begin
      aCanvas := QRPrinter.Canvas;
      aCanvas.Font := Font;
      with QRPrinter do
        PrintToCanvas(QRPrinter.Canvas,
                      OfsX + Size.Left, OfsY + Size.Top,
                      Size.Width, Size.Height,
//                      YFact * 254 * FWidth / 72
                      GetTextHeight('Wy', Font) * 254 / Screen.PixelsPerInch, AutoStretch);
//                      aCanvas.TextHeight('W') / QRPrinter.YFactor, AutoStretch);
    end
    else DegreePrint(OfsX, OfsY);
    inherited Print(OfsX, OfsY);
  end;
end;

function TQRCustomLabel.StoredFont: Boolean;
begin
  Result := not ParentFont;
end;

procedure TQRCustomLabel.DegreePaint;
var
  aRect, ClipRect : TRect;
  LogFont : TLogFont;
  OldFont: HFont;
//  Factor: Integer;
begin

  aRect:=ClientRect;
  ClipRect:=aRect;

  if not Transparent then begin
    Canvas.Brush.Color:=Color;
    Canvas.Brush.Style:=bsSolid;
    Canvas.FillRect(ClientRect);
  end;
  SetBkMode(Canvas.Handle, 1);
//RSV  PaintCorners;
  Canvas.Brush.Style:=bsClear;
  Canvas.Font:=Font;
  Canvas.Font.Size:=round(Canvas.Font.Size * Zoom/100);
  Canvas.Font.Color:=Font.Color;
  CalculateRotatedPosition(Self, Font.Handle, Alignment, aRect,
                           Zoom, False, 0, 0, Caption);
  GetObject(Canvas.Font.Handle, SizeOf(LogFont), @LogFont);
  LogFont.lfEscapement:=Degree;
  OldFont:=SelectObject(Canvas.Handle, CreateFontIndirect(LogFont));

  Canvas.TextRect(ClipRect, aRect.Left, aRect.Top, Caption);
  DeleteObject(SelectObject(Canvas.Handle, OldFont));
//  aRect := GetClientRect;
//  dec(aRect.Bottom);
//  dec(aRect.Right);
//  TQRFixFrame(Frame).PaintIt(Canvas, aRect, Screen.PixelsPerInch / 254 * Zoom/100, Screen.PixelsPerInch / 254 * Zoom/100);
end;

procedure TQRCustomLabel.SetAutoStretch(Value : boolean);
begin
  FAutoStretch := Value;
  Invalidate;
end;

procedure TQRCustomLabel.SetAutoSize(Value: Boolean);
begin
  FAutoSize := Value;
  Invalidate;
end;

procedure TQRCustomLabel.SetCaption(Value : TCaption);
begin
  FCaption := Value;
  FPrintCaption := Value;
  DoneFormat := false;
  FormatLines;
  Invalidate;
end;

procedure TQRCustomLabel.SetName(const Value: TComponentName);
begin
  if ((Caption = '') or (Caption = Name)) then
    Caption := Value;
  inherited SetName(Value);
end;

procedure TQRCustomLabel.SetParent(AParent : TWinControl);
begin
  inherited SetParent(AParent);
  FormatLines;
end;

procedure TQRCustomLabel.SetAlignment(Value : TAlignment);
begin
  inherited SetAlignment(Value);
end;

procedure TQRCustomLabel.SetWordWrap(Value : boolean);
begin
  FWordWrap := Value;
  Invalidate;
end;
{
procedure TQRCustomLabel.SetIndent(const Value: Integer);
begin
  if FIndent <> Value then begin
    if Value >= 0
    then FIndent := Value
    else FIndent := 0;
    FormatLines;
    Invalidate;
  end;
end;
}
procedure TQRCustomLabel.SetDegree(Value: Integer);
begin
  if Value<>FDegree then begin
    if Value<>0 then AutoSize := False;
    while Value>3600 do Value := Value-3600;
    FDegree:=Value;
    Invalidate;
  end;
end;

{ TQRLabel }
{RSV}
{$ifdef USE_QREDITOR}  //RAA
function TQRLabel.GetEditorClass : TQRPrintableEditorClass;
begin
  Result := TQRLabelEditor;
end;

{$endif}
{
function TQRLabel.GetEditorClass : TQRPrintableEditorClass;
begin
  Result := TQRPrintableEditorClass(GetClass('TQRLabelEditor'));
end;
}
function TQRCustomLabel.SaveIndents: Boolean;
begin
  Result := not FIndent.IsDefault;
end;

{ TQRMemo }

function TQRMemo.GetCaptionBased : boolean;
begin
  Result := false;
end;

procedure TQRMemo.Paint;
begin
  if (Lines.Count > 0) and (Caption > '') then
    Caption := '';
  inherited Paint;
end;

//procedure TQRMemo.Print(OfsX, OfsY : integer);
procedure TQRMemo.Print(OfsX, OfsY : extended);
begin
{  if (Lines.Count > 0) then}
  Caption := '';
  inherited Print(OfsX, OfsY);
  Caption := Name;
end;

{ TQRDBText }

constructor TQRDBText.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  DataSourceName := '';
  ComboBox := nil;
  IsMemo := false;
  FBlankIfZero := false;
end;

procedure TQRDBText.SetDataSet(Value : TDataSet);
begin
  FDataSet := Value;
  if Value <> nil then
    Value.FreeNotification(self);
end;

function TQRDBText.GetCaptionBased : boolean;
begin
  Result := not IsMemo;
end;

procedure TQRDBText.SetDataField(Value : string);
begin
  FDataField := Value;
  Caption := Value;
end;

procedure TQRDBText.Loaded;
var
  aComponent : TComponent;
begin
  inherited Loaded;
  if DataSourceName<>'' then
  begin
    aComponent := Owner.FindComponent(DataSourceName);
    if (aComponent <> nil) and (aComponent is TDataSource) then
      DataSet:=TDataSource(aComponent).DataSet;
  end;
end;

procedure TQRDBText.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) then
    if AComponent = FDataSet then
      FDataSet := nil;
end;

procedure TQRDBText.SetMask(Value : string);
begin
  FMask := Value;
end;

procedure TQRDBText.Prepare;
begin
  inherited Prepare;
  if assigned(FDataSet) then
  begin
    Field := FDataSet.FindField(FDataField);
    if Field <> nil then
    begin
      FieldNo := Field.Index;
      FieldOK := true;
      if (Field is TMemoField) or (Field is TBlobField) then
      begin
        FPrintCaption := '';
        IsMemo := true;
      end
        else IsMemo := false;
    end;
  end else
  begin
    Field := nil;
    FieldOK := false;
  end;
end;

//rsv procedure TQRDBText.Print(OfsX, OfsY : integer);
procedure TQRDBText.Print(OfsX, OfsY : extended);
begin
  if IsEnabled then
  begin
    if FieldOK then
    begin
      if FDataSet.DefaultFields then
        Field := FDataSet.Fields[FieldNo];
    end
    else
      Field := nil;
    if assigned(Field) then
    begin
      try
        if (Field is TMemoField) or
           (Field is TBlobField) then
        begin
          Lines.Text := TMemoField(Field).AsString;
        end else
          if (Mask = '') or (Field is TStringField) then
            if not (Field is TBlobField) then
              if ( FBlankIfZero and ( Field.Value = 0)) // TSK
                then FPrintCaption := ''
                else FPrintCaption := Field.DisplayText
            else
              FPrintCaption := Field.AsString
          else
          begin
            if (Field is TIntegerField) or
               (Field is TSmallIntField) or
               (Field is TWordField) then
               FPrintCaption := FormatFloat(Mask, TIntegerField(Field).Value * 1.0)
            else
              if (Field is TFloatField) or
                 (Field is TCurrencyField) or
                 (Field is TBCDField) then
                  FPrintCaption := FormatFloat(Mask,TFloatField(Field).Value)
              else
                if (Field is TDateTimeField) or
                   (Field is TDateField) or
                   (Field is TTimeField) then
                  FPrintCaption := FormatDateTime(Mask,TDateTimeField(Field).Value);
          end;
      except
        FPrintCaption := '';
      end;
    end else
      FPrintCaption := '';
    DoneFormat := false;
    inherited Print(OfsX,OfsY);
  end;
end;

procedure TQRDBText.Unprepare;
begin
  Field := nil;
  inherited Unprepare;
  if DataField <> '' then
    SetDataField(DataField) { Reset component caption }
  else
    SetDataField(Name);
end;

function TQRDBText.UseRightToLeftAlignment: Boolean;
begin
  Result := QRDBUseRightToLeftAlignment(Self, Field);
end;

{ TQRExpr }

constructor TQRExpr.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  Evaluator := TQREvaluator.Create;
  FExpression := '';
  FMask := '';
end;

destructor TQRExpr.Destroy;
begin
  Evaluator.Free;
  inherited Destroy;
end;

function TQRExpr.GetValue : TQREvResult;
begin
  if Evaluator.Prepared then
    result := Evaluator.Value
  else
    result.Kind := resError;
end;

procedure TQRExpr.Reset;
begin
   Evaluator.Reset;
end;

procedure TQRExpr.SetMaster(AComponent : TComponent);
begin
  FMaster := AComponent;
end;

procedure TQRExpr.QRNotification(Sender : TObject; Operation : TQRNotifyOperation);
begin
  inherited QRNotification(Sender, Operation);
  case Operation of
    qrMasterDataAdvance : Evaluator.DoAggregate;
  end;
end;

{$ifdef USE_QREDITOR}  //RAA
function TQRExpr.GetEditorClass : TQRPrintableEditorClass;
begin
  Result := TQRExprEditor;
end;

{
function TQRExpr.GetEditorClass : TQRPrintableEditorClass;
begin
  Result := TQRPrintableEditorClass(GetClass('TQRExprEditor'));
end;
}
{$endif}

procedure TQRExpr.Prepare;
begin
  inherited Prepare;
  Evaluator.DataSets := ParentReport.AllDataSets;
  Evaluator.Environment := ParentReport.Functions;
  Evaluator.Prepare(FExpression);
  if assigned(FMaster) then
  begin
    if Master is TQuickRep then
      TCustomQuickRep(Master).AddNotifyClient(Self)
    else
      if Master is TQRSubDetail then
        TQRSubDetail(Master).AddNotifyClient(Self);
  end else
    if Evaluator.IsAggreg then ParentReport.AddNotifyClient(Self);
  Reset;
end;

procedure TQRExpr.Unprepare;
begin
  Evaluator.DataSets := nil;
  Evaluator.Unprepare;
  inherited Unprepare;
  SetExpression(Expression); { Reset component caption... }
end;

//rsvprocedure TQRExpr.Print(OfsX, OfsY : integer);
procedure TQRExpr.Print(OfsX, OfsY : extended);
var
  aValue : TQREvResult;
begin
  if IsEnabled then
  begin
    aValue := Evaluator.Value;
    case aValue.Kind of
      resInt : FPrintCaption := FormatFloat(Mask, aValue.IntResult * 1.0);
      resString : FPrintCaption := aValue.strResult;
      resDouble : //TSK
        begin
          if FCurrency
            then FPrintCaption := FloatToStrF(aValue.DblResult, ffCurrency, 19, CurrencyDecimals)
            else FPrintCaption := FormatFloat(Mask,aValue.DblResult);
        end;
      resBool : if aValue.booResult then FPrintCaption := 'True' else FPrintCaption := 'False'; {<-- do not resource }
      resError : FPrintCaption := aValue.strResult;
    end;
    inherited Print(OfsX, OfsY);
    if ResetAfterPrint then Reset;
  end;
end;

procedure TQRExpr.SetExpression(Value : string);
begin
  FExpression := Value;
  if Value='' then
    Caption := '(' + SqrNone + ')'
  else
    Caption := Value;
  Invalidate;
end;

procedure TQRExpr.SetMask(Value : string);
begin
  FMask := Value;
  SetExpression(Expression);
end;

procedure TQRExpr.SetCurrency(const Value: Boolean);
begin
  FCurrency := Value;
  SetExpression(Expression);
end;

{ TQRSysData }

constructor TQRSysData.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  FText := '';
  CreateCaption;
end;

//rsv procedure TQRSysData.Print(OfsX,OfsY : integer);
procedure TQRSysData.Print(OfsX,OfsY : extended);
begin
  case FData of
    qrsTime : Caption := FText + FormatDateTime('t',SysUtils.Time);
    qrsDate : Caption := FText + FormatDateTime('c',SysUtils.Date);
    qrsDateTime : Caption := FText + FormatDateTime('c',Now);
    qrsPageNumber : Caption := FText + IntToStr(ParentReport.PageNumber);
    qrsReportTitle: Caption := FText + ParentReport.ReportTitle;
    qrsDetailCount: if ParentReport is TQuickRep then
      Caption := FText+IntToStr(TQUickRep(ParentReport).RecordCount);
    qrsDetailNo : if ParentReport is TQuickRep then
      Caption := FText+IntToStr(TQuickRep(ParentReport).RecordNumber);
  end;
  inherited Print(OfsX,OfsY);
end;

procedure TQRSysData.CreateCaption;
begin
  case FData of
    qrsTime : Caption := FText + '(' + SqrTime + ')';
    qrsDate : Caption := FText + '(' + SqrDate + ')';
    qrsDateTime : Caption := FText + '(' + SqrDateTime + ')';
    qrsPageNumber : Caption := FText + '(' + SqrPageNum + ')';
    qrsReportTitle: Caption := FText + '(' + SqrReportTitle + ')';
    qrsDetailCount: Caption := FText + '(' + SqrDetailCount + ')';
    qrsDetailNo : Caption := Ftext + '(' + SqrDetailNo + ')';
  end;
  Invalidate;
end;

procedure TQRSysData.SetData(Value : TQRSysDataType);
begin
  FData := Value;
  CreateCaption;
end;

procedure TQRSysData.SetText(Value : String);
begin
  FText := Value;
  CreateCaption;
end;

{ TQRShape }

constructor TQRShape.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  width := 65;
  Height := 65;
  FPen := TPen.Create;
  FBrush := TBrush.Create;
  FBrush.OnChange := StyleChanged;
  FPen.OnChange := StyleChanged;
end;

procedure TQRShape.StyleChanged(Sender : TObject);
begin
  invalidate;
end;

procedure TQRShape.SetShape(Value : TQRShapeType);
begin
  if FShape <> value then
  begin
    FShape := Value;
    Invalidate;
  end
end;

procedure TQRShape.SetBrush(Value : TBrush);
begin
  FBrush.Assign(Value);
end;

procedure TQRShape.SetPen(Value : TPen);
begin
  FPen.Assign(value);
end;

procedure TQRShape.Paint;
begin
  inherited paint;
  with Canvas do
  begin
    Pen.Assign(FPen);
    Pen.Width := muldiv(Pen.Width,Zoom,100);
    if (FPen.Width > 0) and (Pen.Width<1) then Pen.Width := 1;
    Brush := FBrush;
    Case FShape of
      qrsRectangle : Rectangle(0,0,Width,Height);
      qrsCircle : Ellipse(0,0,Width,Height);
      qrsHorLine : begin
          MoveTo(0,Height div 2);
          LineTo(Width,Height div 2);
        end;
      qrsVertLine : begin
          MoveTo(Width div 2,0);
          LineTo(Width div 2,Height);
        end;
      qrsTopAndBottom : begin
          MoveTo(0,0);
          LineTo(Width,0);
          MoveTo(0,Height-1);
          LineTo(Width,Height-1);
        end;
      qrsRightAndLeft : begin
          MoveTo(0,0);
          LineTo(0,Height);
          MoveTo(Width-1,0);
          LineTo(Width-1,Height);
        end;
      qrsDownLine: begin
          MoveTo(0,0);
          LineTo(Width-1, Height-1);
        end;
      qrsUpLine: begin
          MoveTo(0,Height-1);
          LineTo(Width-1,0);
        end;
      qrsTriangle1: begin
          Polygon([Point(0,0),Point(Width-1,0),Point(0,Height-1)]);
        end;
      qrsTriangle2: begin
          Polygon([Point(0,0),Point(Width-1,0),Point(Width-1,Height-1)]);
        end;
      qrsTriangle3: begin
          Polygon([Point(0,0),Point(Width-1,Height-1),Point(0,Height-1)]);
        end;
      qrsTriangle4: begin
          Polygon([Point(Width-1,0),Point(0,Height-1),Point(Width-1,Height-1)]);
        end;
    end
  end
end;
                                
//RSV procedure TQRShape.Print(OfsX,OfsY : Integer);
procedure TQRShape.Print(OfsX,OfsY : extended);
var
 oPen: TPen;
begin
  oPen := QRPrinter.Canvas.Pen;
  if ParentReport.FinalPass and IsEnabled then
  try
    QRPrinter.Canvas.Brush.Assign(Brush);
    QRPrinter.Canvas.Pen := Pen;
    QRPrinter.Canvas.Pen.Width :=
         (QRPrinter.XSize(Pen.Width*254/Screen.PixelsPerInch)+
          QRPrinter.YSize(Pen.Width*254/Screen.PixelsPerInch)) shr 1;
    if (FPen.Width > 0) and (QRPrinter.Canvas.Pen.Width < 1) then QRPrinter.Canvas.Pen.Width := 1;
//    if QRPrinter.Destination = qrdPrinter
//    then QRPrinter.Canvas.Pen.Width := Round(QRPrinter.Canvas.Pen.Width,254,72);
    with QRPrinter do
    begin
      with Canvas do
      begin
        case FShape of
          qrsRectangle : Rectangle(XPos(OfsX + Size.Left), YPos(OfsY + Size.Top),
            XPos(OfsX+Size.Left + Size.Width), YPos(OfsY + Size.Top + Size.Height));
          qrsCircle : Ellipse(XPos(OfsX + Size.Left), YPos(OfsY + Size.Top),
            XPos(OfsX+Size.Left + Size.Width), YPos(OfsY + Size.Top + Size.Height));
          qrsHorLine : begin
              MoveTo(XPos(OfsX + Size.Left), YPos(OfsY + Size.Top + Size.Height / 2));
              LineTo(XPos(OfsX + Size.Left + Size.Width),YPos(OfsY + Size.Top + Size.Height/2));
            end;
          qrsVertLine : begin
              MoveTo(XPos(OfsX+Size.Left + Size.Width / 2), YPos(OfsY + Size.Top));
              LineTo(XPos(OfsX+Size.Left + Size.Width / 2), Ypos(OfsY + Size.Height + Size.Top));
            end;
          qrsTopAndBottom : begin
              MoveTo(XPos(OfsX + Size.Left), YPos(OfsY + Size.Top));
              LineTo(Xpos(OfsX + Size.Left + Size.Width), YPos(OfsY + Size.Top));
              MoveTo(Xpos(OfsX + Size.Left), YPos(OfsY + Size.Top + Size.Height));
              LineTo(Xpos(OfsX + Size.Left + Size.Width), Ypos(OfsY + Size.Top + Size.Height));
            end;
          qrsRightAndLeft : Begin
              MoveTo(XPos(OfsX + Size.Left), YPos(OfsY + Size.Top));
              LineTo(Xpos(OfsX + Size.Left), YPos(OfsY + Size.Top + Size.Height));
              MoveTo(XPos(OfsX + Size.Left + Size.Width), YPos(OfsY + Size.Top));
              LineTo(XPos(OfsX + Size.Left + Size.Width), YPos(OfsY + Size.Top + Size.Height));
            end;
          qrsDownLine: begin
              MoveTo(XPos(OfsX + Size.Left),YPos(OfsY + Size.Top));
              LineTo(XPos(OfsX+Size.Left + Size.Width),
                     YPos(OfsY + Size.Top + Size.Height));
             end;
          qrsUpLine: begin
              MoveTo(XPos(OfsX + Size.Left),YPos(OfsY + Size.Top + Size.Height));
              LineTo(XPos(OfsX+Size.Left + Size.Width),
                     YPos(OfsY + Size.Top));
            end;
          qrsTriangle1: begin
              Polygon([Point(XPos(OfsX + Size.Left),YPos(OfsY + Size.Top)),
                       Point(XPos(OfsX+Size.Left + Size.Width),YPos(OfsY + Size.Top)),
                       Point(XPos(OfsX + Size.Left),YPos(OfsY + Size.Top + Size.Height))]);
            end;
          qrsTriangle2: begin
              Polygon([Point(XPos(OfsX + Size.Left),YPos(OfsY + Size.Top)),
                              Point(XPos(OfsX+Size.Left + Size.Width),YPos(OfsY + Size.Top)),
                              Point(XPos(OfsX+Size.Left + Size.Width),YPos(OfsY + Size.Top + Size.Height))]);
            end;
          qrsTriangle3: begin
              Polygon([Point(XPos(OfsX + Size.Left),YPos(OfsY + Size.Top)),
                       Point(XPos(OfsX+Size.Left + Size.Width),YPos(OfsY + Size.Top + Size.Height)),
                       Point(XPos(OfsX + Size.Left),YPos(OfsY + Size.Top + Size.Height))]);
            end;
          qrsTriangle4: begin
              Polygon([Point(XPos(OfsX+Size.Left + Size.Width),YPos(OfsY + Size.Top)),
                       Point(XPos(OfsX + Size.Left),YPos(OfsY + Size.Top + Size.Height)),
                       Point(XPos(OfsX+Size.Left + Size.Width),YPos(OfsY + Size.Top + Size.Height))]);
            end;
        end
      end
    end
  finally
    QRPrinter.Canvas.Pen := oPen;
  end
end;

destructor TQRShape.Destroy;
begin
  FPen.Free;
  FBrush.Free;
  inherited Destroy;
end;

{ TQRImage }

  procedure PrintBitmap(aCanvas : TCanvas; Dest : TRect; Bitmap : TBitmap);
  var
    Info : PBitmapInfo;
    InfoSize : DWORD;
    Image : Pointer;
{$ifdef ver80}
    ImageSize : Longint;
{$else}
    ImageSize : DWord;
{$endif}
  begin
    with Bitmap do
    begin
      GetDIBSizes(Handle, InfoSize, ImageSize);
      Info := AllocMem(InfoSize);
      try
        Image := AllocMem(ImageSize);
        try
          GetDIB(Handle, Palette, Info^, Image^);
          if not Monochrome then
            SetStretchBltMode(ACanvas.Handle, STRETCH_DELETESCANS);
          with Info^.bmiHeader do
            StretchDIBits(aCanvas.Handle, Dest.Left, Dest.Top,
              Dest.RIght - Dest.Left, Dest.Bottom - Dest.Top,
              0, 0, biWidth, biHeight, Image, Info^, DIB_RGB_COLORS, SRCCOPY);
        finally
          FreeMem(Image, ImageSize);
        end;
      finally
        FreeMem(Info, InfoSize);
      end;
    end;
  end;

constructor TQRImage.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FPicture := TPicture.Create;
  FPicture.OnChange := PictureChanged;
  Height := 105;
  Width := 105;
end;

destructor TQRImage.Destroy;
begin
  FPicture.Free;
  inherited Destroy;
end;

function TQRImage.GetPalette: HPALETTE;
begin
  Result := 0;
  if FPicture.Graphic is TBitmap then
    Result := TBitmap(FPicture.Graphic).Palette;
end;

procedure TQRImage.Paint;
var
  Dest: TRect;
begin
  if csDesigning in ComponentState then
    with inherited Canvas do
    begin
      Pen.Style := psDash;
      Brush.Style := bsClear;
      Rectangle(0, 0, Width, Height);
    end;
  if Stretch then
    Dest := ClientRect
  else if Center then
    Dest := Bounds((Width - Picture.Width) div 2, (Height - Picture.Height) div 2,
      Picture.Width, Picture.Height)
  else
    Dest := Rect(0, 0, Picture.Width, Picture.Height);
  if Zoom <> 100 then
  begin
    Dest.Right := Dest.Left + (Dest.Right - Dest.Left) * Zoom div 100;
    Dest.Bottom := Dest.Top + (Dest.Bottom - Dest.Top) * Zoom div 100;
  end;
  with inherited Canvas do
    StretchDraw(Dest, Picture.Graphic);
//  Inherited Paint;
end;

//rsv procedure TQRImage.Print(OfsX,OfsY : Integer);
procedure TQRImage.Print(OfsX,OfsY : extended);
var
  Dest : TRect;
  DC, SavedDC : THandle;
begin
  Dest.Top := QRPrinter.YPos(OfsY + Size.Top);
  Dest.Left := QRPrinter.XPos(OfsX + Size.Left);
  Dest.Right := QRPrinter.XPos(OfsX + Size.Width + Size.Left);
  Dest.Bottom := QRPrinter.YPos(OfsY + Size.Height + Size.Top);
  if Stretch then
  begin
    if Picture.Graphic is TBitmap then
      PrintBitmap(QRPrinter.Canvas, Dest, TBitmap(Picture.Graphic))
    else
      with QRPrinter.Canvas do
        StretchDraw(Dest, Picture.Graphic);
  end else
  begin
    IntersectClipRect(QRPrinter.Canvas.Handle, Dest.Left, Dest.Top, Dest.Right, Dest.Bottom);
    DC := GetDC(QRPrinter.Canvas.Handle);
    SavedDC := SaveDC(DC);
    Dest.Right := Dest.Left +
      round(Picture.Width / Screen.PixelsPerInch * 254 * ParentReport.QRPrinter.XFactor);
    Dest.Bottom := Dest.Top +
      round(Picture.Height / Screen.PixelsPerInch * 254 * ParentReport.QRPrinter.YFactor);
    if Center then OffsetRect(Dest,
      (QRPrinter.XSize(Size.Width) -
        round(Picture.Width / Screen.PixelsPerInch * 254 * ParentReport.QRPrinter.XFactor)) div 2,
      (QRPrinter.YSize(Size.Height) -
        round(Picture.Height / Screen.PixelsPerInch * 254 * ParentReport.QRPrinter.YFactor)) div 2);
    if Picture.Graphic is TBitmap then
      PrintBitmap(QRPrinter.Canvas, Dest, TBitmap(Picture.Graphic))
    else
      QRPrinter.Canvas.StretchDraw(Dest, Picture.Graphic);
   RestoreDC(DC, SavedDC);
   SelectClipRgn(QRPrinter.Canvas.Handle, 0);
  end;
  inherited Print(OfsX,OfsY);
end;

function TQRImage.GetCanvas: TCanvas;
var
  Bitmap: TBitmap;
begin
  if Picture.Graphic = nil then
  begin
    Bitmap := TBitmap.Create;
    try
      Bitmap.Width := Width;
      Bitmap.Height := Height;
      Picture.Graphic := Bitmap;
    finally
      Bitmap.Free;
    end;
  end;
  if Picture.Graphic is TBitmap then
    Result := TBitmap(Picture.Graphic).Canvas
  else
    Result := nil;
{    raise EInvalidOperation.CreateRes(SImageCanvasNeedsBitmap)};
end;

procedure TQRImage.SetAutoSize(Value: Boolean);
begin
  FAutoSize := Value;
  PictureChanged(Self);
end;

procedure TQRImage.SetCenter(Value: Boolean);
begin
  if FCenter <> Value then
  begin
    FCenter := Value;
    PictureChanged(Self);
  end;
end;

procedure TQRImage.SetPicture(Value: TPicture);
begin
  FPicture.Assign(Value);
end;

procedure TQRImage.SetStretch(Value: Boolean);
begin
  if Value <> FStretch then
  begin
    FStretch := Value;
    PictureChanged(Self);
  end;
end;

procedure TQRImage.PictureChanged(Sender: TObject);
begin
  if AutoSize and (Picture.Width > 0) and (Picture.Height > 0) then
    SetBounds(Left, Top, Picture.Width, Picture.Height);
  if (Picture.Graphic is TBitmap) and (Picture.Width >= Width) and
    (Picture.Height >= Height) then
    ControlStyle := ControlStyle + [csOpaque] else
    ControlStyle := ControlStyle - [csOpaque];
  Invalidate;
end;

{ TQRDBImage }

constructor TQRDBImage.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csFramed, csOpaque];
  Width := 105;
  Height := 105;
  FPicture := TPicture.Create;
  FPicture.OnChange := PictureChanged;
  FCenter := True;
end;

destructor TQRDBImage.Destroy;
begin
  FPicture.Free;
  inherited Destroy;
end;

procedure TQRDBImage.Prepare;
begin
  inherited Prepare;
  if assigned(FDataSet) then
  begin
    FField := DataSet.FindField(FDataField);
    if Field is TBlobField then
    begin
      Caption := '';
    end;
  end else
    FField := nil;
end;

//rsv procedure TQRDBImage.Print(OfsX, OfsY : integer);
procedure TQRDBImage.Print(OfsX, OfsY : extended);
var
  H: integer;
  Dest: TRect;
  DrawPict: TPicture;
begin
  with QRPrinter.Canvas do
  begin
    Brush.Style := bsSolid;
    Brush.Color := Color;
    DrawPict := TPicture.Create;
    H := 0;
    try
      if assigned(FField) and (FField is TBlobField) then
      begin
        DrawPict.Assign(FField);
        if (DrawPict.Graphic is TBitmap) and
          (DrawPict.Bitmap.Palette <> 0) then
        begin
          H := SelectPalette(Handle, DrawPict.Bitmap.Palette, false);
          RealizePalette(Handle);
        end;
        Dest.Left := QRPrinter.XPos(OfsX + Size.Left);
        Dest.Top := QRPrinter.YPos(OfsY + Size.Top);
        Dest.Right := QRPrinter.XPos(OfsX + Size.Width + Size.Left);
        Dest.Bottom := QRPrinter.YPos(OfsY + Size.Height + Size.Top);
        if Stretch then
        begin
          if (DrawPict.Graphic = nil) or DrawPict.Graphic.Empty then
            FillRect(Dest)
          else
            with QRPrinter.Canvas do
              StretchDraw(Dest, DrawPict.Graphic);
        end else
        begin
          IntersectClipRect(Handle, Dest.Left, Dest.Top, Dest.Right, Dest.Bottom);
          Dest.Right := Dest.Left +
            round(DrawPict.Width / Screen.PixelsPerInch * 254 * ParentReport.QRPrinter.XFactor);
          Dest.Bottom := Dest.Top +
            round(DrawPict.Height / Screen.PixelsPerInch * 254 * ParentReport.QRPrinter.YFactor);
          if Center then OffsetRect(Dest,
            (QRPrinter.XSize(Size.Width) -
              round(DrawPict.Width / Screen.PixelsPerInch * 254 * ParentReport.QRPrinter.XFactor)) div 2,
            (QRPrinter.YSize(Size.Height) -
              round(DrawPict.Height / Screen.PixelsPerInch * 254 * ParentReport.QRPrinter.YFactor)) div 2);
          QRPrinter.Canvas.StretchDraw(Dest, DrawPict.Graphic);
          SelectClipRgn(Handle, 0);
        end;
      end;
    finally
      if H <> 0 then SelectPalette(Handle, H, True);
      DrawPict.Free;
    end;
  end;
  inherited Print(OfsX,OfsY);
end;

procedure TQRDBImage.Unprepare;
begin
  FField := nil;
  inherited Unprepare;
end;

procedure TQRDBImage.SetDataSet(Value: TDataSet);
begin
  FDataSet := Value;
  if Value <> nil then Value.FreeNotification(Self);
end;

procedure TQRDBImage.SetDataField(const Value: string);
begin
  FDataField := Value;
end;

function TQRDBImage.GetPalette: HPALETTE;
begin
  Result := 0;
  if FPicture.Graphic is TBitmap then
    Result := TBitmap(FPicture.Graphic).Palette;
end;

procedure TQRDBImage.SetCenter(Value: Boolean);
begin
  if FCenter <> Value then
  begin
    FCenter := Value;
    Invalidate;
  end;
end;

procedure TQRDBImage.SetPicture(Value: TPicture);
begin
  FPicture.Assign(Value);
end;

procedure TQRDBImage.SetStretch(Value: Boolean);
begin
  if FStretch <> Value then
  begin
    FStretch := Value;
    Invalidate;
  end;
end;

procedure TQRDBImage.Paint;
var
  W, H: Integer;
  R: TRect;
  S: string;
begin
  with Canvas do
  begin
    Brush.Style := bsSolid;
    Brush.Color := Color;
    Font := Self.Font;
    if Field <> nil then
      S := Field.DisplayLabel
    else S := Name;
    S := '(' + S + ')';
    W := TextWidth(S);
    H := TextHeight(S);
    R := ClientRect;
    TextRect(R, (R.Right - W) div 2, (R.Bottom - H) div 2, S);
  end;
  Inherited Paint;
end;

procedure TQRDBImage.PictureChanged(Sender: TObject);
begin
  FPictureLoaded := True;
  Invalidate;
end;

procedure TQRDBImage.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = DataSet) then
    DataSet := nil;
end;

procedure TQRDBImage.LoadPicture;
begin
  if not FPictureLoaded and (Field is TBlobField) then
    Picture.Assign(FField);
end;


{ TQRCusomRichText }

constructor TQRCustomRichText.Create(AOwner : TComponent);
begin
  FRichEdit := nil;
  inherited Create(AOwner);
  FRichEdit := TQRRichEdit.Create(Self);
  FRichEdit.Parent := self;
  FRichEdit.BorderStyle := bsNone;
  AutoStretch := false;
  PrintFinished := true;
  Width := 100;
  Height := 100;
end;

destructor TQRCustomRichText.Destroy;
begin
  FRichEdit.Free;
  inherited Destroy;
end;

function TQRCustomRichText.GetControlsAlignment: TAlignment;
begin
  Result := FRichEdit.Alignment;
end;

function TQRCustomRichText.GetAlignment : TAlignment;
begin
  result := FRichEdit.Alignment;
end;

function TQRCustomRichText.GetColor : TColor;
begin
  result := FRichEdit.Color;
end;

function TQRCustomRichText.GetFont : TFont;
begin
  result := FRichEdit.Font;
end;

function TQRCustomRichText.GetLines : TStrings;
begin
  result := FRichEdit.Lines;
end;

//procedure TQRCustomRichText.Print(OfsX, OfsY : integer);
procedure TQRCustomRichText.Print(OfsX, OfsY : extended);
var
  Range: TFormatRange;
  LogX, LogY : integer;
  TextLength : integer;
  HasExpanded : boolean;
  OldMapMode : integer;
  ARichEdit : TCustomRichEdit;
  Expanded : extended;

  function Render(RenderIt : boolean) : integer;
  begin
    if RenderIt then
      Result := SendMessage(ARichEdit.Handle, EM_FORMATRANGE, 0, Longint(@Range))
//      Result := SendMessage(ARichEdit.Handle, EM_DISPLAYBAND, 0, Longint(@Range.rc))
    else
      Result := SendMessage(ARichEdit.Handle, EM_FORMATRANGE, 1, Longint(@Range));
  end;

begin
  if assigned(FParentRichEdit) then
    ARichEdit := ParentRichEdit
  else
    ARichEdit := FRichEdit;
  FillChar(Range, SizeOf(TFormatRange), 0);

  with Range do
  begin
    hdc := ParentReport.QRPrinter.Canvas.Handle;
    hdcTarget := hdc;
    LogX := GetDeviceCaps(hdc, LOGPIXELSX);
    LogY := GetDeviceCaps(hdc, LOGPIXELSY);

    rc := Rect(QRPrinter.XPos(OfsX + Size.Left) * 1440 div LogX,
               QRPrinter.YPos(OfsY + Size.Top) * 1440 div LogY,
               QRPrinter.XPos(OfsX + Size.Left + Size.Width) * 1440 div LogX,
               QRPrinter.YPos(OfsY + Size.Top + Size.Height) * 1440 div LogY);
    rcPage := Rect(0, 0, QRPrinter.XSize(QRPrinter.PaperWidth) * 1440 div LogX,
                         QRPrinter.YSize(QRPrinter.PaperLength) * 1440 div LogY);

    if PrintFinished then
      LastChar := 0;
    HasExpanded := false;
    Expanded := 0;
    TextLength := ARichEdit.GetTextLen;
    chrg.cpMax := -1;
    chrg.cpMin := LastChar;
    OldMapMode := SetMapMode(hdc, MM_TEXT);

//rsv
    SendMessage(ARichEdit.Handle, EM_FORMATRANGE, 0, 0);

    LastChar := Render(false);
    if (LastChar < TextLength) and AutoStretch and (Parent is TQRCustomBand) and (TextLength > 0) then
    begin
      PrintFinished := false;
      while (LastChar <= TextLength) and TQRCustomBand(Parent).CanExpand(50) do
      begin
        TQRCustomBand(Parent).ExpandBand(50, Expanded, HasExpanded);
        rc.Bottom := QRPrinter.YPos(OfsY + Size.Top + Size.Height + Expanded) * 1440 div LogY;
        LastChar := Render(false);
      end;
      LastChar := Render(true);
      if (LastChar >= TextLength) or (LastChar = -1) then
      begin
        LastChar := TextLength;
        PrintFinished := true;
      end
    end else
    begin
      LastChar := Render(true);
      PrintFinished := true;
      inherited;
    end;
  end;

  SetMapMode(ParentReport.QRPrinter.Canvas.Handle, OldMapMode);
  if PrintFinished then
    SendMessage(ARichEdit.Handle, EM_FORMATRANGE, 0, 0);

end;

procedure TQRCustomRichText.SetAlignment(Value : TAlignment);
begin
  FRichEdit.Alignment := Value;
end;

procedure TQRCustomRichText.SetColor(Value : TColor);
begin
  FRichEdit.Color := Value;
end;

procedure TQRCustomRichText.SetFont(Value : TFont);
begin
  FRichEdit.Font := Value;
end;

procedure TQRCustomRichText.SetBounds(ALeft, ATop, AWidth, AHeight : integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if FRichEdit <> nil then FRichEdit.SetBounds(1, 1, AWidth - 2, AHeight - 2);
end;

procedure TQRCustomRichText.SetLines(Value : TStrings);
begin
  FRichEdit.Lines := Value;
  if assigned(FParentRichEdit) then
    FParentRichEdit.Lines := Value;
end;

procedure TQRCustomRichText.SetParentRichEdit(Value : TRichEdit);
begin
  FParentRichEdit := Value;
  if Value<>nil then
    FRichEdit.Lines := Value.Lines;
end;

{ TQRDBRichText }

procedure TQRDBRichText.SetDataSet(Value : TDataSet);
begin
  FDataSet := Value;
  if Value<>nil then
    Value.FreeNotification(self);
end;

procedure TQRDBRichText.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) then
    if AComponent=FDataSet then
      FDataSet := nil;
end;

procedure TQRDBRichText.Prepare;
begin
  inherited Prepare;
  if assigned(FDataSet) then
  begin
    Field := FDataSet.FindField(FDataField);
    if (Field is TBlobField) or (Field is TMemoField) then
    begin
      Caption := '';
    end;
  end else
    Field := nil;
end;

//rsv procedure TQRDBRichText.Print(OfsX, OfsY : integer);
procedure TQRDBRichText.Print(OfsX, OfsY : extended);
begin
  if assigned(Field) then
    if (Field is TMemoField) or
       (Field is TBlobField) then
      Lines.Assign(Field);
//Lines.Text := 'Test';
  inherited Print(OfsX,OfsY);
end;

procedure TQRDBRichText.Unprepare;
begin
  Field := nil;
  inherited Unprepare;
end;

{ TQRExprMemo }

type
  TMemoEvaluator = class(TQREvaluator)
  public
    Line : integer;
    Position : integer;
  end;

constructor TQRMerger.Create;
begin
  FOrgLines := TStringList.Create;
  FMergedLines := nil;
  FMerged := false;
  FPrepared := false;
  Expressions := nil;
  FDataSets := TList.Create;
end;

destructor TQRMerger.Destroy;
begin
  if Prepared then Unprepare;
  if FOrgLines <> nil then
    FOrgLines.Free;
  if FDataSets <> nil then
    FDataSets.Free;
  inherited Destroy;
end;

function TQRMerger.GetOrgLines : TStrings;
begin
  Result := FOrgLines;
end;

function TQRMerger.GetMergedLines : TStrings;
begin
  if Merged then
    Result := FMergedLines
  else
    Result := nil;
end;

procedure TQRMerger.SetOrgLines(Value : TStrings);
begin
  if FOrgLines <> nil then
    FOrgLines.Free;
  FOrgLines := Value;
end;

procedure TQRMerger.Prepare;
var
  I, start, stop, Len : integer;
  Expr : String;
  aLine : string;
  aEvaluator : TMemoEvaluator;
begin
  if Prepared then Unprepare;
  Expressions := TList.Create;
  FMergedLines := TStringList.Create;
  if Lines.Count > 0 then
  begin
    FStrippedLines := TStringList.Create;
    try
      for I := 0 to Lines.Count - 1 do
      begin
        aLine := FOrgLines[I];
        Start := AnsiPos('{', aLine);
        while Start > 0 do
        begin
          stop := AnsiPos('}', aLine);
          Len := Stop - Start - 1;
          if Len > 0 then
          begin
            Expr := copy(aLine, start + 1, Len);
            Delete(aLine, Start, Len + 2);
            aEvaluator := TMemoEvaluator.Create;
            aEvaluator.DataSets := DataSets;
            aEvaluator.Prepare(Expr);
            aEvaluator.Line := I;
            aEvaluator.Position := Start;
            Expressions.Add(aEvaluator);
          end;
          Start := AnsiPos('{', aLine);
        end;
        FStrippedLines.Add(aLine);
      end;
    finally
      FPrepared := true;
    end;
  end;
end;

procedure TQRMerger.Unprepare;
var
  I : integer;
begin
  if Prepared then
  try
    for I := 0 to Expressions.Count - 1 do
      TMemoEvaluator(Expressions[I]).Free;
    FStrippedLines.Free;
  finally
    FPrepared := false;
    FMerged := false;
  end;
  FMergedLines.Free;
  Expressions.Free;
end;

procedure TQRMerger.Merge;
var
  I : integer;
  aResult : TQREvResult;
  Replacement : string;
  aLine : string;
begin
  if (Expressions.Count > 0) then
  begin
    if Merged then
      FMergedLines.Clear;
    FMergedLines.Assign(FStrippedLines);
    for I := Expressions.Count - 1 downto 0 do
    begin
      with TMemoEvaluator(Expressions[I]) do
      begin
        aResult := Value;
        case aResult.Kind of
          resInt : Replacement := IntToStr(aResult.IntResult);
          resDouble : Replacement := FloatToStr(aResult.dblResult);
          resString : Replacement := aResult.StrResult;
          resError : Replacement := '';
        end;
        aLine := FMergedLines[Line];
        Insert(Replacement, aLine, Position);
        FMergedLines[Line] := aLine;
      end;
    end
  end else
    if not Merged then
      FMergedLines.Assign(FOrgLines);
  FMerged := true;
end;

{ TQRExrMemo component }

constructor TQRExprMemo.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  FRemoveBlankLines := false;
end;

function TQRExprMemo.GetCaptionBased : boolean;
begin
  Result := false;
end;

procedure TQRExprMemo.Prepare;
begin
  Merger := TQRMerger.Create;
  Merger.Lines.Assign(Lines);
  Merger.Datasets.Free;  { Free the object already there }
  Merger.Datasets := ParentReport.AllDatasets;
  Merger.Prepare;
  inherited Prepare;
end;

procedure TQRExprMemo.Unprepare;
begin
  Lines.Assign(Merger.Lines);
  Merger.DataSets := nil;
  Merger.Unprepare;
  Merger.Free;
  inherited Unprepare;
end;

//RSV procedure TQRExprMemo.Print(OfsX, OfsY : integer);
procedure TQRExprMemo.Print(OfsX, OfsY : extended);
var
  I, J : integer;
begin
  Caption := '';
  Merger.Merge;
  if Merger.Prepared then
  begin
    if RemoveBlankLines then
    begin
      J := Merger.MergedLines.Count - 1;
      I := 0;
      while I <= J do
      begin
        if Merger.MergedLines[I] = '' then
        begin
          Merger.MergedLines.Delete(I);
          dec(J);
        end else
          inc(I);
      end;
    end;
    Lines := Merger.MergedLines;
    if (Lines.Count > 0) and (Caption > '') then
      Caption := '';
  end;
  inherited Print(OfsX, OfsY);
  Caption := Name;
end;

procedure TQRExprMemo.Paint;
begin
  if (Lines.Count > 0) and (Caption > '') then
    Caption := '';
  inherited Paint;
end;

function TQRDBCalc.GetPrintMask : string;
begin
  Result := Mask;
end;

procedure TQRDBCalc.SetDataField(Value : string);
begin
  FDataField := Value;
  SetOperation(Operation);
end;

procedure TQRDBCalc.SetOperation(Value : TQRCalcOperation);
begin
  FOperation := Value;
  case FOperation of
    qrcSum : Expression := 'Sum('+DataField+')';         {<-- do not resource}
    qrcCount : Expression := 'Count';                    {<-- do not resource}
    qrcMax : Expression := 'Max('+DataField+')';         {<-- do not resource}
    qrcMin : Expression := 'Min('+DataField+')';         {<-- do not resource}
    qrcAverage : Expression := 'Average('+DataField+')'; {<-- do not resource}
  end
end;

procedure TQRDBCalc.SetPrintMask(Value : string);
begin
  Mask := Value;
end;

{TQRCheck}
constructor TQRCheckBox.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
                         {----}
  FPen:=TPen.Create;
  FBrush:=TBrush.Create;
  FBrush.Color:=clWhite;
  FBrush.OnChange:=StyleChanged;
  FPen.OnChange:=StyleChanged;
  FChecked:=True;
{
  Frame.DrawTop:=True;
  Frame.DrawLeft:=True;
  Frame.DrawBottom:=True;
  Frame.DrawRight:=True;
}
  Height:=17;
  Width:=17;
  Transparent:=False;
  FCtl3D:=False;
  FCheckmarkDistance:=3;
  FGlyphChecked := TPicture.Create;
  FGlyphUnchecked := TPicture.Create;
  FGlyphChecked.OnChange := PictureChanged;
  FGlyphUnchecked.OnChange := PictureChanged;
  FExpression := '';
  FCheckmarkType := cmCross;
end;

destructor TQRCheckBox.Destroy;
begin
  FPen.Free;
  FBrush.Free;
  FGlyphChecked.Free;
  FGlyphUnchecked.Free;
  inherited Destroy;
end;

procedure TQRCheckBox.PictureChanged(Sender: TObject);
begin
  Invalidate;
end;

procedure TQRCheckBox.SetBrush(Value : TBrush);
begin
  FBrush.Assign(Value);
  Invalidate;
end;

procedure TQRCheckBox.SetGlyphChecked(Value: TPicture);
begin
  FGlyphChecked.Assign(Value);
  Invalidate;
end;

procedure TQRCheckBox.SetGlyphUnchecked(Value: TPicture);
begin
  FGlyphUnchecked.Assign(Value);
  Invalidate;
end;

procedure TQRCheckBox.SetCheckmarkDistance(Value: Word);
begin
  FCheckmarkDistance:=Value;
  Invalidate;
end;

{--------------------------------------------------------------------------------}

procedure TQRCheckBox.SetCheckmarkType(Value: TQRCheckmarkType);
begin
  FCheckmarkType:=Value;
  Invalidate;
end;

procedure TQRCheckBox.SetPen(Value : TPen);
begin
  FPen.Assign(Value);
  Invalidate;
end;

{--------------------------------------------------------------------------------}

procedure TQRCheckBox.SetChecked(Value : boolean);
begin
  if Value <> FChecked then
  begin
    FChecked := Value;
    Invalidate;
  end;
end;

{--------------------------------------------------------------------------------}

procedure TQRCheckBox.SetCtl3D(Value : boolean);
begin
  FCtl3D := Value;
  Invalidate;
end;

procedure TQRCheckBox.Paint;
var
  Dest, aRect: TRect;
  DX, DY, DC: Integer;
begin
  Canvas.Pen.Assign(FPen);
  Canvas.Pen.Width := muldiv(Canvas.Pen.Width,Zoom,100);
  if (FPen.Width > 0) and (Canvas.Pen.Width<1) then Canvas.Pen.Width := 1;
  Canvas.Brush.Assign(FBrush);
//  DC := round(CheckmarkDistance / 72 * Screen.PixelsPerInch * Zoom / 100);
  DC := muldiv(CheckmarkDistance, Screen.PixelsPerInch * Zoom, 7200);
  with Canvas do
    begin
//      DX := muldiv(Frame.Width, Screen.PixelsPerInch * Zoom, 7200);
      DX := muldiv(Frame.Width, Screen.PixelsPerInch * Zoom, 25400);
      if DX < 1 then DX := 1;
      DY := DX;
      If not Transparent Then FillRect(ClientRect);
      If FCtl3d Then
        begin
          Canvas.Pen.Color:=clGray;
          Canvas.Pen.Width:=1;
          Canvas.Pen.Style:=psSolid;
          MoveTo(DX,DY);
          LineTo(Width-DX,DY);
          MoveTo(DX,DY);
          LineTo(DX,Height-DY);
          Canvas.Pen.Color:=clSilver;
          MoveTo(DX,DY);
          LineTo(Width-DX,DY);
          MoveTo(DX,DY);
          LineTo(DX,Height-DY);
          Canvas.Pen:=FPen;
        end;
      if Checked then
        begin
          If CheckmarkType=cmCross Then
            begin
              MoveTo(DC + Byte(Frame.DrawLeft)*DX{FCheckMarkDistance},
                     DC + Byte(Frame.DrawTop)*DY{FCheckMarkDistance});
              LineTo(Width - DC - Byte(Frame.DrawRight)*DX{FCheckMarkDistance},
                    Height - DC - Byte(Frame.DrawBottom)*DY{FCheckMarkDistance});
              If Pen.Width=1 Then
                begin
                  MoveTo(DC + Byte(Frame.DrawLeft)*DX{FCheckMarkDistance},
                         Height - 1 - DC - Byte(Frame.DrawBottom)*DY{FCheckMarkDistance});
                  LineTo(Width - DC - Byte(Frame.DrawRight)*DX{FCheckMarkDistance},
                         DC + Byte(Frame.DrawTop)*DY{FCheckMarkDistance} -1);
                end
              Else
                begin
                  MoveTo(DC + Byte(Frame.DrawLeft)*DX{FCheckMarkDistance},
                         Height - DC - Byte(Frame.DrawBottom)*DY{FCheckMarkDistance});
                  LineTo(Width - DC - Byte(Frame.DrawRight)*DX{FCheckMarkDistance},
                         DC + Byte(Frame.DrawTop)*DY{FCheckMarkDistance});
                end
            end
          Else
          If CheckmarkType=cmHook Then
            begin
              MoveTo(DC + Byte(Frame.DrawLeft)*DX{FCheckMarkDistance},
                     ((Height-2*DC-(Byte(Frame.DrawTop)+Byte(Frame.DrawBottom))*DY{FCheckMarkDistance}) DIV 2) +
                      DC+ Byte(Frame.DrawTop)*DY{FCheckmarkDistance});
              LineTo(((Width-2*DC-(Byte(Frame.DrawLeft)+Byte(Frame.DrawRight))*DX{FCheckMarkDistance}) DIV 3) +
                       DC+ Byte(Frame.DrawLeft)*DX{FCheckmarkDistance},
                       Height - DC - Byte(Frame.DrawBottom)*DY{FCheckmarkDistance});
              LineTo(Width - DC - Byte(Frame.DrawRight)*DX{FCheckmarkDistance},
                     DC + Byte(Frame.DrawTop)*DY{FCheckmarkDistance});
            end
          Else
          If CheckmarkType=cmGlyph Then
            begin
              Dest := ClientRect;
              Dest.Left:=Dest.Left+CheckmarkDistance;
              Dest.Top:=Dest.Top+CheckmarkDistance;
              Dest.Bottom:=Dest.Bottom-CheckmarkDistance;
              Dest.Right:=Dest.Right-CheckmarkDistance;
              if Zoom <> 100 then
                begin
                  Dest.Right := Dest.Left + (Dest.Right - Dest.Left) * Zoom div 100;
                  Dest.Bottom := Dest.Top + (Dest.Bottom - Dest.Top) * Zoom div 100;
                end;
              with inherited Canvas do
                StretchDraw(Dest, GlyphChecked.Graphic);
            end;
        end
      Else {checked=false}
        begin
          If CheckmarkType=cmGlyph then
            begin
              Dest := ClientRect;
              Dest.Left:=Dest.Left+CheckmarkDistance;
              Dest.Top:=Dest.Top+CheckmarkDistance;
              Dest.Bottom:=Dest.Bottom-CheckmarkDistance;
              Dest.Right:=Dest.Right-CheckmarkDistance;
              if Zoom <> 100 then
                begin
                  Dest.Right := Dest.Left + (Dest.Right - Dest.Left) * Zoom div 100;
                  Dest.Bottom := Dest.Top + (Dest.Bottom - Dest.Top) * Zoom div 100;
                end;
              with inherited Canvas do
                StretchDraw(Dest, GlyphUnchecked.Graphic);
            end;
        end;
    end;
//RSV  PaintCorners;
  aRect := GetClientRect;
{
  dec(aRect.Bottom);
  dec(aRect.Right);
}
//  TQRFixFrame(Frame).PaintIt(Canvas, aRect, Screen.PixelsPerInch / 254 * Zoom / 100, Screen.PixelsPerInch / 254 * Zoom/100);
  TQRFixFrame(Frame).PaintIt(Canvas, aRect, Screen.PixelsPerInch * Zoom / 25400, Screen.PixelsPerInch * Zoom/25400);
  PaintCorners;
end;

procedure TQRCheckbox.StyleChanged(Sender : TObject);
begin
  Invalidate;
end;

procedure TQRCheckBox.Print(OfsX, OfsY : Extended);
var
  DXl, DYt,
  DXr, DYb,
  cTop, cHeight,
  cLeft, cWidth,
  cmDist: Extended;
  Dest: TRect;
begin
  PrintFinished := False;
{
  DXl := Frame.Width / QRPrinter.XFactor;
  DXr := DXl*Byte(Frame.DrawRight);
  DXl := DXl*Byte(Frame.DrawLeft);
  DYt := Frame.Width / QRPrinter.YFactor;
  DYb := DYt*Byte(Frame.DrawBottom);
  DYt := DYt*Byte(Frame.DrawTop);
}
  DXl := Frame.Width * 254/72;
  DXr := DXl*Byte(Frame.DrawRight);
  DXl := DXl*Byte(Frame.DrawLeft);
  DYt := Frame.Width * 254/72;
  DYb := DYt*Byte(Frame.DrawBottom);
  DYt := DYt*Byte(Frame.DrawTop);
  if Trim(FExpression) <> EmptyStr
  then FChecked:=EvalBoolExpression(Expression, TQuickRep(ParentReport));

  if ParentReport.FinalPass and Enabled then
    begin
      QRPrinter.Canvas.Brush.Assign(FBrush);
      QRPrinter.Canvas.Pen.Assign(FPen);
      QRPrinter.Canvas.Pen.Width :=
           (QRPrinter.XSize(Pen.Width*254/Screen.PixelsPerInch)+
            QRPrinter.YSize(Pen.Width*254/Screen.PixelsPerInch)) shr 1;
      if (FPen.Width > 0) and (QRPrinter.Canvas.Pen.Width < 1) then QRPrinter.Canvas.Pen.Width := 1;

      with QRPrinter do
        begin
          with Canvas do
            begin
              If not Transparent Then
                FillRect(rect(XPos(OfsX + Size.Left),
                              YPos(OfsY + Size.Top),
                              XPos(OfsX + Size.Left + Size.Width),
                              YPos(OfsY + Size.Top + Size.Height)));

              If FCtl3d Then
                begin
                  QRPrinter.Canvas.Pen.Color:=clGray;
                  QRPrinter.Canvas.Pen.Width:=1;
                  QRPrinter.Canvas.Pen.Style:=psSolid;
                  MoveTo(XPos(OfsX + Size.Left + DXl),
                         YPos(OfsY + Size.Top + DYt));
                  LineTo(XPos(OfsX + Size.Left + Size.Width-DXr),
                         YPos(OfsY + Size.Top + DYt));
                  MoveTo(XPos(OfsX + Size.Left + DXl),
                         YPos(OfsY + Size.Top + DYt));
                  LineTo(XPos(OfsX + Size.Left + DXl),
                         YPos(OfsY + Size.Top + DYt));
                  QRPrinter.Canvas.Pen.Color:=clSilver;
                  MoveTo(XPos(OfsX + Size.Left + DXl),
                         YPos(OfsY + Size.Top + DYt));
                  LineTo(XPos(OfsX + Size.Left + Size.Width-DXr),
                         YPos(OfsY + Size.Top + DYt));
                  MoveTo(XPos(OfsX + Size.Left + DXr),
                         YPos(OfsY + Size.Top + DYt));
                  LineTo(XPos(OfsX + Size.Left + DXl),
                         YPos(OfsY + Size.Top + Size.Height-DYb));
                  QRPrinter.Canvas.Pen:=FPen;
                end;
              cmdist:=FCheckmarkDistance * 254/72;
              cTop := Size.Top + DYt + cmdist;
              cHeight := Size.Height - DYt - DYb - 2*cmdist;
              cLeft := Size.Left + DXl + cmdist;
              cWidth := Size.Width  - DXl - DXr - 2*cmdist;
              if Checked then
                begin
                  If CheckmarkType=cmCross Then
                    begin
                      MoveTo(XPos(OfsX + cLeft),
                             YPos(OfsY + cTop));
                      LineTo(XPos(OfsX + cLeft + cWidth),
                             YPos(OfsY + cTop + cHeight));

                      MoveTo(XPos(OfsX + cLeft),
                             YPos(OfsY + cTop + cHeight));
                      LineTo(XPos(OfsX + cLeft + cWidth),
                             YPos(OfsY + cTop));
                    end
                  Else
                  If CheckmarkType=cmHook Then
                    begin
                      MoveTo(XPos(OfsX + cLeft),
                             YPos(OfsY + cTop + cHeight/2));
                      LineTo(XPos(OfsX + cLeft + cWidth/3),
                             YPos(OfsY + cTop + cHeight));
                      LineTo(XPos(OfsX + cLeft + cWidth),
                             YPos(OfsY + cTop));
                    end
                  Else
                  If CheckmarkType=cmGlyph Then
                    begin
                      Dest.Top := QRPrinter.YPos(OfsY + cTop);
                      Dest.Left := QRPrinter.XPos(OfsX + cLeft);
                      Dest.Right := QRPrinter.XPos(OfsX + cWidth + cLeft);
                      Dest.Bottom := QRPrinter.YPos(OfsY + cHeight + cTop);
                      if GlyphChecked.Graphic is TBitmap then
                        PrintBitmap(QRPrinter.Canvas, Dest, TBitmap(GlyphChecked.Graphic))
                      else
                        with QRPrinter.Canvas do
                          StretchDraw(Dest, GlyphChecked.Graphic);
                    end;
                end
              Else {checked=false}
                begin
                  If CheckmarkType=cmGlyph then
                    begin
                      Dest.Top := QRPrinter.YPos(OfsY + cTop);
                      Dest.Left := QRPrinter.XPos(OfsX + cLeft);
                      Dest.Right := QRPrinter.XPos(OfsX + cWidth + cLeft);
                      Dest.Bottom := QRPrinter.YPos(OfsY + cHeight + cTop);
                      if GlyphUnchecked.Graphic is TBitmap then
                        PrintBitmap(QRPrinter.Canvas, Dest, TBitmap(GlyphUnchecked.Graphic))
                      else
                        with QRPrinter.Canvas do
                          StretchDraw(Dest, GlyphUnchecked.Graphic);
                    end;
                end;
            end;
        end;
    end;
    PrintFinished := True;
//    if ParentReport.FinalPass and Frame.AnyFrame
//    then inherited Print(OfsX, OfsY);
{
  if ParentReport.FinalPass and Frame.AnyFrame then
    TQRPFrame(Frame).PaintIt(ParentReport.QRPrinter.Canvas,
      Rect(ParentReport.QRPrinter.XPos(OfsX + Size.Left),
      ParentReport.QRPrinter.YPos(OfsY + Size.Top),
      ParentReport.QRPrinter.XPos(OfsX + Size.Left + Size.Width),
      ParentReport.QRPrinter.YPos(OfsY + Size.Top + Size.Height)),
      ParentReport.QRPrinter.XFactor,
      ParentReport.QRPrinter.YFactor);
}
end;

{TQRGrid}
constructor TQRGrid.Create(AOwner : TComponent);
var
  I, J : Integer;
  L    : TList;
begin
  inherited Create(AOwner);
  FCellColumns:=TList.Create;
  FColumnSizes:=TStringList.Create;
  FRowSizes:=TStringList.Create;
  Width:=65;
  Height:=65;
  FPen:=TPen.Create;
  FBrush:=TBrush.Create;
  FBrush.Color:=clWhite;
  FBrush.OnChange:=StyleChanged;
  FPen.OnChange:=StyleChanged;
  FRows:=3;
  FColumns:=3;
  For I:=1 to 3 Do
    begin
      L:=TList.Create;
      For J:=1 to 3 Do L.Add(TQRGridCell.Create);
      FCellColumns.Add(L);
    end;
  Transparent:=False;
  FFirstRowHeight:=0;
  FFirstColumnWidth:=0;
  FFirstRowColor:=clWhite;
  FFirstColumnColor:=clWhite;
  FFirstColumnJoined:=False;
  FFirstRowJoined:=False;
  FFirstCellColor:=clWhite;
  FDrawFirstRowLine:=True;
end;

function TQRGrid.GetCells(ACol, ARow: Integer): TQRGridCell;
begin
  try
    Result:=TQRGridCell(TList(FCellColumns[ACol])[ARow]);
  except
    Result:=NIL;
  end;
end;

procedure TQRGrid.SetCells(ACol, ARow: Integer; const Value: TQRGridCell);
begin
end;

procedure TQRGrid.SetColumnSizes(Value: TStrings);
var
  PAll, X: Integer;
begin
  FColumnSizes.Assign(Value);
  While FColumnSizes.Count>FColumns Do FColumnSizes.Delete(FColumnSizes.Count-1);
  PAll:=0;
  For X:=1 to FColumnSizes.Count Do
    begin
      try
        PAll:=PAll+StrToInt(FColumnSizes[X-1]);
      except
        FColumnSizes[X-1]:='20';
        PAll:=PAll+20;
      end;
    end;
  If FColumnSizes.Count=Columns Then
    begin
      If Frame.DrawLeft Then PAll:=PAll+Frame.Width;
      If Frame.DrawRight Then PAll:=PAll+Frame.Width;
      Width:=Pall;
    end;
  Invalidate;
end;

procedure TQRGrid.SetRowSizes(Value: TStrings);
var
  PAll, X: Integer;
begin
  FRowSizes.Assign(Value);
  While FRowSizes.Count>FRows Do FRowSizes.Delete(FRowSizes.Count-1);
  PAll:=0;
  For X:=1 to FRowSizes.Count Do
    begin
      try
        PAll:=PAll+StrToInt(FRowSizes[X-1]);
      except
        FRowSizes[X-1]:='20';
        PAll:=PAll+20;
      end;
    end;
  If FRowSizes.Count=Rows Then
    begin
      If Frame.DrawTop Then PAll:=PAll+Frame.Width;
      If Frame.DrawBottom Then PAll:=PAll+Frame.Width;
      Height:=Pall;
    end;
  Invalidate;
end;


destructor TQRGrid.Destroy;
var
  I, J : Integer;
  L    : TList;
begin
  FPen.Free;
  FBrush.Free;
  FColumnSizes.Free;
  FRowSizes.Free;

  For I:=1 to FCellColumns.Count Do
    begin
      L:=TList(FCellColumns[I-1]);
      For J:=L.Count DownTo 1 Do TQRGridCell(L[J-1]).Free;
      L.Free;
    end;
{
  repeat
    if FCellColumns.Count <> 0
    then begin
      L:=TList(FCellColumns[0]);
      repeat
        if L.Count <> 0
        then TQRGridCell(L[0]).Free;
      until L.Count=0;
      L.Free;
    end;
  until FCellColumns.Count = 0;
}
  FCellColumns.Free;
  inherited Destroy;
end;

procedure TQRGrid.SetColumns(Value: Word);
var
  L: TList;
  I: Integer;
begin
  FColumns:=Value;
  While FColumnSizes.Count>FColumns Do FColumnSizes.Delete(FColumnSizes.Count-1);
  While FCellColumns.Count>Value Do
    begin
      L:=TList(FCellColumns[FCellColumns.Count-1]);
      For I:=1 to L.Count Do TQRGridCell(L[I-1]).Free;
{
      repeat
        if L.Count<>0 then TQRGridCell(L[0]).Free;
      until L.Count=0;
}
      FCellColumns.Delete(FCellColumns.Count-1);
      L.Free;
    end;
  While FCellColumns.Count<Value Do
    begin
      L:=TList.Create;
      For I:=1 to Rows Do L.Add(TQRGridCell.Create);
      FCellColumns.Add(L);
    end;
  Invalidate;
end;

procedure TQRGrid.SetRows(Value: Word);
var
  I : Integer;
  L : TList;
begin
  FRows:=Value;
  While FRowSizes.Count>FRows Do FRowSizes.Delete(FRowSizes.Count-1);
  For I:=1 to FColumns Do
    begin
      L:=FCellColumns[I-1];
      While L.Count>Value Do
        begin
          TQRGridCell(L[L.Count-1]).Free;
          L.Delete(L.Count-1);
        end;
      While L.Count<Value Do L.Add(TQRGridCell.Create);
    end;
  Invalidate;
end;

procedure TQRGrid.StyleChanged(Sender : TObject);
begin
(*  If (Sender is TBrush) AND (Not (FFirstRowJoined or FFirstColumnJoined)) Then
    begin
      FFirstCellColor:=FBrush.Color;
      FFirstColumnColor:=FBrush.Color;
      FFirstRowColor:=FBrush.Color;
    end;*)
  Invalidate;
end;

procedure TQRGrid.SetBrush(Value : TBrush);
begin
  If (Not (FFirstRowJoined or FFirstColumnJoined)) AND
     (Value.Color<>FBrush.Color) Then
    begin
      If FFirstCellColor=FBrush.Color Then FFirstCellColor:=Value.Color;
      If FFirstColumnColor=FBrush.Color Then FFirstColumnColor:=Value.Color;
      If FFirstRowColor=FBrush.Color Then FFirstRowColor:=Value.Color;
    end;
  FBrush.Assign(Value);
  Invalidate;
end;

procedure TQRGrid.SetPen(Value : TPen);
begin
  FPen.Assign(value);
end;

procedure TQRGrid.SetDrawFirstRowLine(Value: Boolean);
begin
  FDrawFirstRowLine:=Value;
  Invalidate;
end;
{
procedure TQRGrid.SetFirstRowHeight(Value: Word);
begin
  FFirstRowHeight:=Value;
  Invalidate;
end;

procedure TQRGrid.SetFirstColumnWidth(Value: Word);
begin
  FFirstColumnWidth:=Value;
  Invalidate;
end;
}
procedure TQRGrid.SetFirstCellColor(Value: TColor);
begin
  FFirstCellColor:=Value;
  Invalidate;
end;

procedure TQRGrid.SetFirstRowColor(Value: TColor);
begin
  FFirstRowColor:=Value;
  Invalidate;
end;

procedure TQRGrid.SetFirstColumnColor(Value: TColor);
begin
  FFirstColumnColor:=Value;
  Invalidate;
end;

procedure TQRGrid.SetFirstColumnJoined(Value: Boolean);
begin
  FFirstColumnJoined:=Value;
  Invalidate;
end;

procedure TQRGrid.SetFirstRowJoined(Value: Boolean);
begin
  FFirstRowJoined:=Value;
  Invalidate;
end;

procedure TQRGrid.Paint;
var
  aRect: TRect;
  AllWidth, AllHeight,
  X, Y,
  RestHeight, RestWidth, P, DX, DY: Integer;
  BC: TColor;

  function GetColWidth(I: Integer): Integer;
  begin
    If I<FColumnSizes.Count Then
      Result:=muldiv(StrToInt(FColumnSizes[I]),zoom,100)
    Else
    If FColumnSizes.Count>0 Then
      Result:=Round(RestWidth/(Columns-FColumnSizes.Count))
    Else
      Result:=AllWidth DIV Columns;
  end;

  function GetRowHeight(I: Integer): Integer;
  begin
    If I<FRowSizes.Count Then
      Result:=//StrToInt(FRowSizes[I])
               muldiv(StrToInt(FRowSizes[I]),zoom,100)
    Else
    If FRowSizes.Count>0 Then
      Result:=Round(RestHeight/(Rows-FRowSizes.Count))
    Else
      Result:=AllHeight DIV Rows;
  end;

begin
  with Canvas do
    begin
//      Pen := FPen;
      Pen.Assign(FPen);
      Pen.Width := muldiv(Pen.Width,Zoom,100);
      if (FPen.Width > 0) and (Pen.Width<1) then Pen.Width := 1;
      Brush := FBrush;
      If not Transparent Then FillRect(ClientRect);

      AllWidth:=Width;
      If Frame.DrawLeft Then
        begin
          DX := round(Frame.Width * Screen.PixelsPerInch * Zoom / 7200);
          AllWidth:=AllWidth-DX;
        end
      Else
        DX:=0;
      If Frame.DrawRight Then AllWidth:=AllWidth-round(Frame.Width * Screen.PixelsPerInch * Zoom / 7200);

      AllHeight:=Height;
      If Frame.DrawTop Then
        begin
          DY := round(Frame.Width * Screen.PixelsPerInch * Zoom / 7200);
          AllHeight:=AllHeight-DX;
        end
      Else
        DY:=0;
      If Frame.DrawBottom Then AllHeight:=AllHeight-round(Frame.Width * Screen.PixelsPerInch * Zoom / 7200);

      RestWidth:=AllWidth;
      For X:=1 to FColumnSizes.Count Do RestWidth:=RestWidth-StrToInt(FColumnsizes[X-1]);
      RestHeight:=AllHeight;
      For X:=1 to FRowSizes.Count Do RestHeight:=RestHeight-StrToInt(FRowsizes[X-1]);

      If Not Transparent Then
        begin
          BC:=Brush.Color;
          If FirstRowColor<>BC Then
            begin
              Brush.Color:=FirstRowColor;
              FillRect(Rect(0,0,Width,GetRowHeight(0)+DY));
            end;
          If FirstColumnColor<>BC Then
            begin
              Brush.Color:=FirstColumnColor;
              FillRect(Rect(0,0,GetColWidth(0)+DX,Height));
            end;
          If FirstCellColor<>BC Then
            begin
              Brush.Color:=FirstCellColor;
              FillRect(Rect(0,0,GetColWidth(0)+DX,GetRowHeight(0)+DY));
            end;
        end;

      If Rows>1 Then
        begin
          P:=DY;
          For Y:=1 to Rows-1 Do
            begin
              P:=P+GetRowHeight(Y-1);
              If (Y<>1) OR DrawFirstRowLine Then
                begin
                  If FirstColumnJoined Then
                    MoveTo(GetColWidth(0)+DX*byte(Frame.DrawLeft),P)
                  Else
                    MoveTo(DX,P);
                  LineTo(Width-DX*byte(Frame.DrawRight),P);
                end;
            end;
        end;
      If Columns>1 Then
        begin
          P:=DX;
          For X:=1 to Columns-1 Do
            begin
              P:=P+GetColWidth(X-1);
              If FirstRowJoined Then
                MoveTo(P,GetRowHeight(0)+DY)
              Else
                MoveTo(P,DY);
              LineTo(P,Height-DY*byte(Frame.DrawBottom));
            end;
        end;
    end;
//RSV  PaintCorners;
  aRect := GetClientRect;
//  dec(aRect.Bottom);
//  dec(aRect.Right);
  TQRFixFrame(Frame).PaintIt(Canvas, aRect, Screen.PixelsPerInch * Zoom/25400, Screen.PixelsPerInch * Zoom/25400);
  PaintCorners;
end;

procedure TQRGrid.Print(OfsX, OfsY: Extended);
var
  FrameWidth,
  FrameHeight,
  PX, PY, P,
  RestWidth,
  RestHeight,
  AllWidth,
  AllHeight  : Extended;
  X, Y       : Integer;
  BC         : TColor;
  C          : TQRGridCell;

  function GetColWidth(I: Integer): Extended;
  begin
    If I<FColumnSizes.Count Then
//RSV
     Result:=StrToFloat(FColumnSizes[I]) / Screen.PixelsPerInch * 254
//      Result:=StrToFloat(FColumnSizes[I])/QRPrinter.XFactor
    Else
    If FColumnSizes.Count>0 Then
      Result:=RestWidth/(Columns-FColumnSizes.Count)
    Else
      Result:=AllWidth/Columns;
  end;

  function GetRowHeight(I: Integer): Extended;
  begin
    If I<FRowSizes.Count Then
//
      Result:=StrToFloat(FRowSizes[I]) / Screen.PixelsPerInch * 254
//      Result:=StrToFloat(FRowSizes[I]) / QRPrinter.XFactor
    Else
    If FRowSizes.Count>0 Then
      Result:=RestHeight/(Rows-FRowSizes.Count)
    Else
      Result:=AllHeight/Rows;
  end;

  function min_w(a,b: Extended): Extended;
  begin
    Result := a;
    if b < a then Result := b;
  end;

const
  PenStyles: array[TPenStyle] of Word =
    (PS_SOLID, PS_DASH, PS_DOT, PS_DASHDOT, PS_DASHDOTDOT, PS_NULL,
     PS_INSIDEFRAME{, PS_NULL, PS_NULL});

var
  AX, AY: Real;
  ACaption: String;
  PW,
  iAX, iAY: Integer;
  ARect: TRect;
  HA: THorizAlignment;
  VA: TVertAlignment;

//  OldPen: HPen;
  aPen,
  oPen: TPen;
//  LogPen: TLogPen;
  LogBrush: TLogBrush;
  Mask: Integer;

  Cnt: Integer;
  Points: array of TPoint;
  Values: array of DWord;

begin
  PrintFinished := False;
  If (Columns=0) or (Rows=0) Then Exit;
  FrameWidth  := Frame.Width*72/254;
  FrameHeight := FrameWidth;

  if ParentReport.FinalPass and Enabled then
    begin
      QRPrinter.Canvas.Brush := Brush;
      QRPrinter.Canvas.Pen.Assign(FPen);
      QRPrinter.Canvas.Pen.Width :=
           (QRPrinter.XSize(Pen.Width*254/Screen.PixelsPerInch)+
            QRPrinter.YSize(Pen.Width*254/Screen.PixelsPerInch)) shr 1;
      if (FPen.Width > 0) and (QRPrinter.Canvas.Pen.Width < 1) then QRPrinter.Canvas.Pen.Width := 1;

      with QRPrinter, Canvas do
        begin

          {---- draw grid ---}
          If not Transparent Then
            FillRect(rect(XPos(OfsX + Size.Left+ FrameWidth*byte(Frame.DrawLeft)),
                          YPos(OfsY + Size.Top + FrameHeight*byte(Frame.DrawTop)),
                          XPos(OfsX+Size.Left + Size.Width-FrameWidth*byte(Frame.DrawRight)),
                          YPos(OfsY + Size.Top + Size.Height - FrameHeight*byte(Frame.DrawBottom))
                          )
                     );
          AllWidth:=Size.Width - FrameWidth*(byte(Frame.DrawLeft)+byte(Frame.DrawRight));

          AllHeight:=Size.Height - FrameHeight*(byte(Frame.DrawTop)+byte(Frame.DrawBottom));

          RestWidth:=AllWidth;
          For X:=1 to FColumnSizes.Count Do RestWidth:=RestWidth-GetColWidth(X-1);
          RestHeight:=AllHeight;
          For X:=1 to FRowSizes.Count Do RestHeight:=RestHeight-GetRowHeight(X-1);

          If Not Transparent Then
            begin
              BC:=Brush.Color;
              If FirstRowColor<>BC Then
                begin
                  Brush.Color:=FirstRowColor;
                  FillRect(rect(XPos(OfsX + Size.Left),
                                YPos(OfsY + Size.Top + FrameHeight*Byte(Frame.DrawTop)),
                                XPos(OfsX + Size.Left + Size.Width),
                                YPos(OfsY + Size.Top + GetRowHeight(0) + FrameHeight*Byte(Frame.DrawTop)
                                                                       {- FrameHeight*Byte(Frame.DrawBottom)})));
                end;
              If FirstColumnColor<>BC Then
                begin
                  Brush.Color:=FirstColumnColor;
                  FillRect(rect(XPos(OfsX + Size.Left),
                                YPos(OfsY + Size.Top + FrameHeight*Byte(Frame.DrawTop)),
                                XPos(OfsX + Size.Left + GetColWidth(0) + FrameWidth*Byte(Frame.DrawLeft)),
                                YPos(OfsY + Size.Top + Size.Height + FrameHeight*Byte(Frame.DrawTop)
                                                                   {- FrameHeight*Byte(Frame.DrawBottom)})));
                end;
              If FirstCellColor<>BC Then
                begin
                  Brush.Color:=FirstCellColor;
                  FillRect(rect(XPos(OfsX + Size.Left),
                                YPos(OfsY + Size.Top + FrameHeight*Byte(Frame.DrawTop) ),
                                XPos(OfsX + Size.Left + GetColWidth(0) + FrameWidth*Byte(Frame.DrawLeft)),
                                YPos(OfsY + Size.Top + GetRowHeight(0) + FrameHeight*Byte(Frame.DrawTop)
                                                                       {- FrameHeight*Byte(Frame.DrawBottom)})));
                end;
            end;

          {--- draw cell captions ---}

          PX:=FrameWidth*Byte(Frame.DrawLeft);
          For X:=0 to Columns-1 Do
            begin
              PY:=FrameHeight*Byte(Frame.DrawTop);
              For Y:=0 to Rows-1 Do
                begin
                  If Cells[X,Y].Caption<>'' Then
                    begin
                      C:=Cells[X,Y];
                      aCaption:=C.Caption;

                      AX := OfsX + Size.Left + PX;
                      AY := OfsY + Size.Top + PY;

                      If Not C.ParentFont Then Canvas.Font.Assign(C.Font)
                      Else Canvas.Font.Assign(Self.Font);

                      If C.HorizAlignment=haNone Then
                        HA:=CellHorizAlignment
                      Else
                        HA:=C.HorizAlignment;

                      aRect := Rect(XPos(AX),
                                    YPos(AY),
                                    XPos(AX  + GetColWidth(X)),
                                    YPos(AY  + GetRowHeight(Y)));
                      InflateRect(aRect,-Canvas.Pen.Width,-Canvas.Pen.Width);
                      iAX := aRect.Left;
                      case HA of
                        haNone,
                        haLeft  : SetTextAlign(Canvas.Handle, TA_Left + TA_Top + TA_NoUpdateCP);
                        haRight : begin
                                    SetTextAlign(Canvas.Handle, TA_Right + TA_Top + TA_NoUpdateCP);
                                    iAX := aRect.Right;
                                    if fsItalic in Font.Style then
                                      iAX := iAX - ParentReport.TextWidth(Font, ' ');
                                  end;
                        haCenter: begin
                                    SetTextAlign(Canvas.Handle, TA_Center + TA_Top + TA_NoUpdateCP);
                                    iAX := iAX + (aRect.Right-aRect.Left) div 2;
                                  end;
                      end;

                      If C.VertAlignment=vaNone Then
                        VA:=CellVertAlignment
                      Else
                        VA:=C.VertAlignment;
                      iAY := aRect.Top;
                      Case VA of
                        vaNone,
                        vaTop    : iAY := iAY+QRPrinter.Canvas.Pen.Width;
                        vaCenter : iAY := aRect.Top+(aRect.Bottom-aRect.Top-ParentReport.TextHeight(Font,'Yy')) div 2;
                        vaBottom : iAY := aRect.Bottom-ParentReport.TextHeight(Font,'Yy');
                      end;

                      if (X = 0) and (Y=0)
                      then Canvas.Brush.Color := FirstCellColor
                      else if (X = 0) then Canvas.Brush.Color := FirstColumnColor
                        else if (Y = 0) then Canvas.Brush.Color := FirstRowColor
                          else Canvas.Brush.Color := Self.Brush.Color;
                      Canvas.FillRect(aRect);
                      ExtTextOut(Canvas.Handle, iAX, iAY,
                                 ETO_CLIPPED, @aRect, PChar(aCaption), length(aCaption), nil);
                    end;
                  PY:=PY+GetRowHeight(Y);
                end;
              PX:=PX+GetColWidth(X);
            end;


          {--- Draw Line ---}

//     PW := QRPrinter.Canvas.Pen.Width;


     PW := (round(QRPrinter.XFactor * Frame.Width * 254/72)+
            round(QRPrinter.YFactor * Frame.Width * 254/72)) shr 1
     ;

//     if GetVersion < $80000000 then
     if (Rows > 1) or (Columns > 1) then
     begin
//>RSV NEW

       oPen := QRPrinter.Canvas.Pen;
       if GetVersion < $80000000 then begin
         LogBrush.lbStyle := BS_SOLID;
         LogBrush.lbColor := ColorToRGB(Self.Pen.Color);
         LogBrush.lbHatch := 0;
         Mask := PenStyles[Self.Pen.Style]
         or PS_GEOMETRIC or PS_ENDCAP_FLAT or PS_JOIN_MITER;
         aPen := TPen.Create;
       end else aPen := Self.Pen;
       try

         SetLength(Points,2*(Rows+Columns-2));
         SetLength(Values,(Rows+Columns-2));
         Cnt := 0;
         if GetVersion < $80000000 then begin
           aPen.Handle := ExtCreatePen(Mask, PW, LogBrush, 0, nil);
         end;
         QRPrinter.Canvas.Pen := aPen;
         If Rows>1 Then begin
           P := FrameHeight*Byte(Frame.DrawTop);
           For Y:=1 to Rows-1 Do begin
             P:=P+GetRowHeight(Y-1);
             If (Y<>1) OR DrawFirstRowLine Then begin
(*
               If FirstColumnJoined Then
                 MoveTo(XPos(OfsX + Size.Left + GetColWidth(0)){+PW shr 1}
                       +(PW+(PW mod 2)) shr 1,
                        YPos(OfsY + Size.Top + P))
               Else
                 MoveTo(XPos(OfsX + Size.Left)
                       +(PW+(PW mod 2)) shr 1,
                        YPos(OfsY + Size.Top + P));
               LineTo(XPos(OfsX + Size.Left + Size.Width
                      )-(PW+(PW mod 2)) shr 1,
                      YPos(OfsY + Size.Top + P));
*)

               If FirstColumnJoined Then
                 Points[Cnt] :=  Point(XPos(OfsX + Size.Left + GetColWidth(0))
                 {+PW shr 1}+(PW+(PW mod 2)) shr 1,
                        YPos(OfsY + Size.Top + P))
               Else
                 Points[Cnt] :=  Point(XPos(OfsX + Size.Left)+(PW+(PW mod 2)) shr 1,
                        YPos(OfsY + Size.Top + P));
               Points[Cnt+1] :=  Point(XPos(OfsX + Size.Left + Size.Width
                      )-(PW+(PW mod 2)) shr 1,
                      YPos(OfsY + Size.Top + P));
               Values[Cnt shr 1] := 2;
               Inc(Cnt,2);

             end;
           end;
         end;
         If Columns>1 Then begin
           P:= FrameWidth*Byte(Frame.DrawLeft);
           For X:=1 to Columns-1 Do begin
             P:=P+GetColWidth(X-1);
(*
             If FirstRowJoined Then
                MoveTo(XPos(OfsX + Size.Left + P),
                       YPos(OfsY + Size.Top + GetRowHeight(0)){+PW shr 1}
                       +(PW+(PW mod 2)) shr 1)
              Else
                MoveTo(XPos(OfsX + Size.Left + P),
                       YPos(OfsY + Size.Top)
                       +(PW+(PW mod 2)) shr 1);
              LineTo(XPos(OfsX + Size.Left + P),
                     YPos(OfsY + Size.Top + Size.Height
                      )-(PW+(PW mod 2)) shr 1);
*)

             If FirstRowJoined Then
                Points[Cnt] :=  Point(XPos(OfsX + Size.Left + P),
                       YPos(OfsY + Size.Top + GetRowHeight(0)){+PW shr 1}
                       +(PW+(PW mod 2)) shr 1)
              Else
                Points[Cnt] :=  Point(XPos(OfsX + Size.Left + P),
                       YPos(OfsY + Size.Top)+(PW+(PW mod 2)) shr 1);
              Points[Cnt+1] :=  Point(XPos(OfsX + Size.Left + P),
                     YPos(OfsY + Size.Top + Size.Height
                      )-(PW+(PW mod 2)) shr 1);
               Values[Cnt shr 1] := 2;
               Inc(Cnt,2);

           end;
         end;

         if Cnt > 0
         then PolyPolyline(QRPrinter.Canvas.Handle,
           Pointer(Points)^,
           Pointer(Values)^,
           Cnt shr 1);
         SetLength(Points,0);
         SetLength(Values,0);

       finally
         QRPrinter.Canvas.Pen := oPen;
         if GetVersion < $80000000 then begin
           DeleteObject(aPen.Handle);
           aPen.Free;
         end;
       end
     end
(*
     else try
          QRPrinter.Canvas.Pen.Color := Self.Pen.Color;
          If Rows>1 Then
            begin
              P := FrameHeight*Byte(Frame.DrawTop);
              For Y:=1 to Rows-1 Do
                begin
                  P:=P+GetRowHeight(Y-1);
                  If (Y<>1) OR DrawFirstRowLine Then
                    begin
                      If FirstColumnJoined Then
                        MoveTo(XPos(OfsX + Size.Left + GetColWidth(0))+(PW+PW mod 2) shr 1,
                               YPos(OfsY + Size.Top + P))
                      Else
                        MoveTo(XPos(OfsX + Size.Left){+PW shr 1},
                               YPos(OfsY + Size.Top + P));
                      LineTo(XPos(OfsX + Size.Left + Size.Width
                             )-(PW+PW mod 2) shr 1,
                             YPos(OfsY + Size.Top + P));
                    end;
                end;
            end;
          If Columns>1 Then
            begin
              P:= FrameWidth*Byte(Frame.DrawLeft);
              For X:=1 to Columns-1 Do
                begin
                  P:=P+GetColWidth(X-1);
                  If FirstRowJoined Then
                    MoveTo(XPos(OfsX + Size.Left + P),
                           YPos(OfsY + Size.Top + GetRowHeight(0))+(PW+PW mod 2) shr 1)
                  Else
                    MoveTo(XPos(OfsX + Size.Left + P),
                           YPos(OfsY + Size.Top){ + PW shr 1});
                  LineTo(XPos(OfsX + Size.Left + P),
                         YPos(OfsY + Size.Top + Size.Height
                           )-(PW+PW mod 2) shr 1);
                end;
            end;
      finally
      end*);

        end;
    end;
    PrintFinished := True;
    inherited Print(OfsX, OfsY);
end;


constructor TQRGridCell.Create;
begin
  Font:=TFont.Create;
  ParentFont:=True;
end;

destructor TQRGridCell.Destroy;
begin
  Font.Free;
  Inherited Destroy;
end;

{ TQRIndentSpace }

constructor TQRIndentSpace.Create(AOwner: TQRPrintable);
begin
  inherited Create;
  FOwner := AOwner;
  FRect := Rect(1,0,1,0);
end;

destructor TQRIndentSpace.Destroy;
begin
  FOwner := nil;
  inherited Destroy;
end;

function TQRIndentSpace.GetExtValue(const Index: Integer): Extended;
begin
  case Index of
  0: Result := FRect.Left;
  1: Result := FRect.Right;
  2: Result := FRect.Top;
  3: Result := FRect.Bottom;
  else Result := 0;
  end;
  if Result <> 0
  then if Assigned(FOwner) and Assigned(FOwner.ParentReport) and Assigned(FOwner.ParentReport.QRPrinter)
    then //if index in [0..1]
//      then Result := Result*FOwner.ParentReport.QRPrinter.XFactor*254/72
//      else Result := Result*FOwner.ParentReport.QRPrinter.YFactor*254/72
// TSK       Result := Result*254/72
      Result := Result * 254 / ( 72 * 2 )
//      Result := Result/Screen.PixelsPerInch*72
// TSK   else Result := Result*Screen.PixelsPerInch/72;
    else Result := Result * Screen.PixelsPerInch / ( 72 * 2 );

  if (Result < 1) and (GetValue(Index) > 0) then Result := 1;
{
  if Result <> 0
  then if Assigned(FOwner) and Assigned(FOwner.ParentReport) and Assigned(FOwner.ParentReport.QRPrinter)
    then Result := Result* 254 / 72
    else Result := Result* Screen.PixelsPerInch / 254;
}
end;

function TQRIndentSpace.GetValue(const Index: Integer): Integer;
begin
  case Index of
  0: Result := FRect.Left;
  1: Result := FRect.Right;
  2: Result := FRect.Top;
  3: Result := FRect.Bottom;
  else Result := 0;
  end;
end;

procedure TQRIndentSpace.SetValue(const Index: Integer; Value: Integer);
begin
  if GetValue(Index) <> Value then begin
    if Value < 0 then Value := 0;
    case Index of
    0: FRect.Left   := Value;
    1: FRect.Right  := Value;
    2: FRect.Top    := Value;
    3: FRect.Bottom := Value;
    end;
    Invalidate;
  end;
end;

procedure TQRIndentSpace.Invalidate;
begin
  if Assigned(FOwner) then begin
    if FOwner is TQRCustomLabel then TQRCustomLabel(FOwner).FormatLines;
    FOwner.Invalidate;
  end;
end;

function TQRIndentSpace.IsDefault: Boolean;
begin
  Result := EqualRect(FRect, Rect(1,0,1,0));
end;

end.
