{ :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
  :: Bugs Free QuickReport 3.0 for Delphi and C++Builder     ::
  ::                                                         ::
  :: rGridRpt.PAS - DBGrid Printing                          ::
  ::                                                         ::
  :: Copyright (c) 2000 Bugs Free QR Team                    ::
  ::                                                         ::
  :: web: http://bfqr.nm.ru                                  ::
  ::                                                         ::
  ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: }

unit rGridRpt;

interface

uses
  Controls, SysUtils,  DBGrids, DB, QuickRpt, QRCtrls,
  Graphics, QRPrntr, Printers, Windows, Forms,
  Classes;

type
  TrGridLines = (glVertical, glHorizontal, glBoth, glNone);

  TQRDBGridReport = class;

  TrQRBand = class(TPersistent)
  private
    FParent:     TQRDBGridReport;
    CustomControl : TControl;
    FValue:      TStrings;
    FAlignment:  TAlignment;
    FFont:       TFont;
    FColor:      TColor;
    FParentFont: Boolean;
    FFrame:      TQRFrame;
    FValueAsExpr: Boolean;
    procedure SetValue(Value: TStrings);
    procedure SetFont(Value: TFont);
    procedure SetColor(Value: TColor);
    procedure SetParentFont(Value: Boolean);
    procedure FontChanged(Sender : TObject);
    function StoredFont : Boolean;
    procedure SetFrame(Value : TQRFrame);
  public
    constructor Create(AOwner: TQRDBGridReport); virtual;
    destructor Destroy; override;
  published
    property Value: TStrings read FValue write SetValue;
    property Alignment: TAlignment read FAlignment write FAlignment default taCenter;
    property Font: TFont read FFont write SetFont stored StoredFont;
    property Color: TColor read FColor write SetColor default clWhite;
    property ParentFont: Boolean read FParentFont write SetParentFont default True;
    property Frame : TQRFrame read FFrame write SetFrame;
    property ValueAsExpr: Boolean read FValueAsExpr write FValueAsExpr default False;
  end;

  TQRDBGridReport = class(TComponent)
  private
    FPage:                TQRPage;
    FDBGrid:              TDBGrid;
    FQuickRep:            TQuickRep;
    FAutoStretch:         Boolean;
    FPrintGridLines:      TrGridLines;
    FSelectedRecordsOnly: Boolean;
    FUseGridSettings:     Boolean;
    FReportTitle:          String;
    FFont:                TFont;
    // Bands
    FTitle:               TrQRBand;
    FPageHeader:          TrQRBand;
    FColumnHeader:        TrQRBand;
    FDetail:              TrQRBand;
    FPageFooter:          TrQRBand;
    FSummary:             TrQRBand;
    // Events
    FBeforePrint:         TNotifyEvent;
    FBookMarkStr:         String;
    procedure SetFont(const Value: TFont);
    procedure SetValue(index : integer; Value : TrQRBand);
    function StoredFont : Boolean;
  protected
    procedure DetailBandBeforePrint(Sender: TQRCustomBand;  var PrintBand: Boolean); virtual;
    procedure PrintOrPreviewReport(Print: Boolean);
    procedure OnAfterPrint(Sender: TObject);
    procedure OnBeforePrint(Sender: TCustomQuickRep;
      var PrintReport: Boolean);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function CreateReportFromGrid(Dialog : Boolean): Boolean;
    procedure PrintReport;
    procedure PreviewReport;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    property QuickRep: TQuickRep read FQuickRep;
  published
    property Page: TQRPage read FPage;
    property BeforePrint: TNotifyEvent read FBeforePrint write FBeforePrint;
    property AutoStretch: Boolean read FAutoStretch write FAutoStretch default False;
    property PrintGridLines: TrGridLines read FPrintGridLines write FPrintGridLines default glBoth;
    property DBGrid: TDBGrid read FDBGrid write FDBGrid;
    property SelectedRecordsOnly: Boolean read FSelectedRecordsOnly write FSelectedRecordsOnly default False;
    property UseGridSettings: Boolean read FUseGridSettings write FUseGridSettings default True;
    property ReportTitle: String read FReportTitle write FReportTitle;
    property Font: TFont read FFont write SetFont stored StoredFont;
    // Bands
    property Title: TrQRBand index 0 read FTitle write SetValue;
    property PageHeader: TrQRBand index 1 read FPageHeader write SetValue;
    property ColumnHeader: TrQRBand index 2 read FColumnHeader write SetValue;
    property Detail: TrQRBand index 3 read FDetail write SetValue ;
    property PageFooter: TrQRBand index 4 read FPageFooter write SetValue;
    property Summary: TrQRBand index 5 read FSummary write SetValue;
  end;

  procedure AssignFrame(Distinct, Source : TQRFrame);

implementation
{
uses
  QRDlg;
}
//Tima  rGridDlg;

function GetPixelsPerInch: Integer;
var DC: HDC;
begin
  DC := GetDC(0);
  Result := GetDeviceCaps(DC, LOGPIXELSY);
  ReleaseDC(0, DC);
end;

function MM(Pixel: Integer): Double;
begin
  Result := Pixel / (GetPixelsPerInch / 25.4);
end;

procedure AssignFrame(Distinct, Source : TQRFrame);
begin
  with Distinct do  begin
    DrawTop    := Source.DrawTop;
    DrawLeft   := Source.DrawLeft;
    DrawBottom := Source.DrawBottom;
    DrawRight  := Source.DrawRight;
    Color      := Source.Color;
    Width      := Source.Width;
  end;
end;

{TQRDBGridReport}

function TQRDBGridReport.StoredFont : Boolean;
begin
  Result := not FUseGridSettings;
end;

procedure TQRDBGridReport.SetValue(Index : Integer; Value : TrQRBand);
begin
  case Index of
    0 : FTitle.Assign(Value);
    1 : FPageHeader.Assign(Value);
    2 : ColumnHeader.Assign(Value);
    3 : Detail.Assign(Value);
    4 : PageFooter.Assign(Value);
    5 : Summary.Assign(Value);
  end;
end;

procedure TQRDBGridReport.DetailBandBeforePrint(Sender: TQRCustomBand; var PrintBand: Boolean);
begin
  PrintBand := FDBGrid.SelectedRows.CurrentRowSelected;
end;

procedure TQRDBGridReport.PrintOrPreviewReport(Print: Boolean);
begin
  if CreateReportFromGrid(True) then try
    QuickRep.DataSet.DisableControls;
    if Print then QuickRep.Print else QuickRep.PreviewModal;
  finally
    QuickRep.DataSet.EnableControls;
  end;
end;

procedure TQRDBGridReport.PreviewReport;
begin
  PrintOrPreviewReport(False);
end;

procedure TQRDBGridReport.PrintReport;
begin
  PrintOrPreviewReport(True);
end;

constructor TQRDBGridReport.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDBGrid              := nil;
  FQuickRep            := TQuickRep.Create(nil);
  FQuickRep.AfterPrint := OnAfterPrint;
  FQuickRep.AfterPreview := OnAfterPrint;
  FQuickRep.BeforePrint  := OnBeforePrint;
  FPage                := TQRPage.Create(QuickRep);
  FSelectedRecordsOnly := False;
  FPrintGridLines      := glBoth;
  FUseGridSettings     := True;
  FFont                := TFont.Create;
  FFont.Name           := 'Arial';
  FFont.Size           := 10;
  // Bands
  FTitle               := TrQRBand.Create(Self);
  FPageHeader          := TrQRBand.Create(Self);
  FColumnHeader        := TrQRBand.Create(Self);
  FDetail              := TrQRBand.Create(Self);
  FPageFooter          := TrQRBand.Create(Self);
  FSummary             := TrQRBand.Create(Self);
end;

destructor TQRDBGridReport.Destroy;
begin
  // Bands
  FTitle.Free;
  FPageHeader.Free;
  FColumnHeader.Free;
  FDetail.Free;
  FPageFooter.Free;
  FSummary.Free;

  FPage.Free;
  FQuickRep.Free;
  FTitle := nil;
  FPageHeader := nil;
  FColumnHeader := nil;
  FDetail := nil;
  FPageFooter := nil;
  FSummary := nil;
  FPage := nil;
  FQuickRep := nil;
  inherited Destroy;
end;

//***
function TQRDBGridReport.CreateReportFromGrid(Dialog : Boolean): Boolean;
var
  PageHeaderBand, DetailBand, ColumnHeaderBand,
  PageFooterBand, SummaryBand, TitleBand : TQRBand;
  ColumnLabel, CustomLabel:  TQRLabel;
  CustomExpr : TQRExpr;
  DetailLabel : TQRDBText;
  x : Integer;
  FDataSet: TDataset;
  CurPlacement : Double;
//Tima  SetupDlg : TDlgGridRptSetup;

 function GetPaperSize(PS: Short): TQRPaperSize;
 var
   I: TQRPaperSize;
 begin
   Result := Default;
   for I := Default to Custom do
     if PS=cQRPaperTranslate[I] then begin
       Result := I;
       Exit;
     end;
 end;

begin
  Result := False;
{Tima
  if Dialog then begin
    SetupDlg := TDlgGridRptSetup.Create(nil);
    SetupDlg.FGridReport := Self;
    try
      if SetupDlg.ShowModal <> mrOK then Exit;
    finally
      SetupDlg.Free;
    end;
  end;
}
{
  if Dialog then begin
    with TQRColumnPageSetupDialog.Create(Application) do
    try
      Options       := Options + [poDisablePrinter];
      Measurements  := pmMillimeters;
      Margin.Left   := Page.LeftMargin;
      Margin.Top    := Page.TopMargin;
      Margin.Right  := Page.RightMargin;
      Margin.Bottom := Page.BottomMargin;
      dmPaperSize   := ord(Page.PaperSize);
      Orientation   := Page.Orientation;
      Columns       := Page.Columns;
      ColumnSpace   := Page.ColumnSpace;
     if Execute then begin
        Self.Page.Units        := TQRUnit(0);//MM;
        Self.Page.Length       := PaperSize.Y;
        Self.Page.Width        := PaperSize.X;
        Self.Page.PaperSize    := GetPaperSize(dmPaperSize);
        Self.Page.LeftMargin   := Margin.Left;
        Self.Page.TopMargin    := Margin.Top;
        Self.Page.RightMargin  := Margin.Right;
        Self.Page.BottomMargin := Margin.Bottom;
        Self.Page.Orientation  := Orientation;
        Self.Page.Columns      := Columns;
        Self.Page.ColumnSpace  := ColumnSpace;
      end
      else Exit;
    finally
      Free;
    end;
  end;
}
  if (QuickRep = nil) or (DBGrid = nil) or
     (DBGrid.DataSource = nil) or
     (DBGrid.DataSource.Dataset = nil) then Exit;

  with QuickRep, QuickRep.Page do begin
    Units := TQRUnit(0);//MM;
    Length := Self.Page.Length;
    Width := Self.Page.Width;
    PaperSize := Self.Page.PaperSize;

    BottomMargin := Self.Page.BottomMargin;
    Columns := Self.Page.Columns;
    ColumnSpace := Self.Page.ColumnSpace;
    LeftMargin := Self.Page.LeftMargin;
    Orientation := Self.Page.Orientation;
    RightMargin := Self.Page.RightMargin;
    Ruler := Self.Page.Ruler;
    TopMargin := Self.Page.TopMargin;
  end;

  FDataSet         := DBGrid.DataSource.DataSet;
  QuickRep.Dataset := FDataSet;

  //  Page Header Band

  CurPlacement := 0;
  if Trim(FPageHeader.Value.Text) <> '' then begin
    PageHeaderBand          := TQRBand.Create(QuickRep);
    PageHeaderBand.Parent   := QuickRep;
    PageHeaderBand.BandType := rbPageHeader;
    PageHeaderBand.Color    := FPageHeader.Color;
    AssignFrame(PageHeaderBand.Frame, FPageHeader.Frame);
    PageHeaderBand.Font.Assign(FPageHeader.Font);

    for x := 0 to FPageHeader.Value.Count - 1 do
      if Trim(FPageHeader.Value[x]) <> '' then begin
        if FPageHeader.ValueAsExpr then begin
          CustomExpr := TQRExpr.Create(QuickRep);
          with CustomExpr do begin
            Parent       := PageHeaderBand;
            Transparent  := True;
            WordWrap     := False;
            Expression   := FPageHeader.Value[x];
            Alignment    := FPageHeader.Alignment;
            AlignToBand  := True;
            Size.Top     := CurPlacement;
            Size.Height  := MM(-FPageHeader.Font.Height) + 1;
            CurPlacement := CurPlacement + Size.Height;
          end;
        end
        else begin
          CustomLabel := TQRLabel.Create(QuickRep);
          with CustomLabel do begin
            Parent       := PageHeaderBand;
            Transparent  := True;
            WordWrap     := False;
            Caption      := FPageHeader.Value[x];
            Alignment    := FPageHeader.Alignment;
            AlignToBand  := True;
            Size.Top     := CurPlacement;
            Size.Height  := MM(-FPageHeader.Font.Height) + 1;
            CurPlacement := CurPlacement + Size.Height;
          end;
        end;
      end;
    PageHeaderBand.Size.Height := CurPlacement;
  end;

  // TitleBand

  TitleBand          := TQRBand.Create(QuickRep);
  TitleBand.Parent   := QuickRep;
  TitleBand.BandType := rbTitle;
  TitleBand.Color    := FTitle.Color;
  AssignFrame(TitleBand.Frame, FTitle.Frame);
  TitleBand.Font.Assign(FTitle.Font);

  if ReportTitle <> '' then begin
    CustomLabel := TQRLabel.Create(QuickRep);
    with CustomLabel do begin
      Parent       := TitleBand;
      Transparent  := True;
      WordWrap     := False;
      Caption      := ReportTitle;
      Size.Top     := 0;
      Size.Height  := MM(-FTitle.Font.Height) + 1;
      Alignment    := FTitle.Alignment;
      AlignToBand  := True;
      CurPlacement := Size.Height;
    end;
  end;

  if Trim(FTitle.Value.Text) <> '' then begin
    for x := 0 to FTitle.Value.Count - 1 do
      if Trim(FTitle.Value[x]) <> '' then begin
        if FTitle.ValueAsExpr then begin
          CustomExpr := TQRExpr.Create(QuickRep);
          with CustomExpr do begin
            Parent       := TitleBand;
            Transparent  := True;
            WordWrap     := False;
            Expression   := FTitle.Value[x];
            Size.Top     := CurPlacement;
            Size.Height  := MM(-FTitle.Font.Height) + 1;
            Alignment    := FTitle.Alignment;
            AlignToBand  := True;
            CurPlacement := CurPlacement + Size.Height;
          end;
        end
        else begin
          CustomLabel := TQRLabel.Create(QuickRep);
          with CustomLabel do begin
            Parent       := TitleBand;
            Transparent  := True;
            WordWrap     := False;
            Caption      := FTitle.Value[x];
            Size.Top     := CurPlacement;
            Size.Height  := MM(-FTitle.Font.Height) + 1;
            Alignment    := FTitle.Alignment;
            AlignToBand  := True;
            CurPlacement := CurPlacement + Size.Height;
          end;
        end
      end;
    TitleBand.Size.Height := CurPlacement;
  end;

  // Column Header Band

  ColumnHeaderBand          := TQRBand.Create(QuickRep);
  ColumnHeaderBand.Parent   := QuickRep;
  ColumnHeaderBand.BandType := rbColumnHeader;
  ColumnHeaderBand.Size.Height := MM(-FColumnHeader.Font.Height) + 1.5;
  ColumnHeaderBand.Color    := FColumnHeader.Color;
  AssignFrame(ColumnHeaderBand.Frame, FColumnHeader.Frame);
  ColumnHeaderBand.Font.Assign(FColumnHeader.Font);

  // Debail Band

  DetailBand             := TQRBand.Create(QuickRep);
  DetailBand.Parent      := QuickRep;
  DetailBand.BandType    := rbDetail;
  DetailBand.Size.Height := MM(-FDetail.Font.Height) + 1;
  DetailBand.Color       := FDetail.Color;
  AssignFrame(DetailBand.Frame, FDetail.Frame);
  DetailBand.Font.Assign(FDetail.Font);

  // Selected records Only
  if FSelectedRecordsOnly then begin
    FDBGrid.SelectedRows.Refresh;
    DetailBand.BeforePrint := DetailBandBeforePrint;
  end;

  CurPlacement := 0;
  for x := 0 to FDBGrid.Columns.Count - 1 do begin
    if CurPlacement + MM(FDBGrid.Columns[x].Width) > ColumnHeaderBand.Width
    then Break;
    ColumnLabel             := TQRLabel.Create(QuickRep);
    ColumnLabel.Parent      := ColumnHeaderBand;
    ColumnLabel.Caption     := FDBGrid.Columns[x].Title.Caption;
    ColumnLabel.Alignment   := FDBGrid.Columns[x].Title.Alignment;
    if UseGridSettings then ColumnLabel.Font.Assign(FDBGrid.Columns[x].Title.Font);
    ColumnLabel.AutoSize    := False;
    ColumnLabel.Size.Left   := CurPlacement;
    ColumnLabel.Size.Height := ColumnHeaderBand.Size.Height;
    ColumnLabel.Size.Top    := 0;
    ColumnLabel.Size.Width  := MM(FDBGrid.Columns[x].Width);
    ColumnLabel.Transparent := True;
    if FPrintGridLines in [glBoth, glHorizontal]  then begin
      ColumnLabel.Frame.DrawTop := True;
      ColumnLabel.Frame.DrawBottom := True;
    end;
    if FPrintGridLines in [glBoth, glVertical] then begin
      ColumnLabel.Frame.DrawRight := True;
      if x = 0 then ColumnLabel.Frame.DrawLeft := True;
    end;

    DetailLabel             := TQRDBText.Create(QuickRep);
    DetailLabel.Parent      := DetailBand;
    DetailLabel.Alignment   := FDBGrid.Columns[x].Alignment;
    if UseGridSettings then DetailLabel.Font.Assign(FDBGrid.Columns[x].Font);
    DetailLabel.AutoSize    := False;
    DetailLabel.Size.Top    := 0;
    DetailLabel.Size.Left   := CurPlacement;
    DetailLabel.Size.Height := DetailBand.Size.Height;
    DetailLabel.Size.Width  := MM(FDBGrid.Columns[x].Width);
    DetailLabel.DataSet     := FDataSet;
    DetailLabel.Datafield   := FDBGrid.Columns[x].FieldName;
    DetailLabel.AutoStretch := FAutoStretch;
    DetailLabel.Transparent := True;
    if FPrintGridLines in [glBoth, glHorizontal]  then
      DetailLabel.Frame.DrawBottom := True;
    if FPrintGridLines in [glBoth, glVertical] then begin
      DetailLabel.Frame.DrawRight := True;
      if x = 0 then DetailLabel.Frame.DrawLeft := True;
    end;

    CurPlacement := CurPlacement + DetailLabel.Size.Width;
  end;

  // Footer Band

  PageFooterBand          := TQRBand.Create(QuickRep);
  PageFooterBand.Parent   := QuickRep;
  PageFooterBand.BandType := rbPageFooter;
  PageFooterBand.Color    := FPageFooter.Color;
  AssignFrame(PageFooterBand.Frame, FPageFooter.Frame);
  PageFooterBand.Font.Assign(FPageFooter.Font);

  CurPlacement := 0;
  if Trim(FPageFooter.Value.Text) <> '' then begin
    for x := 0 to FPageFooter.Value.Count - 1 do
      if Trim(FPageFooter.Value[x]) <> '' then begin
        if FPageFooter.ValueAsExpr then begin
          CustomExpr := TQRExpr.Create(QuickRep);
          with CustomExpr do begin
            Parent      := PageFooterBand;
            Transparent := True;
            WordWrap    := False;
            Expression := FPageFooter.Value[x];
            Alignment   := FPageFooter.Alignment;
            AlignToBand := True;
            Size.Top    := CurPlacement;
            Size.Height := MM(-FPageFooter.Font.Height) + 1;
            CurPlacement := CurPlacement + Size.Height;
          end;
        end
        else begin
          CustomLabel := TQRLabel.Create(QuickRep);
          with CustomLabel do begin
            Parent      := PageFooterBand;
            Transparent := True;
            WordWrap    := False;
            Caption     := FPageFooter.Value[x];
            Alignment   := FPageFooter.Alignment;
            AlignToBand := True;
            Size.Top    := CurPlacement;
            Size.Height := MM(-FPageFooter.Font.Height) + 1;
            CurPlacement := CurPlacement + Size.Height;
          end;
        end;
      end;
  end;

  CustomLabel := TQRLabel.Create(QuickRep);
  with CustomLabel do begin
    Parent      := PageFooterBand;
    Transparent := True;
    Caption := DateTimeToStr(Now);
    Alignment   := taRightJustify;
    AlignToBand := True;
    Size.Top    := CurPlacement;
    Size.Height := MM(-FPageFooter.Font.Height) + 1;
    PageFooterBand.Size.Height := CurPlacement + Size.Height;
  end;

  // Summary Band

  SummaryBand          := TQRBand.Create(QuickRep);
  SummaryBand.Parent   := QuickRep;
  SummaryBand.BandType := rbSummary;
  SummaryBand.Color    := FSummary.Color;
  AssignFrame(SummaryBand.Frame, FSummary.Frame);
  SummaryBand.Font.Assign(FSummary.Font);
  CurPlacement := 0;
  if Trim(FSummary.Value.Text) <> '' then begin
    for x := 0 to FSummary.Value.Count - 1 do
      if Trim(FSummary.Value[x]) <> '' then begin
        if FSummary.ValueAsExpr then begin
          CustomExpr := TQRExpr.Create(QuickRep);
          with CustomExpr do begin
            Parent       := SummaryBand ;
            Transparent  := True;
            WordWrap     := False;
            Expression   := FSummary.Value[x];
            Size.Top     := CurPlacement;
            Size.Height  := MM(-FSummary.Font.Height) + 1;
            AlignToBand  := True;
            Alignment    := FSummary.Alignment;
            CurPlacement := CurPlacement + Size.Height;
          end;
        end
        else begin
          CustomLabel := TQRLabel.Create(QuickRep);
          with CustomLabel do begin
            Parent       := SummaryBand ;
            Transparent  := True;
            WordWrap     := False;
            Caption      := FSummary.Value[x];
            Size.Top     := CurPlacement;
            Size.Height  := MM(-FSummary.Font.Height) + 1;
            AlignToBand  := True;
            Alignment    := FSummary.Alignment;
            CurPlacement := CurPlacement + Size.Height;
          end;
        end;
      end;
    SummaryBand.Size.Height := CurPlacement;
  end;
  if Assigned(BeforePrint) then BeforePrint(Self);
  Result := True;
end;

procedure TQRDBGridReport.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) then
    if (AComponent = FDBGrid) and (FDBGrid <> nil) then FDBGrid := nil;
end;

procedure TQRDBGridReport.SetFont(const Value: TFont);
begin
  FFont.Assign(Value);
end;

procedure TQRDBGridReport.OnAfterPrint(Sender: TObject);
begin
  if (QuickRep <> nil) and (QuickRep.DataSet <> nil) then
    with QuickRep.DataSet do begin
      BookMark := FBookMarkStr;
//      EnableControls;
    end;
end;

procedure TQRDBGridReport.OnBeforePrint(Sender: TCustomQuickRep;
  var PrintReport: Boolean);
begin
  if (QuickRep <> nil) and (QuickRep.DataSet <> nil) then
    with QuickRep.DataSet do begin
      FBookMarkStr := Bookmark;
//      DisableControls;
    end;
end;

{TrQRBand}

constructor TrQRBand.Create(AOwner: TQRDBGridReport);
begin
  FParent     := AOwner;
  FValue      := TStringList.Create;
  FAlignment  := taCenter;
  FColor      := clWhite;
  FFont       := TFont.Create;
  FFont.Assign(AOwner.Font);
  FParentFont := True;
  FFont.OnChange := FontChanged;
  FFrame      := TQRFrame.Create;//(AOwner.QuickRep);
  FFrame.Parent := AOwner.QuickRep;
  FValueAsExpr := False;
end;

destructor TrQRBand.Destroy;
begin
  FValue.Free;
  FFrame.Free;
  CustomControl.Free;
  inherited Destroy;
end;

procedure TrQRBand.SetColor(Value: TColor);
begin
  FColor := Value;
end;

procedure TrQRBand.SetValue(Value: TStrings);
begin
  FValue.Assign(Value);
end;

procedure TrQRBand.SetFont(Value: TFont);
begin
  FFont.Assign(Value);
end;

procedure TrQRBand.SetParentFont(Value: Boolean);
begin
  if FParentFont <> Value then begin
    FParentFont := Value;
    if FParentFont and (FParent <> nil) then
      FFont.Assign(TQRDBGridReport(FParent).Font);
  end;
end;

procedure TrQRBand.FontChanged(Sender : TObject);
begin
  FParentFont := False;
end;

function TrQRBand.StoredFont : Boolean;
begin
  Result := not FParentFont;
end;

procedure TrQRBand.SetFrame(Value : TQRFrame);
begin
  FFrame.Assign(Value);
  //FParent.QuickRep.Invalidate;
end;

end.
