{===================================================================================================

  Comming Library

    .

====================================================================================================

  :

  01.09.2007 [*] .
  22.01.2003 [-] .
  29.11.2002 [*]    CommingDesign  CommingRuntime
  14.08.2002 [+]   .
  13.08.2002 [+] NoClick:   .
             [*]   (20  24).
             [*]    .
             [+] AutoSize:   .
             [-]   .
  12.08.2002 [+]   (   TsohoSpeedButton).
  09.09.1998 [!]  Comming Library.

====================================================================================================
  Copyright 19982007 TSK Developer Laboratory, Moscow.                     Konstantin Tsvetkov
                                                                              e-mail: dev@tsklab.ru
                                                                              http://www.tsklab.ru
  CVS: $Id: CommingButton.pas,v 1.13 2007/09/03 06:00:09 tsk Exp $
===================================================================================================}
unit CommingButton;

interface

uses  Windows, Messages, SysUtils, Classes, Controls, Graphics, ExtCtrls, Forms, ActnList,
      // Comming:
      CommingOther;

type

  TButtonState = (bsUp, bsDisabled, bsDown, bsExclusive);

  TcoButton = class(TGraphicControl)
  private
    FDown: Boolean;
    FDragging: Boolean;
    FFlat: Boolean;
    FMenuTracking:boolean;
    FMouseInControl: Boolean;
    FAlignment:TAlignment;
    FImageList: TImageList;
    FImageIndex: LongInt;
    FAutoSize: Boolean;
    FNoClick: Boolean;
    procedure   SetAlignment(Value: TAlignment);
    procedure   SetFlat(Value: Boolean);
    procedure   SetImageIndex(Value: LongInt);
    procedure   UpdateTracking;
    procedure   CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure   CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure   CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
    procedure   CMSysColorChange(var Message: TMessage); message CM_SYSCOLORCHANGE;
    procedure   CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure   CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure   SetImageList(const Value: TImageList);
    procedure   SetAutoSizeX( const Value: Boolean );
    procedure   CalcLayout;
    procedure   SetNoClick( Value: Boolean );
    function    GetHeightX: Integer;
    procedure   SetHeightX( Value: Integer);
    function GetWidthText: Integer;
  protected
    FState: TButtonState;
    procedure   ActionChange(Sender: TObject; CheckDefaults: Boolean); override;
    procedure   Loaded; override;
    procedure   MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure   DoMouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure   MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure   Paint; override;
    property    MouseInControl: Boolean read FMouseInControl;
    procedure   Notification(AComponent: TComponent;Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
  published
    property    Alignment: TAlignment read FAlignment write SetAlignment default taLeftJustify;
    property    AutoSize: Boolean read FAutoSize write SetAutoSizeX default True;
    property    NoClick: Boolean read FNoClick write SetNoClick default False;
    property    Images: TImageList read FImageList write SetImageList;
    property    ImageIndex: LongInt read FImageIndex write SetImageIndex default -1;
    property    Height: Integer read GetHeightX write SetHeightX;
    property    Caption;
    property    Enabled;
    property    Flat: Boolean read FFlat write SetFlat default True;
    property    Font;
    property    ParentFont;
    property    ParentShowHint;
    property    ShowHint;
    property    Visible;
    property    OnClick;
    property    Anchors;
    property    Action;
    property    Constraints;
    property    WidthText: Integer read GetWidthText stored False;
  end;

implementation

//==================================================================================================
{ TcoButton }
//==================================================================================================
constructor TcoButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetBounds( 0, 0, 24, 24 );
  ControlStyle  := [csCaptureMouse];
  ParentFont    := True;
  Color         := clBtnFace;
  FAlignment    := taLeftJustify;
  FFlat         := True;
  FAutoSize     := True;
  FNoClick      := False;
end;
//==================================================================================================
destructor TcoButton.Destroy;
begin
  inherited Destroy;
//  TButtonGlyph(FGlyph).Free;
end;
//==================================================================================================
procedure TcoButton.Paint;
  const DownStyles: array[Boolean] of Integer = (BDR_RAISEDINNER, BDR_SUNKENOUTER);
  var FramRect: TRect;
      TopLay: Integer;
begin
  //  :
  if not Enabled then begin
    FState := bsDisabled;
    FDragging := False;
  end else if FState = bsDisabled then
    if FDown then FState := bsExclusive
             else FState := bsUp;
  //  Canvas.
  Canvas.Font := Self.Font;
  Canvas.Brush.Color := Self.Color;
  //  :
  FramRect := Rect(0, 0, Width, Height);
  if FFlat then begin
    if ( FState in [bsDown, bsExclusive] )
        or ( FMouseInControl and (FState <> bsDisabled ) and not NoClick )
        or ( csDesigning in ComponentState ) then
      DrawEdge(Canvas.Handle, FramRect, DownStyles[FState in [bsDown, bsExclusive]], BF_RECT )
  end else begin
    DrawEdge(Canvas.Handle, FramRect, DownStyles[FState in [bsDown, bsExclusive]], BF_RECT )
  end;
  //  
  TopLay := iif( Height = 24, 4, 2 );
  if Caption <> '' then begin
    if Enabled then
      if Alignment = taLeftJustify then begin
        Canvas.TextOut( 24, TopLay, Caption );
      end else begin
        Canvas.TextOut( Width - 4 - Canvas.TextWidth(Caption), TopLay, Caption );
      end
    else
      if Alignment = taLeftJustify then begin
        Canvas.Brush.Style := bsClear;
        Canvas.Font.Color := clWhite;
        Canvas.TextOut( 25, TopLay + 1, Caption );
        Canvas.Font.Color := clDkGray;
        Canvas.TextOut( 24, TopLay, Caption );
      end else begin
        Canvas.Brush.Style := bsClear;
        Canvas.Font.Color := clWhite;
        Canvas.TextOut( Width - 3 - Canvas.TextWidth(Caption), TopLay + 1, Caption );
        Canvas.Font.Color := clDkGray;
        Canvas.TextOut( Width - 4 - Canvas.TextWidth(Caption), TopLay, Caption );
      end
  end;
  //  
  if Assigned( FImageList ) then
    if Enabled and not NoClick
      then FImageList.Draw(Canvas, 4, TopLay, FImageIndex, True )
      else FImageList.Draw(Canvas, 4, TopLay, FImageIndex, False );
end;
//==================================================================================================
procedure TcoButton.UpdateTracking;
var P: TPoint;
begin
  if FFlat then begin
    if Enabled then begin
      GetCursorPos(P);
      FMouseInControl := not (FindDragTarget(P, True) = Self);
      if FMouseInControl then Perform(CM_MOUSELEAVE, 0, 0)
      else Perform(CM_MOUSEENTER, 0, 0);
    end;
  end;
end;

procedure TcoButton.Loaded;
//var
//  State: TButtonState;
begin
  inherited Loaded;
//  if Enabled then State := bsUp
//  else State := bsDisabled;
//  TButtonGlyph(FGlyph).CreateButtonGlyph(State);
end;

procedure TcoButton.DoMouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  DoClick: Boolean;
begin
  if FDragging and (Button = mbLeft) then begin
    FDragging := False;
    DoClick := (X >= 0) and (X < ClientWidth) and (Y >= 0) and (Y <= ClientHeight);
    if DoClick then begin
(*      if Assigned(FDropDownMenu) then begin
        FDown := not FDown;
        FState := bsUp;
        Repaint;
      end;      *)
//      FState := bsExclusive;
//      Repaint;
//      Click;
    end
    else begin
      FDown := false;
      Repaint;
    end;
  end;
end;

procedure TcoButton.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
   P:TPoint;
   Msg:TMsg;
begin
  if FMenuTracking then Exit;
  inherited MouseDown(Button, Shift, X, Y);
  if (Button = mbLeft) and Enabled and not NoClick then begin
    FDown := not FDown;
    if FDown then FState := bsDown
    else FState := bsUp;
    Repaint;
    FDragging := True;
    FMenuTracking := True;
    try
//      P := GetDropDownMenuPos;
//      if CheckMenuDropDown(PointToSmallPoint(P), False) then
//        MouseUp(Button, Shift, X, Y);
      if PeekMessage(Msg, 0, 0, 0, PM_NOREMOVE) then begin
        if (Msg.Message = WM_LBUTTONDOWN) or (Msg.Message = WM_LBUTTONDBLCLK) then
        begin
          P := ScreenToClient(Msg.Pt);
//          if (P.X >= 0) and (P.X < ClientWidth) and (P.Y >= 0)
//            and (P.Y <= ClientHeight) then KillMessage(0, Msg.Message);
        end;
      end;
    finally
      FMenuTracking := False;
      FDown := false;
    end;
  end;
        
end;

procedure TcoButton.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  NewState: TButtonState;
begin
  inherited MouseMove(Shift, X, Y);
  if FDragging then begin
    if not FDown then NewState := bsUp
    else NewState := bsExclusive;
    if (X >= 0) and (X < ClientWidth) and (Y >= 0) and (Y <= ClientHeight) then
      if FDown then NewState := bsExclusive else NewState := bsDown;
    if NewState <> FState then begin
      FState := NewState;
      Invalidate;
    end;
  end
end;
    
procedure TcoButton.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  DoClick: Boolean;
begin
  inherited MouseUp(Button, Shift, X, Y);
  if FDragging then begin
    FDragging := False;
    DoClick := (X >= 0) and (X < ClientWidth) and (Y >= 0) and (Y <= ClientHeight);
//    if FGroupIndex = 0 then begin
      { Redraw face in-case mouse is captured }
      FState := bsUp;
      FMouseInControl := False;
      if DoClick and not (FState in [bsExclusive, bsDown]) then
        Invalidate;
{    end
    else
      if DoClick then begin
        SetDown(not FDown);
        if FDown then Repaint;
      end
      else begin
        if FDown then FState := bsExclusive;
        Repaint;
      end;}
    if DoClick and not FMenuTracking then Click;
    UpdateTracking;
  end;
end;
//==================================================================================================
procedure TcoButton.SetAlignment( Value: TAlignment );
begin
  if ( FAlignment <> Value ) and ( Value <> taCenter) then begin
    FAlignment := Value;
    if FAlignment = taRightJustify then AutoSize := False;
    Invalidate;
  end;
end;
//==================================================================================================
procedure TcoButton.SetAutoSizeX( const Value: Boolean );
begin
  if FAutoSize <> Value then begin
    FAutoSize := Value;
    if FAutoSize then Alignment := taLeftJustify;
    CalcLayout;
    Invalidate;
  end;
end;
//==================================================================================================
(*procedure TcoButton.UpdateExclusive;
var
  Msg: TMessage;
begin
  if (FGroupIndex <> 0) and (Parent <> nil) then
  begin
    Msg.Msg := CM_BUTTONPRESSED;
    Msg.WParam := FGroupIndex;
    Msg.LParam := Longint(Self);
    Msg.Result := 0;
    Parent.Broadcast(Msg);
  end;
end;*)

(*procedure TcoButton.SetDown(Value: Boolean);
begin
  {if FGroupIndex = 0 then }Value := False;
  if Value <> FDown then
  begin
//    if FDown and (not FAllowAllUp) then Exit;
    FDown := Value;
    if Value then
    begin
      if FState = bsUp then Invalidate;
      FState := bsExclusive
    end
    else
    begin
      FState := bsUp;
      Repaint;
    end;
//    if Value then UpdateExclusive;
  end;
end;        *)
//==================================================================================================
procedure TcoButton.SetFlat( Value: Boolean );
begin
  if Value <> FFlat then begin
    FFlat := Value;
    Invalidate;
  end;
end;
//==================================================================================================
{procedure TcoButton.SetGroupIndex(Value: Integer);
begin
  if FGroupIndex <> Value then
  begin
    FGroupIndex := Value;
    UpdateExclusive;
  end;
end;}


(*procedure TcoButton.WMLButtonDblClk(var Message: TWMLButtonDown);
begin
  if not FMenuTracking then
  begin
    inherited;
    if FDown then DblClick;
  end;
end;           *)

procedure TcoButton.CMEnabledChanged( var Message: TMessage );
begin
  UpdateTracking;
  Repaint;
//  Invalidate;
end;

(*ocedure TcoButton.CMButtonPressed(var Message: TMessage);
var
  Sender: TcoButton;
begin
  if Message.WParam = 0 {FGroupIndex} then
  begin
    Sender := TcoButton(Message.LParam);
    if Sender <> Self then
    begin
{      if Sender.Down and FDown then
      begin
        FDown := False;
        FState := bsUp;
        Invalidate;
      end;}
      FAllowAllUp := Sender.AllowAllUp;
    end;
  end;
end;*)
    
{procedure TcoButton.CMDialogChar(var Message: TCMDialogChar);
begin
  with Message do
    if IsAccel(CharCode, Caption) and Enabled and Visible and
      (Parent <> nil) and Parent.Showing then begin
      if GroupIndex = 0 then Click
      else begin
        Down := not Down;
        Click;
      end;
      Result := 1;
     end
     else inherited;
end;                         }

procedure TcoButton.CMFontChanged(var Message: TMessage);
begin
  Invalidate;
end;

procedure TcoButton.CMTextChanged(var Message: TMessage);
begin
  if AutoSize then CalcLayout;
  Invalidate;
end;

procedure TcoButton.CMSysColorChange(var Message: TMessage);
begin
//  with TButtonGlyph(FGlyph) do begin
    Invalidate;
//    CreateButtonGlyph(FState);
//  end;
end;

procedure TcoButton.CMMouseEnter(var Message: TMessage);
begin
  inherited;
  { Don't draw a border if DragMode <> dmAutomatic since this button is meant to
    be used as a dock client. }
  if FFlat and not FMouseInControl and Enabled and (DragMode <> dmAutomatic)
    and (GetCapture = 0) then
  begin
    FMouseInControl := True;
    Repaint;
  end;
end;

procedure TcoButton.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  if FFlat and FMouseInControl and Enabled and not FDragging then begin
    FMouseInControl := False;
    Invalidate;
  end;
end;

//{$IFDEF Delphi4_or_5}
procedure TcoButton.ActionChange(Sender: TObject; CheckDefaults: Boolean);
begin
  inherited;
  if Sender is TCustomAction then
    with TCustomAction(Sender) do
    begin
      if not CheckDefaults or (Self.ImageIndex = -1) then
        Self.ImageIndex := ImageIndex;
//        Self.Caption    := Caption;
//        Self.Images     := Images;
    end;
end;
//{$ENDIF}

(*function TcoButton.GetDropDownMenuPos: TPoint;
begin
  if Assigned(FDropDownMenu) then begin
    if MenuPosition = dmpBottom then begin
      case FDropDownMenu.Alignment of
        paLeft: Result := Point(-1, Height);
        paRight: Result := Point(Width + 1, Height);
        else Result := Point(Width div 2, Height);
      end;
    end
    else
    begin
      case FDropDownMenu.Alignment of
        paLeft: Result := Point(Width, -1);
        paRight: Result := Point(-1, -1);
        else Result := Point(Width div 2, Height);
      end;
    end;
  end else Result := Point(0, 0);
end;            *)

procedure TcoButton.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
//  if (AComponent = DropDownMenu) and (Operation = opRemove) then
//    DropDownMenu := nil;
  if (AComponent = FImageList) and (Operation = opRemove) then
    FImageList := nil;
end;

(*procedure TcoButton.SetDropDownMenu(Value: TPopupMenu);
begin
  FDropDownMenu := Value;
{$IFDEF WIN32}
  if Value <> nil then Value.FreeNotification(Self);
{$ENDIF}
//  if FMarkDropDown then Invalidate;
end;

function TcoButton.CheckMenuDropDown(const Pos: TSmallPoint;
  Manual: Boolean): Boolean;
var
  Form: TCustomForm;
begin
  Result := False;
  if csDesigning in ComponentState then Exit;
  if Assigned(FDropDownMenu) and (DropDownMenu.AutoPopup or Manual) then
  begin
    Form := GetParentForm(Self);
    if Form <> nil then Form.SendCancelMode(nil);
    DropDownMenu.PopupComponent := Self;
    with ClientToScreen(SmallPointToPoint(Pos)) do DropDownMenu.Popup(X, Y);
    Result := True;
  end;
end;                            *)
//==================================================================================================
procedure TcoButton.SetImageIndex(Value: Integer);
begin
  if FImageIndex = Value then Exit;
  FImageIndex := Value;
  Repaint;
end;
//==================================================================================================
procedure TcoButton.SetImageList( const Value: TImageList );
begin
  FImageList := Value;
  if AutoSize then CalcLayout;
  Repaint;
end;
//==================================================================================================
procedure TcoButton.CalcLayout;
  var ImPlace: Integer;
      LeftDelta: Integer;
begin
  if Length( Caption ) > 0 then begin
    ImPlace := iif( Assigned( FImageList ), 28, 8 );
    Canvas.Font := Self.Font;
    LeftDelta := ( Width - Canvas.TextWidth( Caption ) - ImPlace );
    Width := Width - LeftDelta;
{    if akRight in Anchors then begin
      Left := Left + LeftDelta;
    end;}
  end
end;
//==================================================================================================
procedure TcoButton.SetNoClick( Value: Boolean );
begin
//  Value := Value and Assigned( OnClick );
  if FNoClick <> Value then begin
    FNoClick := Value;
    Invalidate;
  end;
end;
//==================================================================================================
function TcoButton.GetHeightX: Integer;
begin
  Result := inherited Height;
end;
//==================================================================================================
procedure TcoButton.SetHeightX( Value: Integer );
begin
  if Value < 22 then Value := 20;
  if Value > 23 then Value := 24;
  inherited Height := Value;
end;
//==================================================================================================
function TcoButton.GetWidthText: Integer;
begin
  Canvas.Font := Self.Font;
  Result      := Canvas.TextWidth( Caption );
end;
//==================================================================================================
end.
