{===================================================================================================

  Comming Library

    ComboBox   .

====================================================================================================

  :

  01.09.2007 [*] .
  29.11.2002 [*]    CommingDesign  CommingRuntime
  11.10.2002 [+] TcoCustomDBComboBox: Color.
  26.08.2002 [+] TcoDBComboBox: Color.
  14.03.2002 [+] Enabled:     .
  15.12.2001 [+] ReadOnly:  .
  16.08.2001 [-]   $IFDEF.
  02.07.2001 [*]   Delphi6.
  13.06.2001 [-]    "".
  08.01.2001 [+] "".
  20.12.2000 [+]  .
  09.09.1998 [!]  Comming Library.

====================================================================================================
  Copyright 19982007 TSK Developer Laboratory, Moscow.                     Konstantin Tsvetkov
                                                                              e-mail: dev@tsklab.ru
                                                                              http://www.tsklab.ru
  CVS: $Id: CommingDBComboBox.pas,v 1.8 2007/09/03 06:00:09 tsk Exp $
===================================================================================================}
unit CommingDBComboBox;

interface

uses  Windows, DbCtrls, Messages, Menus, Graphics, Classes, Controls, DB, StdCtrls, DBConsts,
      // Comming:
      CommingColor, CommingDisplay;

type

{ TCustomDBComboBox }

  TcoCustomDBComboBox = class( TCustomComboBox )
  private
    FDataLink: TFieldDataLink;
    FPaintControl: TPaintControl;
    FEmptyValue: String;
    procedure DataChange(Sender: TObject);
    procedure EditingChange(Sender: TObject);
    function  GetDataField: string;
    function  GetDataSource: TDataSource;
    function  GetField: TField;
    function  GetReadOnly: Boolean;
    procedure SetDataField(const Value: string);
    procedure SetDataSource(Value: TDataSource);
    procedure SetEditReadOnly;
    procedure SetItems(Value: TStrings);
    procedure SetReadOnly(Value: Boolean);
    procedure   UpdateData(Sender: TObject);
    function  GetComboText: string; virtual;
    procedure SetComboText(const Value: string); virtual;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure CMGetDataLink(var Message: TMessage); message CM_GETDATALINK;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
  protected
    procedure Change; override;
    procedure Click; override;
    procedure ComboWndProc(var Message: TMessage; ComboWnd: HWnd; ComboProc: Pointer); override;
    procedure CreateWnd; override;
    procedure DropDown; override;
    function GetPaintText: string; virtual;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure WndProc(var Message: TMessage); override;
    property ComboText: string read GetComboText write SetComboText;
    property DataField: string read GetDataField write SetDataField;
    property    DataSource: TDataSource read GetDataSource write SetDataSource;
    property    ReadOnly: Boolean read GetReadOnly write SetReadOnly default False;
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    function    ExecuteAction(Action: TBasicAction): Boolean; override;
    function    UpdateAction(Action: TBasicAction): Boolean; override;
    function    UseRightToLeftAlignment: Boolean; override;
    property    Field: TField read GetField;
    property    Items write SetItems;
    property    Text;
    property    EmptyValue: String read FEmptyValue write FEmptyValue;
  end;

{ TcoDBComboBox }

  TcoDBComboBox = class(TcoCustomDBComboBox)
  private
    FValues: TStrings;
    FEnableValues: Boolean;
//    FFlat: Boolean;
    procedure   SetEnableValues(Value: Boolean);
    procedure   SetValues(Value: TStrings);
    procedure   ValuesChanged(Sender: TObject);
    procedure   WMPaint (var Msg : TWMPaint); message WM_PAINT;
    function    GetEnabled_: Boolean;
    procedure   SetEnabled_(const Value: Boolean);
//    procedure SetFlat(const Value: Boolean);
  protected
//    procedure SetStyle(Value: TComboBoxStyle); override;
    function    GetComboText: string; override;
    function    GetPaintText: string; override;
    procedure   SetComboText(const Value: string); override;
  public
    constructor   Create(AOwner: TComponent); override;
    destructor    Destroy; override;
  published
//    property Align;  // Polaris
      { must be published before Items }
    property    Style default csOwnerDrawFixed; // TSK   // Polaris
    property Color;
    property Ctl3D;
    property DataField;
    property DataSource;
    property DragMode;
    property DragCursor;
    property DropDownCount;
    property Enabled: Boolean read GetEnabled_ write SetEnabled_ default True;
    property EnableValues: Boolean read FEnableValues write SetEnableValues default True;   // Polaris
    property EmptyValue;
    property Font;
    property    Height default 22; // TSK
    property Anchors;
    property BiDiMode;
    property Constraints;
    property DragKind;
    property ParentBiDiMode;
//    property ImeMode;
//    property ImeName;
    property ItemHeight default 16;
    property Items;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property Sorted;
    property TabOrder;
    property TabStop;
    property Values: TStrings read FValues write SetValues;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawItem;
    property OnDropDown;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMeasureItem;
    property OnStartDrag;
    property OnContextPopup;
//    property OnEndDock;
//    property OnStartDock;
//    property  Flat: Boolean read FFlat write SetFlat;
  end;

// procedure Register;

implementation

(*uses DBUtils;*)

{ TcoCustomDBComboBox }

constructor TcoCustomDBComboBox.Create( AOwner: TComponent );
begin
  inherited Create( AOwner );
  ControlStyle := ControlStyle + [csReplicatable];
  FDataLink := TFieldDataLink.Create;
  FDataLink.Control := Self;
  FDataLink.OnDataChange := DataChange;
  FDataLink.OnUpdateData := UpdateData;
  FDataLink.OnEditingChange := EditingChange;
  FPaintControl := TPaintControl.Create(Self, 'COMBOBOX');
  if not ( csDesigning in ComponentState ) then begin
    Color       := clDimWindow;
  end;
end;

destructor TcoCustomDBComboBox.Destroy;
begin
{$IFDEF WIN32}
  FPaintControl.Free;
{$ENDIF}
  FDataLink.OnDataChange := nil;
  FDataLink.OnUpdateData := nil;
  FDataLink.Free;
  FDataLink := nil;
  inherited Destroy;
end;

procedure TcoCustomDBComboBox.Loaded;
begin
  inherited Loaded;
  if (csDesigning in ComponentState) then DataChange(Self);
end;

procedure TcoCustomDBComboBox.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (FDataLink <> nil) and
    (AComponent = DataSource) then DataSource := nil;
end;

procedure TcoCustomDBComboBox.CreateWnd;
begin
  inherited CreateWnd;
  SetEditReadOnly;
end;

procedure TcoCustomDBComboBox.DataChange(Sender: TObject);
begin
  if DroppedDown then Exit;
  if FDataLink.Field <> nil then ComboText := FDataLink.Field.Text
  else if csDesigning in ComponentState then ComboText := Name
  else ComboText := '';
end;

procedure TcoCustomDBComboBox.UpdateData(Sender: TObject);
begin
  if ComboText = FEmptyValue
    then FDataLink.Field.Clear
    else FDataLink.Field.Text := ComboText;
end;

procedure TcoCustomDBComboBox.SetComboText(const Value: string);
var
  I: Integer;
  Redraw: Boolean;
begin
  if Value <> ComboText then begin
    if Style <> csDropDown then begin
      Redraw := (Style <> csSimple) and HandleAllocated;
      if Redraw then SendMessage(Handle, WM_SETREDRAW, 0, 0);
      try
        if Value = '' then I := -1 else I := Items.IndexOf(Value);
        ItemIndex := I;
      finally
        if Redraw then begin
          SendMessage(Handle, WM_SETREDRAW, 1, 0);
          Invalidate;
        end;
      end;
      if I >= 0 then Exit;
    end;
    if Style in [csDropDown, csSimple] then Text := Value;
  end;
end;

function TcoCustomDBComboBox.GetComboText: string;
var
  I: Integer;
begin
  if Style in [csDropDown, csSimple] then Result := Text
  else begin
    I := ItemIndex;
    if I < 0 then Result := '' else Result := Items[I];
  end;
end;

procedure TcoCustomDBComboBox.Change;
begin
  FDataLink.Edit;
  inherited Change;
  FDataLink.Modified;
end;

procedure TcoCustomDBComboBox.Click;
begin
  FDataLink.Edit;
  inherited Click;
  FDataLink.Modified;
end;

procedure TcoCustomDBComboBox.DropDown;
begin
{$IFNDEF WIN32}
  FDataLink.Edit;
{$ENDIF}
  // TSK
//  if not ReadOnly then
  inherited DropDown;
end;

function TcoCustomDBComboBox.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

procedure TcoCustomDBComboBox.SetDataSource(Value: TDataSource);
begin
//{$IFDEF RX_D4}
  if not (FDataLink.DataSourceFixed and (csLoading in ComponentState)) then
//{$ENDIF}
    FDataLink.DataSource := Value;
{$IFDEF WIN32}
  if Value <> nil then Value.FreeNotification(Self);
{$ENDIF}
end;

function TcoCustomDBComboBox.GetDataField: string;
begin
  Result := FDataLink.FieldName;
end;

procedure TcoCustomDBComboBox.SetDataField(const Value: string);
begin
  FDataLink.FieldName := Value;
end;

function TcoCustomDBComboBox.GetReadOnly: Boolean;
begin
  Result := FDataLink.ReadOnly;
end;

procedure TcoCustomDBComboBox.SetReadOnly(Value: Boolean);
begin
  FDataLink.ReadOnly := Value;
// TSK
  if Value then Color := clBtnFace else begin
    if not ( csDesigning in ComponentState )
      then Color := clDimWindow
      else Color := clWindow;
  end;
  RecreateWnd;
end;

function TcoCustomDBComboBox.GetField: TField;
begin
  Result := FDataLink.Field;
end;

procedure TcoCustomDBComboBox.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key, Shift);
  if Key in [VK_BACK, VK_DELETE, VK_UP, VK_DOWN, 32..255] then begin
    if not FDataLink.Edit and (Key in [VK_UP, VK_DOWN]) then
      Key := 0;
  end;
end;

procedure TcoCustomDBComboBox.KeyPress(var Key: Char);
begin
  inherited KeyPress(Key);
  if (Key in [#32..#255]) and (FDataLink.Field <> nil) and
    not FDataLink.Field.IsValidChar(Key) then
  begin
    MessageBeep(0);
    Key := #0;
  end;
  case Key of
    ^H, ^V, ^X, #32..#255:
      FDataLink.Edit;
    #27:
      begin
        FDataLink.Reset;
        SelectAll;
{$IFNDEF WIN32}
        Key := #0;
{$ENDIF}
      end;
  end;
end;

procedure TcoCustomDBComboBox.EditingChange(Sender: TObject);
begin
  SetEditReadOnly;
end;

procedure TcoCustomDBComboBox.SetEditReadOnly;
begin
  if (Style in [csDropDown, csSimple]) and HandleAllocated then
    SendMessage({$IFDEF WIN32} EditHandle {$ELSE} FEditHandle {$ENDIF},
      EM_SETREADONLY, Ord(not FDataLink.Editing), 0);
end;

procedure TcoCustomDBComboBox.WndProc(var Message: TMessage);
begin
  if not (csDesigning in ComponentState) then
    case Message.Msg of
      WM_COMMAND:
        if TWMCommand(Message).NotifyCode = CBN_SELCHANGE then
          if not FDataLink.Edit then begin
            if Style <> csSimple then
              PostMessage(Handle, CB_SHOWDROPDOWN, 0, 0);
            Exit;
          end;
      CB_SHOWDROPDOWN:
        if Message.WParam <> 0 then FDataLink.Edit
        else if not FDataLink.Editing then DataChange(Self); {Restore text}
{$IFDEF WIN32}
      WM_CREATE,
      WM_WINDOWPOSCHANGED,
      CM_FONTCHANGED:
        FPaintControl.DestroyHandle;
{$ENDIF}
    end;
  inherited WndProc(Message);
end;

procedure TcoCustomDBComboBox.ComboWndProc(var Message: TMessage; ComboWnd: HWnd;
  ComboProc: Pointer);
begin
  if not (csDesigning in ComponentState) then
    case Message.Msg of
      WM_LBUTTONDOWN:
{$IFDEF WIN32}
        if (Style = csSimple) and (ComboWnd <> EditHandle) then
{$ELSE}
        if (Style = csSimple) and (ComboWnd <> FEditHandle) then
{$ENDIF}
          if not FDataLink.Edit then Exit;
    end;
  inherited ComboWndProc(Message, ComboWnd, ComboProc);
end;

procedure TcoCustomDBComboBox.CMExit(var Message: TCMExit);
begin
  try
    FDataLink.UpdateRecord;
  except
    SelectAll;
    if CanFocus then SetFocus;
    raise;
  end;
  inherited;
end;

{$IFDEF WIN32}
procedure TcoCustomDBComboBox.CMGetDatalink(var Message: TMessage);
begin
  Message.Result := Longint(FDataLink);
end;

procedure TcoCustomDBComboBox.WMPaint(var Message: TWMPaint);
var
  S: string;
  R: TRect;
  P: TPoint;
  Child: HWND;
begin
  if csPaintCopy in ControlState then begin
    S := GetPaintText;
    if Style = csDropDown then begin
      SendMessage(FPaintControl.Handle, WM_SETTEXT, 0, Longint(PChar(S)));
      SendMessage(FPaintControl.Handle, WM_PAINT, Message.DC, 0);
      Child := GetWindow(FPaintControl.Handle, GW_CHILD);
      if Child <> 0 then begin
        Windows.GetClientRect(Child, R);
        Windows.MapWindowPoints(Child, FPaintControl.Handle, R.TopLeft, 2);
        GetWindowOrgEx(Message.DC, P);
        SetWindowOrgEx(Message.DC, P.X - R.Left, P.Y - R.Top, nil);
        IntersectClipRect(Message.DC, 0, 0, R.Right - R.Left, R.Bottom - R.Top);
        SendMessage(Child, WM_PAINT, Message.DC, 0);
      end;
    end
    else begin
      SendMessage(FPaintControl.Handle, CB_RESETCONTENT, 0, 0);
      if Items.IndexOf(S) <> -1 then begin
        SendMessage(FPaintControl.Handle, CB_ADDSTRING, 0, Longint(PChar(S)));
        SendMessage(FPaintControl.Handle, CB_SETCURSEL, 0, 0);
      end;
      SendMessage(FPaintControl.Handle, WM_PAINT, Message.DC, 0);
    end;
  end
  else inherited;
//  RepaintFlatBorder( FPaintControl.Handle );  // TSK
end;
{$ENDIF}

function TcoCustomDBComboBox.GetPaintText: string;
begin
  if FDataLink.Field <> nil then Result := FDataLink.Field.Text
  else Result := '';
end;

procedure TcoCustomDBComboBox.SetItems(Value: TStrings);
begin
  Items.Assign(Value);
  DataChange(Self);
end;

(*{$IFNDEF WIN32}
function TcoCustomDBComboBox.GetStyle: TComboBoxStyle;
begin
  Result := inherited Style;
end;
{$ENDIF}

(*procedure TcoCustomDBComboBox.SetStyle(Value: TComboBoxStyle);
begin
//{$IFDEF WIN32}
  if (Value = csSimple) and Assigned(FDatalink) and FDatalink.DatasourceFixed then
// TSK   _DBError(SNotReplicatable);
    DatabaseError(SNotReplicatable);
  inherited SetStyle(Value);
(*{$ELSE}
  if Value = csSimple then ControlStyle := ControlStyle - [csFixedHeight]
  else ControlStyle := ControlStyle + [csFixedHeight];
  inherited Style := Value;
  RecreateWnd;
//{$ENDIF}
end;*)

//{$IFDEF RX_D4}
function TcoCustomDBComboBox.UseRightToLeftAlignment: Boolean;
begin
  Result := DBUseRightToLeftAlignment(Self, Field);
end;

function TcoCustomDBComboBox.ExecuteAction(Action: TBasicAction): Boolean;
begin
  Result := inherited ExecuteAction(Action) or (FDataLink <> nil) and
    FDataLink.ExecuteAction(Action);
end;

function TcoCustomDBComboBox.UpdateAction(Action: TBasicAction): Boolean;
begin
  Result := inherited UpdateAction(Action) or (FDataLink <> nil) and
    FDataLink.UpdateAction(Action);
end;
//{$ENDIF}

{ TcoDBComboBox }

constructor TcoDBComboBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FValues := TStringList.Create;
  TStringList(FValues).OnChange := ValuesChanged;
//  EnableValues := False;
  FEnableValues := True;    // Polaris
  Style         := csOwnerDrawFixed;  // csDropDownList;  // Polaris
  Height        := 22;
  if not ( csDesigning in ComponentState ) then begin
    Color       := clDimWindow;
  end;
//  Height := 20; //TSK
end;

destructor TcoDBComboBox.Destroy;
begin
  TStringList(FValues).OnChange := nil;
  FValues.Free;
  inherited Destroy;
end;

procedure TcoDBComboBox.ValuesChanged(Sender: TObject);
begin
  if FEnableValues then DataChange(Self);
end;

function TcoDBComboBox.GetPaintText: string;
var
  I: Integer;
begin
  Result := '';
  if FDataLink.Field <> nil then begin
    if FEnableValues then begin
      I := Values.IndexOf(FDataLink.Field.Text);
      if I >= 0 then Result := Items.Strings[I]
    end
    else Result := FDataLink.Field.Text;
  end;
end;

function TcoDBComboBox.GetComboText: string;
var
  I: Integer;
begin
  if (Style in [csDropDown, csSimple]) and (not FEnableValues) then
    Result := Text
  else begin
    I := ItemIndex;
    if (I < 0) or (FEnableValues and (FValues.Count < I + 1)) then
      Result := ''
    else
      if FEnableValues then Result := FValues[I]
      else Result := Items[I];
  end;
end;

procedure TcoDBComboBox.SetComboText(const Value: string);
var
  I: Integer;
  Redraw: Boolean;
begin
  if Value <> ComboText then begin
    if Style <> csDropDown then begin
      Redraw := (Style <> csSimple) and HandleAllocated;
      if Redraw then SendMessage(Handle, WM_SETREDRAW, 0, 0);
      try
        if Value = '' then I := -1 else
          if FEnableValues then I := Values.IndexOf(Value)
          else I := Items.IndexOf(Value);
        if I >= Items.Count then I := -1;
        ItemIndex := I;
      finally
        if Redraw then begin
          SendMessage(Handle, WM_SETREDRAW, 1, 0);
          Invalidate;
        end;
      end;
      if I >= 0 then Exit;
    end;
    if Style in [csDropDown, csSimple] then Text := Value;
  end;
end;

procedure TcoDBComboBox.SetEnableValues(Value: Boolean);
begin
  if FEnableValues <> Value then begin
    if Value and (Style in [csDropDown, csSimple]) then
      Style := csDropDownList;
    FEnableValues := Value;
    DataChange(Self);
  end;
end;

procedure TcoDBComboBox.SetValues(Value: TStrings);
begin
  FValues.Assign(Value);
end;

(*procedure TcoDBComboBox.SetStyle(Value: TComboboxStyle);
begin
  if (Value in [csSimple, csDropDown]) and FEnableValues then
//    Value := csDropDownList;
    FEnableValues := False;  // Polaris
  inherited SetStyle(Value);
end;*)

(*procedure RepaintFlatBorder( Handle: HWND );
  var DC : HDC;
      aRect : TRect;
      BrushFrame: HBrush;
begin
  BrushFrame := CreateSolidBrush({clYellow} GetSysColor( COLOR_BTNFACE ) );
  DC := GetWindowDC( Handle );
  try
    GetWindowRect( Handle, aRect );
    OffsetRect( aRect, -aRect.Left, -aRect.Top );
    FrameRect( DC, aRect, BrushFrame);
    InflateRect(aRect, -1, -1);
    DrawEdge( DC, aRect, BDR_SUNKENOUTER, BF_RECT )
  finally
    ReleaseDC( Handle, DC);
    DeleteObject( BrushFrame );
  end;
end;       *)

procedure TcoDBComboBox.WMPaint( var Msg: TWMPaint );
//  var x: TRect;
begin
  inherited;
(*  x := BoundsRect;
  x.Bottom := x.Bottom - 4;
  BoundsRect := x;
  if FFlat then *) RepaintFlatBorder( Handle );
end;

(*procedure TcoDBComboBox.SetFlat( const Value: Boolean );
begin
  FFlat := Value;
  if FFlat then begin
    Style := csOwnerDrawFixed;  // TSK
  end;
  RecreateWnd;
end;*)
//==================================================================================================
function TcoDBComboBox.GetEnabled_: Boolean;
begin
  Result := inherited Enabled;
end;
//==================================================================================================
procedure TcoDBComboBox.SetEnabled_(const Value: Boolean);
begin
  inherited Enabled := Value;
  if Enabled and not ReadOnly then begin
    if not ( csDesigning in ComponentState )
      then Color := clDimWindow
      else Color := clWindow;
  end else Color := clBtnFace;
  if Enabled then Font.Color := clWindowText else Font.Color := clGrayText;
  RecreateWnd;
end;
//==================================================================================================
end.


