{===================================================================================================

  Comming Library

       EhLib.

====================================================================================================

  :

  10.05.2007 [+]   -.
  16.01.2004 [+] SortMarkingChanged:      
                  .
  02.09.2003 [+] SettingMarker:    .
  27.03.2003 [+] ColumnByField:    DataSet.
  17.02.2003 [*] ShowHintData:    Hint.
  16.02.2003 [+] Hint  .
  13.02.2003 [*]     .
  30.11.2002 [*]  .
  29.11.2002 [-]   TAB.
             [*]    CommingDesign  CommingRuntime.
  17.07.2002 [+]   SettingMarker.
  18.04.2002 [+]  ScrollBar  .
  08.01.2002 [*]  :  .
  15.12.2001 [+]   (,  Height).
             [*]   "no parent window".
  02.07.2001 [*]   Delphi6.
  12.04.2001 [+]  .  .
  09.09.1998 [!]  Comming Library.

====================================================================================================
  Copyright 19982007 TSK Developer Laboratory, Moscow.                     Konstantin Tsvetkov
                                                                              e-mail: dev@tsklab.ru
                                                                              http://www.tsklab.ru
  CVS: $Id: CommingDBGridEh.pas,v 1.14 2007/10/29 12:43:39 tsk Exp $
===================================================================================================}
unit CommingDBGridEh;

interface

uses  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Grids, DBGrids,
      DBGridEh, Math, DB, ADODB,
      // Comming:
      CommingColor, CommingOther;

type
  TcoDBGridEh = class( TDBGridEh )
  private     { Private declarations }
    FAutoHeightBinding: Boolean;
    FShowHintData: Boolean;
    FShowHintDataActive: Boolean;
    function    GetEmptyRow: Integer;
    procedure   SetAutoHeightBinding( const Value: Boolean );
    procedure   SetHeight( const Value: Integer );
    function    GetHeight: Integer;
    procedure   SetShowHintData( const Value: Boolean );
    procedure   HintDataShow( Sender: TObject );
    procedure   CMHintShow(var Message: TCMHintShow); message CM_HINTSHOW;
  public      { Public declarations }
    constructor Create( AOwner: TComponent ); override;
    procedure   HeightBinding( Par: TControl = nil );
    procedure   SettingMarker( PosSet: String );
    procedure   ClearMarker;
    function    ColumnByField( Field: TField ): TColumnEh;
    procedure   SortMarkingChanged;
  published   { Published declarations }
    property    Height: Integer read GetHeight write SetHeight;
    property    AutoHeightBinding: Boolean read FAutoHeightBinding write SetAutoHeightBinding;
    property    ShowHintData: Boolean read FShowHintData write SetShowHintData default True;
    property    EmptyRow: Integer read GetEmptyRow stored False;
  end;

implementation

//==================================================================================================
{ TcoDBGridEh }
//==================================================================================================
constructor TcoDBGridEh.Create( AOwner: TComponent );
begin
  inherited Create( AOwner );
  Flat                            := True;
  AutoFitColWidths                := True;
  if not ( csDesigning in ComponentState ) then begin
    Color                         := clDimWindow;
  end;
  FAutoHeightBinding              := False;
  Options                         := [ dgTitles, dgColLines, dgRowLines, dgTabs, dgRowSelect,
                                       dgAlwaysShowSelection, dgConfirmDelete, dgCancelOnExit ];
  VertScrollBar.VisibleMode       := sbAlwaysShowEh;
  ColumnDefValues.Title.Alignment := taCenter;
  ShowHintData                    := True;
end;
//==================================================================================================
function TcoDBGridEh.GetEmptyRow: Integer;
  var C: Integer;
      R: Integer;
begin
  R := DefaultRowHeight + iif( dgRowLines in Options, GridLineWidth, 0 ); //   .
  C := iif( dgTitles in Options, RowHeights[0], 0 );                      // 
  if FooterRowCount > 0 then C := C + ( R * FooterRowCount ) + 1;         // 
  Result := (( ClientHeight - C ) mod R ) - 1;
end;
//==================================================================================================
procedure TcoDBGridEh.SetAutoHeightBinding( const Value: Boolean );
begin
  FAutoHeightBinding := Value;
end;
//==================================================================================================
function TcoDBGridEh.GetHeight: Integer;
begin
  Result := inherited Height;
end;
//==================================================================================================
procedure TcoDBGridEh.SetHeight( const Value: Integer );
begin
  inherited Height := Value;
  if FAutoHeightBinding and ( GetEmptyRow <> 0 ) then HeightBinding;
end;
//==================================================================================================
procedure TcoDBGridEh.HeightBinding( Par: TControl );
begin
  if ( Align in [ alTop, alBottom ] ) then Height := Height - GetEmptyRow;
  if ( Align in [ alClient, alNone ] ) and ( Par <> nil ) then Par.Height := Par.Height - GetEmptyRow;
end;
//==================================================================================================
procedure TcoDBGridEh.SettingMarker( PosSet: String );
  var i: Integer;
begin
 for i := 0 to Columns.Count - 1 do
  if i < Length( PosSet ) then
    case PosSet[i+1] of
      'U','u' : Columns[i].Title.SortMarker := smUpEh;
      'D','d' : Columns[i].Title.SortMarker := smDownEh
      else Columns[i].Title.SortMarker := smNoneEh;
    end
  else Columns[i].Title.SortMarker := smNoneEh;
end;
//==================================================================================================
procedure TcoDBGridEh.ClearMarker;
  var i: Integer;
begin
 for i := 0 to Columns.Count - 1 do Columns[i].Title.SortMarker := smNoneEh;
end;
//==================================================================================================
procedure TcoDBGridEh.SetShowHintData( const Value: Boolean );
begin
  if FShowHintData <> Value then begin
    if Value and ( not Assigned( OnDblClick )) then begin
      FShowHintData := True;
      ShowHint      := True;
      OnDblClick    := HintDataShow;
    end else if FShowHintData then begin
      OnDblClick    := nil;
      FShowHintData := False;
    end else begin
      FShowHintData := False;
    end;
  end;
end;
//==================================================================================================
procedure TcoDBGridEh.HintDataShow( Sender: TObject );
begin
  FShowHintDataActive := True;
end;
//==================================================================================================
procedure TcoDBGridEh.CMHintShow( var Message: TCMHintShow );
  var Data: String;
      i: Integer;
begin
  if FShowHintDataActive then begin
    Data := '';
    with DataSource.DataSet do
      for i := 0 to FieldCount - 1
        do if Fields[i].Visible then Data := Data + iif( i = 0, '', #10 ) + Fields[i].DisplayLabel + ': ' + Fields[i].DisplayText;
    Message.HintInfo.HintStr      := Data;
//    Message.HintInfo.HideTimeout  := 0;
    FShowHintDataActive := False;
  end;
  inherited
end;
//==================================================================================================
function TcoDBGridEh.ColumnByField( Field: TField ): TColumnEh;
  var i: Integer;
begin
  Result := Columns[ 0 ];
  for i := 0 to ColCount - 1 do
    if Columns[ i ].Field = Field then Result := Columns[ i ]
end;
//==================================================================================================
procedure TcoDBGridEh.SortMarkingChanged;
  var i: Integer;
      s: String;
      b: TBookmark;
begin
  Screen.Cursor := crHourGlass;
  b := DataSource.DataSet.GetBookmark;
  s := '';
  for i := 0 to SortMarkedColumns.Count - 1 do begin
    if Empty( SortMarkedColumns[i].EditMask )
      then s := s + SortMarkedColumns[i].FieldName
      else s := s + SortMarkedColumns[i].EditMask;
    if ( SortMarkedColumns[i].Title.SortMarker = smDownEh )
      then s := s + ' DESC';
    s := s + ';';
  end;
  TADODataSet(DataSource.DataSet).IndexFieldNames := s;
  DataSource.DataSet.GotoBookmark( b );
  Screen.Cursor := crDefault;
end;
//==================================================================================================
end.
