{==================================================================================================

  Comming Library

     "".

===================================================================================================

  :

  10.12.2002 TSK     CommingDesign  CommingRuntime
  14.11.2002 TSK   TAB.
  18.04.2001 TSK    .
  12.10.1999 TSK   .  .
  09.09.1998 TSK   Comming Library.

===================================================================================================
  Copyright 1998-2002 TSK Developer Laboratory, Moscow.             Konstantin Tsvetkov
                                                                           e-mail: tsklab@mail.ru
===================================================================================================}
unit CommingWaitForm;

interface

uses  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
      Gauges, ExtCtrls, StdCtrls, CommingTracer, CommingOther, Buttons;

type
  TcoFForm = class(TForm)
    FBar: TPanel;
    FProgress: TGauge;
    FDescription: TLabel;
    FIcon: TImage;
    FTracer: TcoTracer;
    FButton: TSpeedButton;
//    procedure FormCreate(Sender: TObject);
    procedure FButtonClick(Sender: TObject);
  private    { Private declarations }
    FContinuation: Boolean;
    FsIcon: Boolean;
    FsBar: Boolean;
    FsBreak: Boolean;
    FsTracer: Boolean;
    procedure  Figuration;
  public     { Public declarations }
    procedure CreateParams(var Params: TCreateParams); override;
  end;

  TcoShowControl = ( scIcon, scProgressBar, scBreakButton, scTimeTracer );
  TcoShowControls = set of TcoShowControl;

  TcoWaitForm = class(TComponent)
  private     { Private declarations }
    FForm: TcoFForm;
    FShowControls: TcoShowControls;
    FMaxValue: Integer;
    FActive: Boolean;
    function     GetCaption(): String;
    procedure    SetCaption( Value: String );
    function     GetContinuation(): Boolean;
    procedure    SetContinuation( Value: Boolean );
    procedure    SetShowControls( Value: TcoShowControls );
    procedure   SetActive(const Value: Boolean);
    function    GetActive: Boolean;
    procedure   SetMaxValue( const Value: Integer );
  protected   { Protected declarations }
  public      { Public declarations }
    constructor Create(AOwner: TComponent); override;
//    destructor  Destroy; override;
    procedure    ShowOn;
    procedure    ShowOff;
    procedure    NextStep;
    procedure   Actuation( Value: Integer = -1 );
  published   { Published declarations }
    property    Description: String read GetCaption write SetCaption;
    property    Continuation: Boolean read GetContinuation write SetContinuation;
    property    ShowControls: TcoShowControls read FShowControls write SetShowControls;
    property    MaxValue: Integer read FMaxValue write SetMaxValue;
    property    Active: Boolean read GetActive write SetActive;
  end;

procedure Register;

implementation
{$R *.DFM}
//==================================================================================================
constructor TcoWaitForm.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if csDesigning in ComponentState  
    then FForm  := TcoFForm.Create(Application)
    else FForm  := TcoFForm.Create(AOwner);
  Continuation  := True;
  ShowControls  := [scProgressBar, scBreakButton, scTimeTracer];
end;
//==================================================================================================
procedure TcoWaitForm.ShowOn;
begin
  Continuation := True;
  with FForm do begin
    Figuration;
    if FsBar then
      with FProgress do begin
        Progress  := 0;
        MaxValue   := FMaxValue;
      end;
    if FsTracer then FTracer.Actuation( FMaxValue );
    Show;
    Update;
  end;
end;
//==================================================================================================
procedure TcoWaitForm.ShowOff;
begin
  FForm.Close;
end;
//==================================================================================================
procedure TcoWaitForm.NextStep;
begin
  with FForm do begin
    FProgress.Progress := FProgress.Progress + 1;
    if FsTracer then FTracer.NextStep;
    Application.ProcessMessages;
  end;
end;
//==================================================================================================
function TcoWaitForm.GetCaption(): String;
begin
  Result := FForm.FDescription.Caption;
end;
//==================================================================================================
procedure TcoWaitForm.SetCaption( Value: String );
begin
  FForm.FDescription.Caption := Value;
  FForm.Update;
end;
//==================================================================================================
function TcoWaitForm.GetContinuation(): Boolean;
begin
  Result := FForm.FContinuation;
end;
//==================================================================================================
procedure TcoWaitForm.SetContinuation( Value: Boolean );
begin
  FForm.FContinuation := Value;
end;
//==================================================================================================
procedure TcoWaitForm.SetShowControls( Value:  TcoShowControls );
begin
  FShowControls   := Value;
  FForm.FsIcon     := ( scIcon in Value );
  FForm.FsBar      := ( scProgressBar in Value );
  FForm.FsBreak    := ( scBreakButton in Value ) and FForm.FsBar;
  FForm.FsTracer  := ( scTimeTracer in Value ) and FForm.FsBar;
  FForm.Figuration;
end;
//==================================================================================================
{procedure TcoFForm.FormCreate(Sender: TObject);
begin
{  SetWindowLong( Handle, GWL_STYLE, GetWindowLong( Handle, GWL_STYLE) and not WS_CAPTION);
  SetWindowPos( Handle, HWND_TOP, Left, Top, Width, Height, SWP_DRAWFRAME );}
{end;}
//==================================================================================================
procedure TcoFForm.Figuration;
begin
  with FIcon do Visible := FsIcon;
  with FTracer do Visible := FsTracer;
  with FDescription do begin
    Left       := iif( FsIcon, 48, 8 );
    Width      := 277 + iif( FsIcon, 0, 40 ) + iif( FsTracer, 0, 96 );
    Height    := iif( FsIcon or FsTracer, 33, 17 );
    Layout    := iif( FsBar, tlBottom, tlCenter );
    Alignment  := iif( FsBar, taLeftJustify, taCenter );
  end;
  with FBar do begin
    Visible := FsBar;
    Top      := iif( FsIcon or FsTracer, 48, 32 );
  end;
  with FButton do begin
    Visible := FsBreak;
    Top      := iif( FsIcon or FsTracer, 48, 32 );
  end;
  Height   := iif( FsBar, iif( FsIcon or FsTracer, 89 , 73 ), 57 );
  Width    := iif( FsBreak or FsTracer, 477, 437 );
end;
//==================================================================================================
procedure TcoFForm.FButtonClick(Sender: TObject);
begin
  FContinuation := False;
end;
//==================================================================================================
procedure TcoWaitForm.SetActive( const Value: Boolean );
begin
  if ( FActive <> Value ) and ( FForm.Active <> Value ) then begin
    if Value then ShowOn else ShowOff;
    FActive := FForm.Active;
  end;
end;
//==================================================================================================
function TcoWaitForm.GetActive: Boolean;
begin
  Result := FForm.Active;
end;
//==================================================================================================
procedure TcoFForm.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do Style := ( Style OR WS_POPUP ) AND NOT WS_DLGFRAME;
end;
//==================================================================================================
{destructor TcoWaitForm.Destroy;
begin
  FForm.Destroy;
  inherited;
end;}
//==================================================================================================
procedure Register;
begin
  RegisterComponents( 'Comming', [TcoWaitForm] );
end;
//==================================================================================================
procedure TcoWaitForm.SetMaxValue(const Value: Integer);
begin
  FMaxValue                 := Value;
  FForm.FProgress.MaxValue  := Value;
  FForm.FTracer.Actuation( Value )
end;
//==================================================================================================
procedure TcoWaitForm.Actuation( Value: Integer );
begin
  if Value < 0 then Value := FMaxValue;
  FForm.FProgress.Progress := 0;
  MaxValue := Value;
end;
//==================================================================================================
end.
